@@ -18,6 +18,7 @@
    #:mpz-powm
    #:mpz-pow
    #:mpz-gcd
+   #:mpz-divisible-p
    #:mpz-lcm
    #:mpz-sqrt
    #:mpz-probably-prime-p
@@ -56,6 +57,8 @@
    ))
 
 (in-package "SB-GMP")
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (setf (sb-int:system-package-p *package*) t))
 
 (defvar *gmp-disabled* nil)
 
@@ -98,11 +101,11 @@
   #+(and win32 x86-64) unsigned-long-long)
 
 (deftype ui ()
-  #-(and win32 x86-64) '(unsigned-byte #.sb-vm:n-word-bits)
+  #-(and win32 x86-64) 'sb-vm:word
   #+(and win32 x86-64) '(unsigned-byte 32))
 
 (deftype si ()
-  #-(and win32 x86-64) '(signed-byte #.sb-vm:n-word-bits)
+  #-(and win32 x86-64) 'sb-vm:signed-word
   #+(and win32 x86-64) '(signed-byte 32))
 
 (define-alien-type nil
@@ -118,7 +121,7 @@
 ;; at some particular point can use mpz_realloc2, or clear variables
 ;; no longer needed."
 ;;
-;; We can therefore allocate a bignum of sufficiant size and use the
+;; We can therefore allocate a bignum of sufficient size and use the
 ;; space for GMP computations without the need for memory transfer
 ;; from C to Lisp space.
 (declaim (inline z-to-bignum z-to-bignum-neg))
@@ -176,14 +179,14 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
             (mp_num (struct gmpint))
             (mp_den (struct gmpint))))
 
-;;; memory initialization functions to support non-alloced results
+;;; Memory initialization functions to support non-allocated results
 ;;; since an upper bound cannot always correctly predetermined
 ;;; (e.g. the memory required for the fib function exceed the number
 ;;; of limbs that are be determined through the infamous Phi-relation
 ;;; resulting in a memory access error.
 
-;; use these for non-prealloced bignum values, but only when
-;; ultimately necessary since copying back into bignum space a the end
+;; Use these for non-preallocated bignum values, but only when
+;; ultimately necessary since copying back into bignum space at the end
 ;; of the operation is about three times slower than the shared buffer
 ;; approach.
 (declaim (inline __gmpz_init __gmpz_clear))
@@ -244,6 +247,7 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
 (declaim (inline __gmpz_mul_2exp
                  __gmpz_fdiv_q_2exp
                  __gmpz_pow_ui
+                 __gmpz_divisible_p
                  __gmpz_probab_prime_p
                  __gmpz_fac_ui
                  __gmpz_2fac_ui
@@ -268,6 +272,10 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
   (b (* (struct gmpint)))
   (e unsigned-long))
 
+(define-alien-routine __gmpz_divisible_p int
+  (n (* (struct gmpint)))
+  (d (* (struct gmpint))))
+
 (define-alien-routine __gmpz_probab_prime_p int
   (n (* (struct gmpint)))
   (reps int))
@@ -324,13 +332,13 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
 
 ;;;; SBCL interface
 
-;;; utility macros for GMP mpz variable and result declaration and
+;;; Utility macros for GMP mpz variable and result declaration and
 ;;; incarnation of associated SBCL bignums
 
 (defmacro with-mpz-results (pairs &body body)
   (loop for (gres size) in pairs
         for res = (gensym "RESULT")
-        collect `(when (> ,size sb-kernel::maximum-bignum-length)
+        collect `(when (> ,size sb-kernel:maximum-bignum-length)
                    (error "Size of result exceeds maxim bignum length")) into checks
         collect `(,gres (struct gmpint)) into declares
         collect `(,res (%allocate-bignum ,size))
@@ -390,7 +398,7 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
         collect `(__gmpz_init (addr ,gres)) into inits
         collect `(,size (abs (slot ,gres 'mp_size)))
           into resinits
-        collect `(when (> ,size (1- sb-kernel::maximum-bignum-length))
+        collect `(when (> ,size (1- sb-kernel:maximum-bignum-length))
                    (error "Size of result exceeds maxim bignum length")) into checks
         collect `(,res (%allocate-bignum (1+ ,size)))
           into resallocs
@@ -459,6 +467,14 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
                  (/= 0 (slot result 'mp_size)))
         (__gmpz_add (addr result) (addr result) (addr gb))))))
 
+(defgmpfun mpz-divisible-p (n d)
+  "Returns T if (ZEROP (MOD N D))."
+  (with-mpz-vars ((n gn) (d gd))
+    (not
+      (zerop
+        (__gmpz_divisible_p (addr gn) (addr gd))))))
+
+
 (defgmpfun mpz-cdiv (n d)
   (let ((size (1+ (max (blength n)
                        (blength d)))))
@@ -596,7 +612,7 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
 
 (defgmpfun mpz-fib2 (n)
   ;; (let ((size (1+ (ceiling (* n (log 1.618034 2)) 64)))))
-  ;; fibonacci number magnitude in bits is assymptotic to n(log_2 phi)
+  ;; fibonacci number magnitude in bits is asymptotic to n(log_2 phi)
   ;; This is correct for the result but appears not to be enough for GMP
   ;; during computation (memory access error), so use GMP-side allocation.
   (check-type n ui)
@@ -799,7 +815,7 @@ pre-allocated bignum. The allocated bignum-length must be (1+ COUNT)."
                              (bignum-data-sap bd))
                        (,gmpfun (addr r) (addr arga) (addr argb)))))
                  (locally (declare (optimize (speed 1)))
-                   (sb-kernel::build-ratio (if (minusp (slot (slot r 'mp_num) 'mp_size))
+                   (sb-kernel:build-ratio (if (minusp (slot (slot r 'mp_num) 'mp_size))
                                                (z-to-bignum-neg num size)
                                                (z-to-bignum num size))
                                            (z-to-bignum den size)))))))))))
