@@ -100,8 +100,8 @@
 ;;; variable.
 (eval-when (:compile-toplevel :load-toplevel)
   (setf *c-functions-in-runtime*
-        #+netbsd '("stat" "lstat" "fstat" "readdir" "opendir")
-        #-netbsd '()))
+        (append #+netbsd '("stat" "lstat" "fstat" "readdir" "opendir")
+                #+(and (or arm64 riscv ppc ppc64) linux) '("stat" "lstat" "fstat"))))
 
 
 ;;; filesystem access
@@ -186,7 +186,6 @@
   (define-call "sync" void never-fails)
   (define-call ("truncate" :options :largefile)
       int minusp (pathname filename) (length off-t))
-  #-win32
   (macrolet ((def-mk*temp (lisp-name c-name result-type errorp dirp values)
                (declare (ignore dirp))
                (if (sb-sys:find-foreign-symbol-address c-name)
@@ -231,11 +230,11 @@
     ;; FIXME: What about Windows?
     (def-mk*temp mkdtemp "mkdtemp" (* char) null-alien t nil))
   (define-call-internally ioctl-without-arg "ioctl" int minusp
-                          (fd file-descriptor) (cmd int))
+                          (fd file-descriptor) (cmd unsigned-long))
   (define-call-internally ioctl-with-int-arg "ioctl" int minusp
-                          (fd file-descriptor) (cmd int) (arg int))
+                          (fd file-descriptor) (cmd unsigned-long) (arg int))
   (define-call-internally ioctl-with-pointer-arg "ioctl" int minusp
-                          (fd file-descriptor) (cmd int)
+                          (fd file-descriptor) (cmd unsigned-long)
                           (arg alien-pointer-to-anything-or-nil))
   (define-entry-point "ioctl" (fd cmd &optional (arg nil argp))
     (if argp
@@ -342,14 +341,27 @@
 the child process in the parent. Forking while multiple threads are running is
 not supported."
     (tagbody
-       (sb-thread::with-all-threads-lock
-         (when (cdr sb-thread::*all-threads*)
-           (go :error))
-         (let ((pid (posix-fork)))
-           #+darwin
-           (when (= pid 0)
-             (darwin-reinit))
-           (return-from fork pid)))
+       (let ()
+         #+sb-thread
+         (sb-impl::finalizer-thread-stop)
+         ;; Acquiring sb-thread::*make-thread-lock* prevents creation
+         ;; of new threads.
+         (sb-thread::with-system-mutex (sb-thread::*make-thread-lock*)
+           ;; Dead threads aren't pruned from *ALL-THREADS* until the Pthread join.
+           ;; Do that now so that the forked process has only the main thread
+           ;; in *ALL-THREADS* and nothing in *JOINABLE-THREADS*.
+           #+pauseless-threadstart (sb-thread::join-pthread-joinables #'identity)
+           (when (let ((avltree sb-thread::*all-threads*))
+                   (or (sb-thread::avlnode-left avltree)
+                       (sb-thread::avlnode-right avltree)))
+             (go :error))
+           (let ((pid (posix-fork)))
+             #+darwin
+             (when (= pid 0)
+               (darwin-reinit))
+             #+sb-thread
+             (setf sb-impl::*finalizer-thread* t)
+             (return-from fork pid))))
      :error
        (error "Cannot fork with multiple threads running.")))
   (export 'fork :sb-posix)
@@ -457,7 +469,6 @@ not supported."
  (define-call "wtermsig" int never-fails (status int))
  (define-call "wifstopped" boolean never-fails (status int))
  (define-call "wstopsig" int never-fails (status int))
- #+nil ; see alien/waitpid-macros.c
  (define-call "wifcontinued" boolean never-fails (status int)))
 
 ;;; mmap, msync
@@ -465,16 +476,15 @@ not supported."
 (progn
  (define-call ("mmap" :options :largefile) sb-sys:system-area-pointer
    (lambda (res)
-     (= (sb-sys:sap-int res) #.(1- (expt 2 sb-vm::n-machine-word-bits))))
+     (= (sb-sys:sap-int res) #.(1- (expt 2 sb-vm:n-machine-word-bits))))
    (addr sap-or-nil) (length size-t) (prot unsigned)
    (flags unsigned) (fd file-descriptor) (offset off-t))
 
  (define-call "munmap" int minusp
    (start sb-sys:system-area-pointer) (length unsigned))
 
-#-win32
-(define-call "msync" int minusp
-  (addr sb-sys:system-area-pointer) (length unsigned) (flags int)))
+ (define-call "msync" int minusp
+   (addr sb-sys:system-area-pointer) (length unsigned) (flags int)))
 #+win32
 (progn
   ;; No attempt is made to offer a full mmap-like interface on Windows.
@@ -817,10 +827,6 @@ not supported."
 
 ;;; environment
 
-(eval-when (:compile-toplevel :load-toplevel)
-  ;; Do this at compile-time as Win32 code below refers to it as
-  ;; sb-posix:getenv.
-  (export 'getenv :sb-posix))
 (defun getenv (name)
   (let ((r (alien-funcall
             (extern-alien "getenv" (function (* char) (c-string :not-null t)))
