@@ -41,14 +41,6 @@
 ;;; cdr is the layout itself.
 (defvar *!initial-layouts*)
 
-;;; a generator for random values suitable for the CLOS-HASH slots of
-;;; LAYOUTs. We use our own RANDOM-STATE here because we'd like
-;;; pseudo-random values to come the same way in the target even when
-;;; we make minor changes to the system, in order to reduce the
-;;; mysteriousness of possible CLOS bugs.
-(define-load-time-global *layout-clos-hash-random-state*
-    (make-random-state))
-
 ;;; a table mapping class names to layouts for classes we have
 ;;; referenced but not yet loaded. This is initialized from an alist
 ;;; created by genesis describing the layouts that genesis created at
@@ -62,9 +54,11 @@
   #-sb-xc-host (progn
                  (/show0 "processing *!INITIAL-LAYOUTS*")
                  (setq *forward-referenced-layouts* (make-hash-table :test 'equal))
-                 (setq *layout-clos-hash-random-state* (make-random-state))
                  (dovector (x *!initial-layouts*)
-                   (setf (layout-clos-hash (cdr x)) (random-layout-clos-hash))
+                   (let ((expected (hash-layout-name (car x)))
+                         (actual (layout-clos-hash (cdr x))))
+                     (unless (= actual expected)
+                       (bug "XC layout hash calculation failed")))
                    (setf (gethash (car x) *forward-referenced-layouts*)
                          (cdr x)))
                  (/show0 "done processing *!INITIAL-LAYOUTS*")))
@@ -82,22 +76,6 @@
   (defun layout-proper-name (layout)
     (classoid-proper-name (layout-classoid layout))))
 
-;;;; support for the hash values used by CLOS when working with LAYOUTs
-
-(defun random-layout-clos-hash ()
-  ;; FIXME: I'm not sure why this expression is (1+ (RANDOM FOO)),
-  ;; returning a strictly positive value. I copied it verbatim from
-  ;; CMU CL INITIALIZE-LAYOUT-HASH, so presumably it works, but I
-  ;; dunno whether the hash values are really supposed to be 1-based.
-  ;; They're declared as INDEX.. Or is this a hack to try to avoid
-  ;; having to use bignum arithmetic? Or what? An explanation would be
-  ;; nice.
-  ;;
-  ;; an explanation is provided in Kiczales and Rodriguez, "Efficient
-  ;; Method Dispatch in PCL", 1990.  -- CSR, 2005-11-30
-  (1+ (random (1- layout-clos-hash-limit)
-              *layout-clos-hash-random-state*)))
-
 ;;; If we can't find any existing layout, then we create a new one
 ;;; storing it in *FORWARD-REFERENCED-LAYOUTS*. In classic CMU CL, we
 ;;; used to immediately check for compatibility, but for
@@ -118,6 +96,7 @@
         (or (and classoid (classoid-layout classoid))
             (values (ensure-gethash name table
                                     (make-layout
+                                     (hash-layout-name name)
                                      (or classoid
                                          (make-undefined-classoid name))))))))))
 
@@ -134,24 +113,21 @@
                             layout-depthoid layout-bitmap) layout)
                 %init-or-check-layout))
 (defun %init-or-check-layout (layout classoid length flags inherits depthoid bitmap)
+  ;; BITMAP is redundant information about the primitive object representation
+  ;; which is fully captured by the defstruct description, and not stored in the
+  ;; host layout.
   (cond ((eq (layout-invalid layout) :uninitialized)
          ;; There was no layout before, we just created one which
          ;; we'll now initialize with our information.
-         (macrolet ((inherit (n) `(if (> struct-depth ,n) (svref inherits ,n) 0)))
-           ;; struct-depth is generally the depthoid when the depthoid is >0,
-           ;; but not for FILE-STREAM which has depthoid=4 but is not a structure.
-           (let ((struct-depth
-                  (if (logtest +structure-layout-flag+ flags) depthoid 0)))
-             (setf (layout-length layout) length
-                   (layout-flags layout) flags
-                   (layout-inherits layout) inherits
-                   (layout-depthoid layout) depthoid
-                   (layout-depth2-ancestor layout) (inherit 2)
-                   (layout-depth3-ancestor layout) (inherit 3)
-                   (layout-depth4-ancestor layout) (inherit 4)
-                   (layout-bitmap layout) bitmap
-                   (layout-classoid layout) classoid
-                   (layout-invalid layout) nil))))
+         (setf (layout-classoid layout) classoid)
+         (set-layout-inherits layout inherits)
+         (assign-layout-slots layout :depthoid depthoid :length length :flags flags)
+         #-sb-xc-host ; Assign target-only slots
+         (when (logtest +structure-layout-flag+ flags)
+           ;; STRING-STREAM and FILE-STREAM have depthoid=4 but are not structure types
+           (setf (layout-bitmap layout) bitmap))
+         ;; Finally, make it valid.
+         (setf (layout-invalid layout) nil))
         ;; FIXME: Now that LAYOUTs are born :UNINITIALIZED, maybe this
         ;; clause is not needed?
         ((not *type-system-initialized*)
@@ -175,7 +151,7 @@
 (defmethod make-load-form ((layout layout) &optional env)
   (declare (ignore env))
   (when (layout-invalid layout)
-    (sb-c::compiler-error "can't dump reference to obsolete class: ~S"
+    (sb-c:compiler-error "can't dump reference to obsolete class: ~S"
                           (layout-classoid layout)))
   (let* ((classoid (layout-classoid layout))
          (name (classoid-name classoid)))
@@ -187,7 +163,7 @@
                 (bug "xc MAKE-LOAD-FORM on undefined layout"))
                (t 0))))
     (unless name
-      (sb-c::compiler-error "can't dump anonymous LAYOUT: ~S" layout))
+      (sb-c:compiler-error "can't dump anonymous LAYOUT: ~S" layout))
     ;; Since LAYOUT refers to a class which refers back to the LAYOUT,
     ;; we have to do this in two stages, like the TREE-WITH-PARENT
     ;; example in the MAKE-LOAD-FORM entry in the ANSI spec.
@@ -197,13 +173,13 @@
      `(find-layout ',name)
      ;; "initialization" form (which actually doesn't initialize
      ;; preexisting LAYOUTs, just checks that they're consistent).
-     `(%init-or-check-layout ',layout
-                             ',(layout-classoid layout)
-                             ',(layout-length layout)
-                             ',(layout-flags layout)
-                             ',(layout-inherits layout)
-                             ',(layout-depthoid layout)
-                             ',(layout-bitmap layout)))))
+     `(%init-or-check-layout ,layout
+                             ,(layout-classoid layout)
+                             ,(layout-length layout)
+                             ,(layout-flags layout)
+                             ,(layout-inherits layout)
+                             ,(layout-depthoid layout)
+                             ,(layout-bitmap layout)))))
 
 ;;; If LAYOUT's slot values differ from the specified slot values in
 ;;; any interesting way, then give a warning and return T.
@@ -347,33 +323,37 @@ between the ~A definition and the ~A definition"
           (setf (classoid-subclasses classoid) nil)))
 
       (if destruct-layout
+          #+sb-xc-host (error "Why mutate a layout in XC host?")
+          #-sb-xc-host
           ;; Destructively modifying a layout is not threadsafe at all.
           ;; Use at your own risk (interactive use only).
           (let ((inherits (layout-inherits layout))
                 (depthoid (layout-depthoid layout)))
-            (aver (logtest +structure-layout-flag+ (layout-%bits layout)))
+            (aver (logtest +structure-layout-flag+ (layout-flags layout)))
             (aver (= (length inherits) depthoid))
-            (macrolet ((inherit (n) `(if (> depthoid ,n) (svref inherits ,n) 0)))
-              (setf (layout-invalid destruct-layout) nil
-                    (layout-inherits destruct-layout) inherits
-                    (layout-depthoid destruct-layout) (layout-depthoid layout)
-                    (layout-depth2-ancestor destruct-layout) (inherit 2)
-                    (layout-depth3-ancestor destruct-layout) (inherit 3)
-                    (layout-depth4-ancestor destruct-layout) (inherit 4)
-                    (layout-length destruct-layout) (layout-length layout)
-                    (layout-bitmap destruct-layout) (layout-bitmap layout)
-                    (layout-info destruct-layout) (layout-info layout)
-                    (classoid-layout classoid) destruct-layout)))
+            #-64-bit (setf (layout-depthoid destruct-layout) (layout-depthoid layout)
+                           (layout-length destruct-layout) (layout-length layout))
+            (setf (layout-flags destruct-layout) (layout-flags layout)
+                  (layout-info destruct-layout) (layout-info layout)
+                  (layout-bitmap destruct-layout) (layout-bitmap layout))
+            (set-layout-inherits destruct-layout inherits)
+            (setf (layout-invalid destruct-layout) nil
+                  (classoid-layout classoid) destruct-layout))
           (setf (layout-invalid layout) nil
                 (classoid-layout classoid) layout))
 
       (dovector (super-layout (layout-inherits layout))
         (let* ((super (layout-classoid super-layout))
-               (subclasses (or (classoid-subclasses super)
-                               (setf (classoid-subclasses super)
-                                     (make-hash-table :test 'eq
-                                                      #-sb-xc-host #-sb-xc-host
-                                                      :synchronized t)))))
+               (subclasses
+                 (or (classoid-subclasses super)
+                     (setf (classoid-subclasses super)
+                           #+sb-xc-host (make-hash-table :test 'eq)
+                           ;; Might as well use CTYPE-HASH-VALUE as a
+                           ;; stable hash since we have it.
+                           #-sb-xc-host
+                           (make-hash-table :hash-function #'type-hash-value
+                                            :test 'eq
+                                            :synchronized t)))))
           (when (and (eq (classoid-state super) :sealed)
                      (not (gethash classoid subclasses)))
             (warn "unsealing sealed class ~S in order to subclass it"
@@ -516,10 +496,8 @@ between the ~A definition and the ~A definition"
 ;;; don't have a corresponding class.
 (def!struct (structure-classoid (:include classoid)
                                 (:copier nil)
-                                (:constructor %make-structure-classoid
-                                    (hash-value name))))
-(defun make-structure-classoid (&key name)
-  (%make-structure-classoid (interned-type-hash name) name))
+                                (:constructor make-structure-classoid
+                                    (&key name &aux (%bits (pack-ctype-bits classoid name))))))
 
 ;;;; classoid namespace
 
@@ -551,12 +529,12 @@ between the ~A definition and the ~A definition"
                ;; modulo potential differences with respect to
                ;; conditions).
                #+sb-xc-host
-               (let ((old (class-of old-value))
-                     (new (class-of new-value)))
+               (let ((old (cl:class-of old-value))
+                     (new (cl:class-of new-value)))
                  (unless (eq old new)
                    (bug "Trying to change the metaclass of ~S from ~S to ~S in the ~
                             cross-compiler."
-                        name (class-name old) (class-name new))))
+                        name (cl:class-name old) (cl:class-name new))))
                #-sb-xc-host
                (let ((old (classoid-of old-value))
                      (new (classoid-of new-value)))
@@ -635,9 +613,10 @@ between the ~A definition and the ~A definition"
 ;;; Again, this should be compiler-only, but easier to make this
 ;;; thread-safe.
 (defun insured-find-classoid (name predicate constructor)
-  (declare (type function predicate constructor))
+  (declare (type function predicate)
+           (type (or function symbol) constructor))
   (let ((table *forward-referenced-layouts*))
-    (with-locked-system-table (table)
+    (with-system-mutex ((hash-table-lock table))
       (let* ((old (find-classoid name nil))
              (res (if (and old (funcall predicate old))
                       old
@@ -698,7 +677,7 @@ between the ~A definition and the ~A definition"
       (if (typep classoid 'standard-classoid)
           (let ((class (classoid-pcl-class classoid)))
             (cond
-              ((sb-pcl:class-finalized-p class)
+              ((sb-mop:class-finalized-p class)
                (sb-pcl::%force-cache-flushes class)
                t)
               ((sb-pcl::class-has-a-forward-referenced-superclass-p class)
@@ -710,7 +689,7 @@ between the ~A definition and the ~A definition"
                       error-context))
                nil)
               (t
-               (sb-pcl:finalize-inheritance class)
+               (sb-mop:finalize-inheritance class)
                t)))
           (bug "~@<Don't know how to ensure validity of ~S (not a STANDARD-CLASSOID) ~
                 for ~A.~%~:@>"
@@ -727,13 +706,6 @@ between the ~A definition and the ~A definition"
                  (%ensure-classoid-valid class2 layout2 errorp))
       (return-from %ensure-both-classoids-valid nil))))
 
-#-sb-xc-host ; No such thing as LAYOUT-OF, never mind the rest
-(defun update-object-layout-or-invalid (object layout)
-  ;; FIXME: explain why this isn't (LAYOUT-FOR-STD-CLASS-P LAYOUT).
-  (if (layout-for-std-class-p (layout-of object))
-      (sb-pcl::check-wrapper-validity object)
-      (sb-c::%layout-invalid-error object layout)))
-
 ;;; Simple methods for TYPE= and SUBTYPEP should never be called when
 ;;; the two classes are equal, since there are EQ checks in those
 ;;; operations.
@@ -899,19 +871,23 @@ between the ~A definition and the ~A definition"
      (system-area-pointer :codes (,sb-vm:sap-widetag)
                           :prototype-form (int-sap 0))
      (weak-pointer :codes (,sb-vm:weak-pointer-widetag)
-      :prototype-form (make-weak-pointer (find-package "CL")))
-     (code-component :codes (,sb-vm:code-header-widetag))
-     #-(or x86 x86-64) (lra :codes (,sb-vm:return-pc-widetag))
+      :prototype-form (make-weak-pointer 0))
+     (code-component :codes (,sb-vm:code-header-widetag)
+                     :prototype-form (fun-code-header #'identity))
+     #-(or x86 x86-64) (lra :codes (,sb-vm:return-pc-widetag)
+                            ;; Make the PROTOTYPE slot unbound.
+                            :prototype-form sb-pcl:+slot-unbound+)
      (fdefn :codes (,sb-vm:fdefn-widetag)
-            :prototype-form (make-fdefn nil))
-     (random-class) ; used for unknown type codes
-
+            :prototype-form (find-or-create-fdefn 'sb-mop:class-prototype))
+     (random-class ; used for unknown type codes
+            ;; Make the PROTOTYPE slot unbound.
+            :prototype-form sb-pcl:+slot-unbound+)
      (function
       :codes (,sb-vm:closure-widetag ,sb-vm:simple-fun-widetag)
       :state :read-only
-      :prototype-form (function (lambda () 42)))
+      :prototype-form #'identity)
 
-     (number :translation number)
+     (number :translation number :prototype-form 0)
      (complex
       :translation complex
       :inherits (number)
@@ -945,11 +921,9 @@ between the ~A definition and the ~A definition"
       :prototype-form
       ;; KLUDGE: doesn't work without AVX2 support from the CPU
       ;; (%make-simd-pack-256-ub64 42 42 42 42)
-      42)
-     (real :translation real :inherits (number))
-     (float
-      :translation float
-      :inherits (real number))
+      sb-pcl:+slot-unbound+)
+     (real :translation real :inherits (number) :prototype-form 0)
+     (float :translation float :inherits (real number) :prototype-form $0f0)
      (single-float
       :translation single-float
       :inherits (float real number)
@@ -967,18 +941,16 @@ between the ~A definition and the ~A definition"
       :codes (,sb-vm:long-float-widetag)
       :prototype-form $0L0)
      (rational
-      :translation rational
-      :inherits (real number))
+      :translation rational :inherits (real number) :prototype-form 0)
      (ratio
       :translation (and rational (not integer))
       :inherits (rational real number)
       :codes (,sb-vm:ratio-widetag)
       :prototype-form 1/42)
      (integer
-      :translation integer
-      :inherits (rational real number))
+      :translation integer :inherits (rational real number) :prototype-form 0)
      (fixnum
-      :translation (integer ,sb-xc:most-negative-fixnum ,sb-xc:most-positive-fixnum)
+      :translation (integer ,most-negative-fixnum ,most-positive-fixnum)
       :inherits (integer rational real number)
       :codes ,(mapcar #'symbol-value sb-vm::fixnum-lowtags)
       :prototype-form 42)
@@ -986,7 +958,7 @@ between the ~A definition and the ~A definition"
       :translation (and integer (not fixnum))
       :inherits (integer rational real number)
       :codes (,sb-vm:bignum-widetag)
-      :prototype-form ,(1+ sb-xc:most-positive-fixnum))
+      :prototype-form ,(1+ most-positive-fixnum))
 
      (array :translation array :codes (,sb-vm:complex-array-widetag)
             :hierarchical-p nil
@@ -1002,7 +974,8 @@ between the ~A definition and the ~A definition"
      (vector
       :translation vector :codes (,sb-vm:complex-vector-widetag)
       :direct-superclasses (array sequence)
-      :inherits (array sequence))
+      :inherits (array sequence)
+      :prototype-form (make-array 0 :adjustable t))
      (simple-vector
       :translation simple-vector :codes (,sb-vm:simple-vector-widetag)
       :direct-superclasses (vector simple-array)
@@ -1017,15 +990,17 @@ between the ~A definition and the ~A definition"
       :direct-superclasses (bit-vector simple-array)
       :inherits (bit-vector vector simple-array
                  array sequence)
-      :prototype-form (make-array 0 :element-type 'bit))
+      :prototype-form #*)
      (string
       :translation string
       :direct-superclasses (vector)
-      :inherits (vector array sequence))
+      :inherits (vector array sequence)
+      :prototype-form "")
      (simple-string
       :translation simple-string
       :direct-superclasses (string simple-array)
-      :inherits (string vector simple-array array sequence))
+      :inherits (string vector simple-array array sequence)
+      :prototype-form "")
      (vector-nil
       :translation (vector nil)
       :codes (,sb-vm:complex-vector-nil-widetag)
@@ -1069,7 +1044,8 @@ between the ~A definition and the ~A definition"
       :prototype-form (make-array 0 :element-type 'character))
      (list
       :translation (or cons (member nil))
-      :inherits (sequence))
+      :inherits (sequence)
+      :prototype-form 'nil)
      (cons
       :codes (,sb-vm:list-pointer-lowtag)
       :translation cons
@@ -1080,6 +1056,16 @@ between the ~A definition and the ~A definition"
       :inherits (symbol list sequence)
       :direct-superclasses (symbol list)
       :prototype-form 'nil)
+
+     ;; KLUDGE: the length must match the subsequent defstruct.
+     (pathname :depth 1
+               :length ,(+ 7 sb-vm:instance-data-start)
+               :prototype-form (make-pathname))
+     (logical-pathname :depth 2
+                       :length ,(+ 7 sb-vm:instance-data-start)
+                       :prototype-form (make-pathname :host "SYS")
+                       :inherits (pathname))
+
      ;; These last few are strange. STREAM has only T as an ancestor,
      ;; so you'd think it would be at depth 1. FILE- and STRING-STREAM
      ;; each have STREAM and T as ancestors, so you'd think they'd be at depth
@@ -1135,13 +1121,17 @@ between the ~A definition and the ~A definition"
               codes
               state
               depth
+              (length 0)
               prototype-form
               (hierarchical-p t) ; might be modified below
               (direct-superclasses (if inherits
                                      (list (car inherits))
                                      '(t))))
         x
-      (declare (ignore codes state translation prototype-form))
+      (declare (ignore codes state translation))
+      ;; instance metatypes and T don't need a prototype, everything else does
+      (unless (or prototype-form depth (eq name 't))
+        (error "Missing prototype in ~S" x))
       (let ((inherits-list (if (eq name t)
                                ()
                                (cons t (reverse inherits))))
@@ -1155,7 +1145,7 @@ between the ~A definition and the ~A definition"
                      (setf (classoid-cell-classoid
                             (find-classoid-cell name :create t))
                            (!make-built-in-classoid
-                             :hash-value (interned-type-hash name)
+                             :%bits (pack-ctype-bits classoid name)
                              :name name
                              :translation (if trans-p :initializing nil)
                              :direct-superclasses
@@ -1182,7 +1172,7 @@ between the ~A definition and the ~A definition"
            (find-and-init-or-check-layout name
                                           depthoid
                                           0 ; flags
-                                          0 ; length
+                                          length
                                           +layout-all-tagged+
                                           inherits-vector)
            :invalidate nil)))))
@@ -1216,8 +1206,12 @@ between the ~A definition and the ~A definition"
 ;;; use those slots as indexes into tables.
 (defun %invalidate-layout (layout)
   (declare (type layout layout))
-  (setf (layout-invalid layout) t
-        (layout-depthoid layout) -1)
+  #+sb-xc-host (warn "Why are we invalidating layout ~S?" layout)
+  (setf (layout-invalid layout) t)
+  (assign-layout-slots layout :depthoid -1)
+  ;; Ensure that the INVALID slot conveying ancillary data describing the
+  ;; invalidity reason is published before causing the invalid layout trap.
+  (sb-thread:barrier (:write))
   (setf (layout-clos-hash layout) 0)
   (let ((inherits (layout-inherits layout))
         (classoid (layout-classoid layout)))
