@@ -9,17 +9,16 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 (defmacro with-finalizer-store ((var) &body body)
-  `(let ((mutex (hash-table-lock (finalizer-id-map **finalizer-store**))))
-     ;; This does not inhibit GC, though the hashtable operations will,
-     ;; as is (currently) required for tables with non-null weakness.
-     (sb!thread::with-recursive-system-lock (mutex)
-       ;; Grab the store again inside the lock in case the array was enlarged
-       ;; after we referenced the mutex but before we acquired it.
-       (let ((,var **finalizer-store**))
-         ,@body))))
+  `(with-system-mutex ((hash-table-lock (finalizer-id-map **finalizer-store**)))
+     ;; Grab the global var inside the lock in case the array was enlarged
+     ;; after we referenced the mutex but before we acquired it.
+     ;; It's OK to reference the FINALIZER-ID-MAP because that is always
+     ;; in element 1 of the array regardless of what happens to the array.
+     (let ((,var **finalizer-store**))
+       ,@body)))
 
 (defmacro finalizer-recycle-bin (store) `(cdr (elt ,store 0)))
 (defmacro finalizer-id-map (store) `(elt ,store 1))
@@ -27,9 +26,8 @@
 
 (defun make-finalizer-store (array-length)
   (let* ((v (make-array (the index array-length)))
-         (ht (make-hash-table :test 'eq :weakness :key)))
-    (setf (%instance-ref ht (get-dsd-index hash-table flags))
-          (logior (%instance-ref ht (get-dsd-index hash-table flags)) 4))
+         (ht (make-system-hash-table :test 'eq :weakness :key :synchronized nil
+                                     :finalizer t)))
     ;; The recycle bin has a dummy item in front so that the simple-vector
     ;; is growable without messing up RUN-PENDING-FINALIZERS when it atomically
     ;; pushes items into the recycle bin - it is unaffected by looking at
@@ -119,7 +117,7 @@ Examples:
                    (t
                     (setq id (incf (finalizer-max-id store)))
                     (unless (< id (length store))
-                      (sb!thread:barrier (:write)
+                      (sb-thread:barrier (:write)
                         ;; We must completely copy the old vector into the new
                         ;; before publishing the new in **FINALIZER-STORE**.
                         ;; Perhaps a cleverer way to size up is to have a tree
@@ -134,6 +132,25 @@ Examples:
                    (gethash object (finalizer-id-map store)) id)))))
   object)
 
+(defun invalidate-fd-streams ()
+  (with-finalizer-store (store)
+    (maphash (lambda (object id)
+               (declare (ignore id))
+               (when (fd-stream-p object)
+                 (push (list object
+                             (ansi-stream-in object)
+                             (ansi-stream-bin object)
+                             (ansi-stream-n-bin object)
+                             (ansi-stream-out object)
+                             (ansi-stream-bout object)
+                             (ansi-stream-sout object)
+                             (ansi-stream-misc object))
+                       *streams-closed-by-slad*)
+                 ;; Nobody asked us to actually close the fd,
+                 ;; so just make it unusable.
+                 (set-closed-flame-by-slad object)))
+             (finalizer-id-map store))))
+
 (defun finalizers-deinit ()
   ;; remove :dont-save finalizers
   ;; Renumber the ID range as well, but leave the array size as-is. We could
@@ -143,8 +160,8 @@ Examples:
   (with-finalizer-store (old-store)
     (without-gcing
       (let ((new-store
-             (make-finalizer-store (max (1+ (finalizer-max-id old-store))
-                                        +finalizers-initial-size+)))
+              (make-finalizer-store (max (1+ (finalizer-max-id old-store))
+                                         +finalizers-initial-size+)))
             (old-objects (finalizer-id-map old-store)))
         (maphash (lambda (object old-id &aux (old (svref old-store old-id)))
                    ;; OLD is either a vector of finalizers or a single finalizer.
@@ -154,7 +171,7 @@ Examples:
                    (awhen (cond ((simple-vector-p old)
                                  (let ((new (remove-if #'consp old)))
                                    (case (length new)
-                                     (0 nil) ; all deleted
+                                     (0 nil)           ; all deleted
                                      (1 (svref new 0)) ; reduced to singleton
                                      (t new))))
                                 ((atom old) old)) ; a single finalizer to be saved
@@ -214,7 +231,10 @@ Examples:
     object))
 
 ;;; Drain the queue of finalizers and return when empty.
-;;; Concurrent invocations of this function are ok.
+;;; Concurrent invocations of this function in different threads are ok.
+;;; Nested invocations (from a GC forced by a finalizer) are not ok.
+;;; See the trace at the bottom of this file.
+(defvar *in-a-finalizer* nil)
 (defun scan-finalizers ()
   ;; This never acquires the finalizer store lock. Code accordingly.
   (let ((hashtable (finalizer-id-map **finalizer-store**)))
@@ -222,6 +242,8 @@ Examples:
      (let ((cell (hash-table-culled-values hashtable)))
        ;; This is like atomic-pop, but its obtains the first cons cell
        ;; in the list, not the car of the first cons.
+       ;; Possible TODO: when no other work remains, free the *JOINABLE-THREADS*,
+       ;; though MAKE-THREAD and JOIN-THREAD do that also, so there's no memory leak.
        (loop (unless cell (return-from scan-finalizers))
              (let ((actual (cas (hash-table-culled-values hashtable)
                                 cell (cdr cell))))
@@ -233,7 +255,7 @@ Examples:
               ;; so always load the vector again before dereferencing.
               (store **finalizer-store**)
               ;; I don't think we need a barrier; this has a data dependency
-              ;; on (CAR CELL) and STORE. (Alpha with threads, anyone?)
+              ;; on (CAR CELL) and STORE.
               (finalizers (svref store id))) ; [1] load
          (setf (svref store id) 0)           ; [2] store
          ;; The ID can be reused right away. Link it into the recycle list,
@@ -246,7 +268,7 @@ Examples:
          ;; Now call the function(s)
          (flet ((call (finalizer)
                   (let ((fun (if (consp finalizer) (car finalizer) finalizer)))
-                    (handler-case (funcall fun)
+                    (handler-case (let ((*in-a-finalizer* t)) (funcall fun))
                       (error (c)
                         (warn "Error calling finalizer ~S:~%  ~S" fun c))))))
            (if (simple-vector-p finalizers)
@@ -271,7 +293,7 @@ Examples:
          (cond ((simple-vector-p finalizers) (fill finalizers 0))
                ((consp finalizers) (rplaca finalizers 0))))))))
 
-#!+sb-thread
+#+sb-thread
 (progn
   ;; *FINALIZER-THREAD* is either a boolean value indicating whether to start a
   ;; thread, or a thread object (which is created no sooner than needed).
@@ -279,30 +301,37 @@ Examples:
   ;; between stopping the thread and writing to disk will be synchronous.
   ;; Restarting a saved core resets the flag to T.
   (define-load-time-global *finalizer-thread* t)
-  (declaim (type (or sb!thread:thread boolean) *finalizer-thread*))
+  (declaim (type (or sb-thread:thread boolean) *finalizer-thread*))
   (define-load-time-global *finalizer-queue-lock*
-      (sb!thread:make-mutex :name "finalizer"))
+      (sb-thread:make-mutex :name "finalizer"))
   (define-load-time-global *finalizer-queue*
-      (sb!thread:make-waitqueue :name "finalizer")))
+      (sb-thread:make-waitqueue :name "finalizer")))
 
 (defun run-pending-finalizers ()
   (when (hash-table-culled-values (finalizer-id-map **finalizer-store**))
-    (cond #!+sb-thread
+    (cond #+sb-thread
           ((%instancep *finalizer-thread*)
-           (sb!thread::with-system-mutex (*finalizer-queue-lock*)
-             (sb!thread:condition-notify *finalizer-queue*)))
-          #!+sb-thread
+           ;; Do absolutely nothing if in a finalizer. If we are then it's obvious
+           ;; that we're in the SCAN-FINALIZERS loop, and that the loop will just
+           ;; go on looping without further provocation.
+           (unless *in-a-finalizer*
+             (with-system-mutex (*finalizer-queue-lock*)
+               (sb-thread:condition-notify *finalizer-queue*))))
+          #+sb-thread
           ((eq *finalizer-thread* t) ; Create a new thread
-           (sb!thread:make-thread
+           (sb-thread::make-ephemeral-thread
+            "finalizer"
             (lambda ()
+              ;; If we already called FINALIZER-THREAD-STOP, then
+              ;; *FINALIZER-THREAD* is NIL, and this WHEN test is false.
               (when (eq t (cas *finalizer-thread* t
-                               sb!thread:*current-thread*))
+                               sb-thread:*current-thread*))
                 ;; Don't enter the loop if this thread lost the
                 ;; competition to become a finalizer thread.
                 (loop
                   (scan-finalizers)
                   ;; Wait for a notification
-                  (sb!thread::with-system-mutex (*finalizer-queue-lock*)
+                  (with-system-mutex (*finalizer-queue-lock*)
                     ;; Don't go to sleep if *FINALIZER-THREAD* became NIL
                     (unless *finalizer-thread*
                       (return))
@@ -310,10 +339,9 @@ Examples:
                     ;; since it is always legal to call SCAN-FINALIZERS
                     ;; even when it has nothing to do.
                     ;; Spurious wakeup is of no concern to us here.
-                    (sb!thread:condition-wait
+                    (sb-thread:condition-wait
                      *finalizer-queue* *finalizer-queue-lock*)))))
-            :name "finalizer"
-            :ephemeral t))
+            nil))
           (t
            (scan-finalizers)))))
 
@@ -321,13 +349,186 @@ Examples:
 ;;; Make no attempt to drain the queue of pending finalizers.
 ;;; (When called from EXIT, the user must invoke a final GC if there is
 ;;; an expectation that GC-based things run. Similarly when saving a core)
+;;; There's a data race involved during SAVE-LISP-AND-DIE (or SB-POSIX:FORK)
+;;; because *FINALIZER-THREAD* is assigned by that thread itself which means that
+;;; it has to run before we can detect that it exists.
+;;; There are few remedies involving some combination of these steps:
+;;; - Assign a THREAD instance into *finalizer-thread* before it shows
+;;;   up in *ALL-THREADS*. If *FINALIZER-THREAD* has a value but is not in
+;;;   *ALL-THREADS*, then it may be blocked on the mutex. Can we terminate it?
+;;; - During save, acquire the *MAKE-THREAD-LOCK* to get a consistent view of
+;;;   which threads are starting and running.
+;;; - the :pauseless-threadstart code adds new threads to *ALL-THREADS*
+;;;   before starting the posix thread. We could look there.
+;;;   The important thing is to avoid thinking that there is a thread running
+;;;   when it doesn't have a posix thread ID yet.
+;;; - Does acquiring the *MAKE-THREAD-LOCK* before assigning *FINALIZER-THREAD*
+;;;   help anything? Perhaps.
+;;; But first I'd like to entirely eliminate the non-pauseless-threadstart code
+;;; so that we don't have to reason about two totally different solutions.
 (defun finalizer-thread-stop ()
-  #!+sb-thread
+  #+sb-thread
   (let ((thread *finalizer-thread*))
-    (when thread ; if it was NIL it can't become a thread. So there's no race.
-      ;; Setting to NIL causes the thread to exit after waking
-      (setq thread (cas *finalizer-thread* thread nil))
-      (when (%instancep thread) ; only if it was a thread, do this
-        (sb!thread::with-system-mutex (*finalizer-queue-lock*)
-          (sb!thread:condition-notify *finalizer-queue*))
-        (sb!thread:join-thread thread))))) ; wait for it
+    ;; valid state transitions:
+    ;;   T to a #<thread>
+    ;;   #<thread> to NIL
+    ;;   T to NIL
+    ;; If it was NIL, it is latched at that state, so we can ignore it.
+    (when (null thread)
+      (return-from finalizer-thread-stop))
+    ;; If user calls SAVE-LISP-AND-DIE or SB-POSIX:FORK in a finalizer...
+    (when (eq sb-thread:*current-thread* *finalizer-thread*)
+      (error "Can't stop finalizer thread from inside finalizer thread"))
+    (let ((oldval (cas *finalizer-thread* thread nil)))
+      ;; If GC had started the thread, and it got to its main body only after
+      ;; we observed *FINALIZER-THREAD* = T, then we could see a transition
+      ;; from T to #<thread> now. That is unlikely, but we must try once
+      ;; again to CAS the value to NIL.
+      (unless (eq oldval thread)
+        (setq thread oldval)
+        ;; Also, we could be racing with someone else trying to stop the
+        ;; thread, so we could see T | #<thread> -> NIL.
+        (when (null thread)
+          (return-from finalizer-thread-stop))
+        (aver (eq (cas *finalizer-thread* thread nil) thread))))
+    (when (%instancep thread)
+      ;; The finalizer thread will exit when we wake it up
+      ;; and it sees that *FINALIZER-THREAD* is NIL.
+      (with-system-mutex (*finalizer-queue-lock*)
+        (sb-thread:condition-notify *finalizer-queue*))
+      (sb-thread:join-thread thread)))) ; wait for it
+
+#|
+;;; This is a display produced by annotating parts of gc-common.c and
+;;; interrupt.c with each thread's output in its own column.
+;;; The main thread is on the right.
+
+;;; This output shows that if the finalizer thread calls a function that
+;;; triggers a GC, the interrupt nesting depth can grow without limit.
+;;; The finalizer thread does not have to be a memory hog - it just has
+;;; to be unlucky enough to be the thread that triggers the collection.
+;;; _Any_ thread can bring the GC trigger up to the threshold,
+;;; and as long as the finalizer thread is the one to cross the
+;;; the threshold, it is tasked with triggering the next scan
+;;; of finalizers, but it MUST NOT do so recursively.
+;;;
+;;; The same thing can happen without using a finalizer thread,
+;;; but it's actually easier to understand the output this way.
+
+Thread 2                                     Main
+--------                                     --------
+|                                            Enter SB-EXT:GC 0
+|                                            Stopping world
+|                                            Stopped world
+|                                            SUB-GC calling gc(0)
+|                                            set auto_gc_trig=4858833
+|                                            completed GC
+|                                            Restarted world
+|                                            SB-EXT:GC calling POST-GC
+|                                            ENTER run-pending-finalizers
+|                                            Enter SB-EXT:GC 0
+|                                            Stopping world
+|                                            Stopped world
+|                                            SUB-GC calling gc(6)
+|                                            set auto_gc_trig=5075473
+|                                            completed GC
+|                                            Restarted world
+|                                            SB-EXT:GC calling POST-GC
+|                                            ENTER run-pending-finalizers
+|                                            Trying to start finalizer thread
+| starting
+|                                            Enter SB-EXT:GC 0
+|                                            Stopping world
+| Caught SIGUSR2, pc=52a946bd
+| STOP_FOR_GC PA=1 inh=N sigmask=0
+| Caught SIGILL, pc=52a946cc code 0x9
+| evt 1 >handle_pending
+| STOP_FOR_GC PA=0 inh=N sigmask=0
+| fake ffcall "stop_for_gc"
+| bind(free_ICI, 1)
+|                                            Stopped world
+|                                            SUB-GC calling gc(6)
+|                                            set auto_gc_trig=520eeb3
+|                                            completed GC
+|                                            Restarted world
+|                                            SB-EXT:GC calling POST-GC
+|                                            ENTER run-pending-finalizers
+|                                            Notify finalizer thread
+| unbind free_ICI -> 0
+| leave STOP_FOR_GC
+| evt 1 <handle_pending
+| gc_trig=52bcf80, setting PA-int
+| Caught SIGILL, pc=52a946cc code 0x9
+| evt 2 >handle_pending
+| ENTER maybe_gc (pc was 52a946ce)
+| fake ffcall "maybe_gc"
+| bind(free_ICI, 1)
+| maybe_gc calling SUB-GC
+| Stopping world
+|                                            Caught SIGUSR2, pc=7ff010deef47
+|                                            STOP_FOR_GC PA=0 inh=N sigmask=0
+|                                            fake ffcall "stop_for_gc"
+|                                            bind(free_ICI, 1)
+| Stopped world
+| SUB-GC calling gc(0)
+| set auto_gc_trig=552b033
+| completed GC
+| Restarted world
+| maybe_gc calling POST-GC
+| ENTER run-pending-finalizers
+|                                            unbind free_ICI -> 0
+|                                            leave STOP_FOR_GC
+| gc_trig=55d9110, setting PA-int
+| Caught SIGILL, pc=52a946cc code 0x9
+| evt 3 >handle_pending
+| ENTER maybe_gc (pc was 52a946ce)
+| fake ffcall "maybe_gc"
+| bind(free_ICI, 2)
+| maybe_gc calling SUB-GC
+| Stopping world
+|                                            Caught SIGUSR2, pc=7ff010deef47
+|                                            STOP_FOR_GC PA=0 inh=N sigmask=0
+|                                            fake ffcall "stop_for_gc"
+|                                            bind(free_ICI, 1)
+| Stopped world
+| SUB-GC calling gc(0)
+| set auto_gc_trig=56c0423
+| completed GC
+| Restarted world
+| maybe_gc calling POST-GC
+| ENTER run-pending-finalizers
+|                                            unbind free_ICI -> 0
+|                                            leave STOP_FOR_GC
+| gc_trig=576e500, setting PA-int
+| Caught SIGILL, pc=52a946cc code 0x9
+| evt 4 >handle_pending
+| ENTER maybe_gc (pc was 52a946ce)
+| fake ffcall "maybe_gc"
+| bind(free_ICI, 3)
+| maybe_gc calling SUB-GC
+| Stopping world
+|                                            Caught SIGUSR2, pc=7ff010deef47
+|                                            STOP_FOR_GC PA=0 inh=N sigmask=0
+|                                            fake ffcall "stop_for_gc"
+|                                            bind(free_ICI, 1)
+| Stopped world
+| SUB-GC calling gc(0)
+| set auto_gc_trig=59dc213
+| completed GC
+| Restarted world
+| maybe_gc calling POST-GC
+| ENTER run-pending-finalizers
+|                                            unbind free_ICI -> 0
+|                                            leave STOP_FOR_GC
+| gc_trig=5a8a2f0, setting PA-int
+| Caught SIGILL, pc=52a946cc code 0x9
+| evt 5 >handle_pending
+| ENTER maybe_gc (pc was 52a946ce)
+| fake ffcall "maybe_gc"
+| bind(free_ICI, 4)
+| maybe_gc calling SUB-GC
+| Stopping world
+|
+;;; This pattern of binding FREE_INTERRUPT_CONTEXT_INDEX to successively
+;;; higher values will continue forever until reaching the limit and crashing.
+|#
