@@ -187,6 +187,7 @@
   #+win32 (handle nil :type (or null (signed-byte 32)))
   #-win32
   serve-event-pipe)
+(declaim (freeze-type process))
 
 (defmethod print-object ((process process) stream)
   (print-unreadable-object (process stream :type t)
@@ -424,11 +425,10 @@ status slot."
 
   (defun find-a-pty ()
     ;; First try to use the Unix98 pty api.
-    (let* ((master-name (coerce (format nil "/dev/ptmx") 'base-string))
-           (master-fd (sb-unix:unix-open master-name
-                                         (logior sb-unix:o_rdwr
-                                                 sb-unix:o_noctty)
-                                         #o666)))
+    (let ((master-fd (sb-unix:unix-open "/dev/ptmx"
+                                        (logior sb-unix:o_rdwr
+                                                sb-unix:o_noctty)
+                                        #o666)))
       (when master-fd
         (grantpt master-fd)
         (unlockpt master-fd)
@@ -447,15 +447,15 @@ status slot."
     ;; No dice, try using the old-school method.
     (dolist (char '(#\p #\q))
       (dotimes (digit 16)
-        (let* ((master-name (coerce (format nil "/dev/pty~C~X" char digit)
-                                    'base-string))
+        (let* ((master-name (with-output-to-string (str nil :element-type 'base-char)
+                              (format str "/dev/pty~C~X" char digit)))
                (master-fd (sb-unix:unix-open master-name
                                              (logior sb-unix:o_rdwr
                                                      sb-unix:o_noctty)
                                              #o666)))
           (when master-fd
-            (let* ((slave-name (coerce (format nil "/dev/tty~C~X" char digit)
-                                       'base-string))
+            (let* ((slave-name (with-output-to-string (str nil :element-type 'base-char)
+                                 (format str "/dev/tty~C~X" char digit)))
                    (slave-fd (sb-unix:unix-open slave-name
                                                 (logior sb-unix:o_rdwr
                                                         sb-unix:o_noctty)
@@ -541,9 +541,9 @@ status slot."
       (declare (type (simple-array (unsigned-byte 8) (*)) octets))
       (let ((size (length octets)))
         ;; Copy string.
-        (sb-kernel:copy-ub8-to-system-area octets 0 string-sap 0 size)
+        (copy-ub8-to-system-area octets 0 string-sap 0 size)
         ;; NULL-terminate it
-        (sb-kernel:system-area-ub8-fill 0 string-sap size 4)
+        (system-area-ub8-fill 0 string-sap size 4)
         ;; Put the pointer in the vector.
         (setf (sap-ref-sap vec-sap vec-index-offset) string-sap)
         ;; Advance string-sap for the next string.
@@ -630,7 +630,7 @@ status slot."
 
 #+win32
 (defun prepare-args (args escape)
-  (with-simple-output-to-string (str)
+  (%with-output-to-string (str)
     (loop for (arg . rest) on args
           do
           (cond ((and escape
@@ -1105,7 +1105,7 @@ Users Manual for details about the PROCESS structure.
                  (sb-unix:unix-close fd)
                  (fail "failed to unlink ~A" name/errno))
                fd)))
-    (let ((dev-null #.(coerce #-win32 "/dev/null" #+win32 "nul" 'base-string)))
+    (let ((dev-null #-win32 "/dev/null" #+win32 "nul"))
       (cond ((eq object t)
              ;; No new descriptor is needed.
              (values -1 nil))
@@ -1279,13 +1279,17 @@ Users Manual for details about the PROCESS structure.
           (t
            (fail "invalid option: ~S" object))))))
 
-#+(or linux sunos hpux)
+#+(or linux sunos haiku)
 (defun software-version ()
   "Return a string describing version of the supporting software, or NIL
   if not available."
   (or sb-sys::*software-version*
       (setf sb-sys::*software-version*
             (string-trim '(#\newline)
-                         (sb-kernel:with-simple-output-to-string (stream)
-                           (run-program "/bin/uname" `("-r")
+                         (%with-output-to-string (stream)
+                           (run-program "/bin/uname"
+                                        ;; "-r" on haiku just prints "1"
+                                        ;; but "-v" prints some detail.
+                                        #+haiku '("-v")
+                                        #-haiku '("-r")
                                         :output stream))))))
