@@ -7,33 +7,22 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!KERNEL")
+(in-package "SB-KERNEL")
 
 (/show0 "target-defstruct.lisp 12")
 
 ;;;; structure frobbing primitives
 
-;;; Allocate a new instance with LENGTH data slots.
-(defun %make-instance (length)
-  (declare (type index length))
-  (%make-instance length))
-
-;;; Given an instance, return its length.
-(defun %instance-length (instance)
-  (declare (type instance instance))
-  (%instance-length instance))
-
 ;;; Return the value from the INDEXth slot of INSTANCE. This is SETFable.
+;;; This is used right away in warm compile by MAKE-LOAD-FORM-SAVING-SLOTS,
+;;; so without it already defined, you can't define it, because you can't dump
+;;; debug info structures. Were it not for that, this would go in 'stubs'.
 (defun %instance-ref (instance index)
   (%instance-ref instance index))
 
-;;; Set the INDEXth slot of INSTANCE to NEW-VALUE.
-(defun %instance-set (instance index new-value)
-  (setf (%instance-ref instance index) new-value))
-
 ;;; Normally IR2 converted, definition needed for interpreted structure
 ;;; constructors only.
-#!+(or sb-eval sb-fasteval)
+#+(or sb-eval sb-fasteval)
 (defun %make-structure-instance (dd slot-specs &rest slot-values)
   (let ((instance (%make-instance (dd-length dd))) ; length = sans header word
         (value-index 0))
@@ -43,7 +32,7 @@
       (destructuring-bind (kind raw-type . index) spec
         (if (eq kind :unbound)
             (setf (%instance-ref instance index)
-                  (sb!sys:%primitive make-unbound-marker))
+                  (sb-sys:%primitive make-unbound-marker))
             (macrolet ((make-case ()
                            `(ecase raw-type
                               ((t)
@@ -59,86 +48,61 @@
                   (incf value-index)
                   (make-case))))))))
 
-(defun %instance-layout (instance)
-  (%instance-layout instance))
-
-(defun %set-instance-layout (instance new-value)
-  (%set-instance-layout instance new-value))
-
-(defun %make-funcallable-instance (len)
-  (%make-funcallable-instance len))
-
-(defun funcallable-instance-p (x)
-  (funcallable-instance-p x))
-
-(defun %funcallable-instance-info (fin i)
-  (%funcallable-instance-info fin i))
-
-(defun %set-funcallable-instance-info (fin i new-value)
-  (%set-funcallable-instance-info fin i new-value))
-
-(defun funcallable-instance-fun (fin)
-  (%funcallable-instance-function fin))
-
-(defun (setf funcallable-instance-fun) (new-value fin)
-  (setf (%funcallable-instance-function fin) new-value))
 
 ;;;; target-only parts of the DEFSTRUCT top level code
 
 ;;; A list of hooks designating functions of one argument, the
 ;;; classoid, to be called when a defstruct is evaluated.
-(!defvar *defstruct-hooks* nil)
+(!define-load-time-global *defstruct-hooks* nil)
 
 ;;; the part of %DEFSTRUCT which makes sense only on the target SBCL
 ;;;
-(defun %target-defstruct (dd)
+(defun assign-equalp-impl (type-name function)
+  (setf (%instance-ref (find-layout type-name) (get-dsd-index layout equalp-impl))
+        function))
+
+(defun %target-defstruct (dd equalp)
   (declare (type defstruct-description dd))
 
-  #!+(and sb-show (host-feature sb-xc))
-  (progn (write `(%target-defstruct ,(dd-name dd))) (terpri))
-
   (when (dd-doc dd)
-    (setf (fdocumentation (dd-name dd) 'structure)
+    (setf (documentation (dd-name dd) 'structure)
           (dd-doc dd)))
 
   (let* ((classoid (find-classoid (dd-name dd)))
          (layout (classoid-layout classoid)))
-    (when (eq (dd-pure dd) t)
-      (setf (layout-pure layout) t))
-    ;; Make a vector of EQUALP slots comparators, indexed by (- word-index data-start).
-    ;; This has to be assigned to something regardless of whether there are
-    ;; raw slots just in case someone mutates a layout which had raw
-    ;; slots into one which does not - although that would probably crash
-    ;; unless no instances exist or all raw slots miraculously contained
-    ;; bits which were the equivalent of valid Lisp descriptors.
-    (setf (layout-equalp-tests layout)
-          (if (eql (layout-bitmap layout) +layout-all-tagged+)
-              #()
-              ;; The initial element of NIL means "do not compare".
-              ;; Ignored words (comparator = NIL) fall into two categories:
-              ;; - pseudo-ignored, which get compared by their
-              ;;   predecessor word, as for complex-double-float,
-              ;; - internal padding words which are truly ignored.
-              ;; Other words are compared as tagged if the comparator is 0,
-              ;; or as untagged if the comparator is a type-specific function.
-              (let ((comparators
+    (setf (%instance-ref layout (get-dsd-index sb-kernel:layout sb-kernel::equalp-impl))
+          (cond ((compiled-function-p equalp) equalp)
+                ((eql (layout-bitmap layout) +layout-all-tagged+)
+                 #'sb-impl::instance-equalp)
+                (t
+                  ;; Make a vector of EQUALP slots comparators, indexed by
+                  ;; (- word-index INSTANCE-DATA-START).
+                  ;; The initial element of NIL means "do not compare".
+                  ;; Ignored words (comparator = NIL) fall into two categories:
+                  ;; - pseudo-ignored, which get compared by their
+                  ;;   predecessor word, as for complex-double-float,
+                  ;; - internal padding words which are truly ignored.
+                  ;; Other words are compared as tagged if the comparator is 0,
+                  ;; or as untagged if the comparator is a type-specific function.
+                  (let ((comparators
                      ;; If data-start is 1, subtract 1 because we don't need
                      ;; a comparator for the LAYOUT slot.
-                     (make-array (- (dd-length dd) sb!vm:instance-data-start)
+                     (make-array (- (dd-length dd) sb-vm:instance-data-start)
                                  :initial-element nil)))
-                (dolist (slot (dd-slots dd) comparators)
-                  ;; -1 because LAYOUT (slot index 0) has no comparator stored.
-                  (setf (aref comparators
-                              (- (dsd-index slot) sb!vm:instance-data-start))
-                        (let ((rsd (dsd-raw-slot-data slot)))
-                          (if (not rsd)
-                              0 ; means recurse using EQUALP
-                              (raw-slot-data-comparer rsd))))))))
+                    (dolist (slot (dd-slots dd) comparators)
+                      (setf (aref comparators
+                                  (- (dsd-index slot) sb-vm:instance-data-start))
+                            (let ((rsd (dsd-raw-slot-data slot)))
+                              (if (not rsd)
+                                  0 ; means recurse using EQUALP
+                                  (raw-slot-data-comparator rsd)))))
+                    (lambda (a b)
+                      (sb-impl::instance-equalp* comparators a b))))))
 
     (dolist (fun *defstruct-hooks*)
       (funcall fun classoid)))
 
-  (values))
+  (dd-name dd))
 
 ;;; Copy any old kind of structure.
 (defun copy-structure (structure)
@@ -162,7 +126,7 @@
         ;; On backends which don't segregate descriptor vs. non-descriptor
         ;; registers, we could speed up this code in an obvious way.
         (macrolet ((copy-loop (tagged-p &optional step)
-                     `(do ((i sb!vm:instance-data-start (1+ i)))
+                     `(do ((i sb-vm:instance-data-start (1+ i)))
                           ((>= i len))
                         (declare (index i))
                         (if ,tagged-p
@@ -179,6 +143,21 @@
                 (t ; bignum - use LOGBITP to avoid consing more bignums
                  (copy-loop (logbitp i bitmap))))))
       res)))
+;;; Like above, but copy all slots (including the LAYOUT) as though boxed.
+;;; If the structure might contain raw slots and the GC is precise,
+;;; this won't ever be called.
+(defun %copy-instance (to from)
+  (declare (structure-object to from) (optimize (safety 0)))
+  (setf (%instance-layout to) (%instance-layout from))
+  (dotimes (i (%instance-length to) to)
+    (setf (%instance-ref to i) (%instance-ref from i))))
+;;; Like %COPY-INSTANCE, but layout was already assigned.
+;;; Similarly, will not be called if raw slots and precise GC.
+(defun %copy-instance-slots (to from)
+  (declare (structure-object to from) (optimize (safety 0)))
+  (loop for i from sb-vm:instance-data-start below (%instance-length to)
+        do (setf (%instance-ref to i) (%instance-ref from i)))
+  to)
 
 ;;; default PRINT-OBJECT method
 
@@ -187,14 +166,16 @@
 ;;; - it works even if the user clobbered an accessor
 ;;; - it works if the slot fails a type-check and the reader was SAFE-P,
 ;;    i.e. was required to perform a check. This is a feature, not a bug.
-(macrolet ((access-fn (dsd)
-             `(acond ((dsd-raw-slot-data ,dsd)
-                      (symbol-function (raw-slot-data-accessor-name it)))
-                     (t #'%instance-ref))))
+(macrolet ((access (dsd)
+             `(let ((i (dsd-index ,dsd)))
+                (acond ((dsd-raw-slot-data ,dsd)
+                        (funcall (raw-slot-data-accessor-fun it) structure i))
+                       (t
+                        (%instance-ref structure i))))))
 
 (defun %default-structure-pretty-print (structure stream name dd)
   (pprint-logical-block (stream nil :prefix "#S(" :suffix ")")
-    (prin1 name stream)
+    (write name :stream stream) ; escaped or not, according to printer controls
     (let ((remaining-slots (dd-slots dd)))
       (when remaining-slots
         (write-char #\space stream)
@@ -206,8 +187,7 @@
                 (output-symbol (dsd-name slot) *keyword-package* stream)
                 (write-char #\space stream)
                 (pprint-newline :miser stream)
-                (output-object (funcall (access-fn slot) structure (dsd-index slot))
-                               stream)
+                (output-object (access slot) stream)
                 (when (null remaining-slots)
                   (return))
                 (write-char #\space stream)
@@ -216,7 +196,7 @@
 (defun %default-structure-ugly-print (structure stream name dd)
   (descend-into (stream)
     (write-string "#S(" stream)
-    (prin1 name stream)
+    (write name :stream stream)
     (do ((index 0 (1+ index))
          (limit (or (and (not *print-readably*) *print-length*)
                     most-positive-fixnum))
@@ -228,8 +208,7 @@
       (let ((slot (first remaining-slots)))
         (output-symbol (dsd-name slot) *keyword-package* stream)
         (write-char #\space stream)
-        (output-object (funcall (access-fn slot) structure (dsd-index slot))
-                       stream)))))
+        (output-object (access slot) stream)))))
 ) ; end MACROLET
 
 (defun default-structure-print (structure stream depth)
@@ -238,7 +217,7 @@
       (print-unreadable-object (structure stream :identity t :type t))
       (let* ((layout (%instance-layout structure))
              (dd (layout-info layout))
-             (name (classoid-name (layout-classoid layout))))
+             (name (layout-classoid-name layout)))
         (cond ((not dd)
                ;; FIXME? this branch may be unnecessary as a consequence
                ;; of change f02bee325920166b69070e4735a8a3f295f8edfd which
@@ -265,5 +244,24 @@
 
 (defmethod print-object ((x structure-object) stream)
   (default-structure-print x stream *current-level-in-print*))
+
+;;; Used internally, but it would be nice to provide something
+;;; like this for users as well.
+(defmacro define-structure-slot-addressor (name &key structure slot (byte-offset 0))
+  (let* ((dd (find-defstruct-description structure t))
+         (slotd (or (and dd (find slot (dd-slots dd) :key #'dsd-name))
+                    (error "Slot ~S not found in ~S." slot structure)))
+         (index (dsd-index slotd)))
+    `(progn
+       (declaim (inline ,name))
+       (defun ,name (instance)
+         (declare (type ,structure instance) (optimize speed))
+         (truly-the
+          word
+          (+ (get-lisp-obj-address instance)
+             ,byte-offset
+             ,(+ (- sb-vm:instance-pointer-lowtag)
+                 (* (+ sb-vm:instance-slots-offset index)
+                    sb-vm:n-word-bytes))))))))
 
 (/show0 "target-defstruct.lisp end of file")
