@@ -89,6 +89,7 @@
              (:copier nil))
   (contents nil :type vector   :read-only t)
   (keyfun   nil :type function :read-only t))
+(declaim (freeze-type priority-queue))
 
 (defmethod print-object ((object priority-queue) stream)
   (print-unreadable-object (object stream :type t :identity t)
@@ -144,6 +145,7 @@ future versions."
   (thread             nil :type (or sb-thread:thread boolean))
   (interrupt-function nil :type (or null function))
   (cancel-function    nil :type (or null function)))
+(declaim (freeze-type timer))
 
 (defmethod print-object ((timer timer) stream)
   (let ((name (%timer-name timer)))
@@ -203,7 +205,7 @@ from now. For timers with a repeat interval it returns true."
 ;;; real time conversion
 
 (defun delta->real (delta)
-  (floor (* delta sb-xc:internal-time-units-per-second)))
+  (floor (* delta internal-time-units-per-second)))
 
 ;;; Public interface
 
@@ -368,15 +370,15 @@ triggers."
         (prog1
             (setf *waitable-timer-handle* (os-create-wtimer))
           (setf *timer-thread*
-                (sb-thread:make-thread
+                (sb-thread::make-ephemeral-thread
+                 "System timer watchdog thread"
                  (lambda ()
                    (loop while
-                        (or (zerop
-                             (os-wait-for-wtimer *waitable-timer-handle*))
-                            *waitable-timer-handle*)
-                        doing (run-expired-timers)))
-                 :ephemeral t
-                 :name "System timer watchdog thread")))))
+                         (or (zerop
+                              (os-wait-for-wtimer *waitable-timer-handle*))
+                             *waitable-timer-handle*)
+                         doing (run-expired-timers)))
+                 nil)))))
 
   (defun itimer-emulation-deinit ()
     (with-scheduler-lock ()
@@ -401,8 +403,8 @@ triggers."
   (let ((min-nsec 100000))
     (if (minusp time)
         (values 0 min-nsec)
-        (multiple-value-bind (s u) (floor time sb-xc:internal-time-units-per-second)
-          (setf u (floor (* (/ u sb-xc:internal-time-units-per-second)
+        (multiple-value-bind (s u) (floor time internal-time-units-per-second)
+          (setf u (floor (* (/ u internal-time-units-per-second)
                             #.(expt 10 9))))
           (if (and (= 0 s) (< u min-nsec))
               ;; 0 0 means "shut down the timer" for setitimer
@@ -467,8 +469,8 @@ triggers."
                    (push timer timers)))
         (run-timers)))))
 
-(defun timeout-cerror ()
-  (cerror "Continue" 'timeout))
+(defun timeout-cerror (&optional seconds)
+  (cerror "Continue" 'timeout :seconds seconds))
 
 (defmacro with-timeout (expires &body body)
   "Execute the body, asynchronously interrupting it and signalling a TIMEOUT
@@ -496,7 +498,7 @@ the system uninterruptible."
        ;; FIXME: a temporary compatibility workaround for CLX, if unsafe
        ;; unwinds are handled revisit it.
        (if (> expires 0)
-           (let ((timer (make-timer #'timeout-cerror)))
+           (let ((timer (make-timer (lambda () (timeout-cerror expires)))))
              (schedule-timer timer expires)
              (unwind-protect (timeout-body)
                (unschedule-timer timer)))
