@@ -11,7 +11,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 ;;;; default initfiles
 
@@ -19,11 +19,11 @@
   (or (let ((sbcl-homedir (sbcl-homedir-pathname)))
         (when sbcl-homedir
           (probe-file (merge-pathnames "sbclrc" sbcl-homedir))))
-      #!+win32
+      #+win32
       (merge-pathnames "sbcl\\sbclrc"
-                       (sb!win32::get-folder-pathname
-                        sb!win32::csidl_common_appdata))
-      #!-win32
+                       (sb-win32::get-folder-pathname
+                        sb-win32::csidl_common_appdata))
+      #-win32
       "/etc/sbclrc"))
 
 (defun userinit-pathname ()
@@ -58,7 +58,7 @@ been specified on the command-line.")
          (%exit)))))
 
 (define-load-time-global *exit-lock* nil)
-(!define-thread-local *exit-in-process* nil)
+(define-thread-local *exit-in-progress* nil)
 (declaim (type (or null real) *exit-timeout*))
 (defvar *exit-timeout* 60
   "Default amount of seconds, if any, EXIT should wait for other
@@ -67,21 +67,86 @@ means to wait indefinitely.")
 
 (defun os-exit-handler (condition)
   (declare (ignore condition))
-  (os-exit *exit-in-process* :abort t))
+  (os-exit *exit-in-progress* :abort t))
 
 (defvar *exit-error-handler* #'os-exit-handler)
 
 (defun call-exit-hooks ()
-  (unless *exit-in-process*
-    (setf *exit-in-process* 0))
+  (unless *exit-in-progress*
+    (setf *exit-in-progress* 0))
   (handler-bind ((serious-condition *exit-error-handler*))
     (call-hooks "exit" *exit-hooks* :on-error :warn)))
 
+;;; TERMINATE-THREAD as it exists is basically unsuitable for anything other than
+;;; interactive use, because it uses INTERRUPT-THREAD to unwind through user code.
+;;; It is only slightly ok to call INTERRUPT-THREAD on a thread when you know
+;;; exactly what it is doing. But even if you know what it's doing, you don't know
+;;; when, so even if it's your code, it might be calling an alien destructor.
+;;; It says right in the docstring of INTERRUPT-THREAD not to call it,
+;;; and then what do we do? We go full steam ahead and call it.
+;;;   https://www.dictionary.com/e/slang/i-cant-even/
+;;;
+;;; The full story here is that some users seem to think that EXIT is responsible
+;;; for stopping any threads. There was a not-insignificant amount of discussion
+;;; in https://groups.google.com/g/sbcl-devel/c/6Xj_d1dM1k0/m/G9OPAbOueDsJ
+;;; But the mechanism employed for %EXIT-OTHER-THREADS is TERMINATE-THREAD,
+;;; and that's just horrible. No matter how well the SBCL code is constructed,
+;;; this approach can cause lockup.
+;;;
+;;; The *EXIT-HOOKS* are fine, but we can't assume that users call JOIN-THREAD in
+;;; the exit hooks - they might ask threads to stop by assigning a global polled
+;;; flag or by CONDITION-BROADCAST.  They may do nothing. It's arbitrary.
+;;; So maybe the threads are stopping "peacefully" and would finish on their own
+;;; and we {could,should} just call JOIN-THREAD on them. But there's no way to know.
+;;; Instead we just abort random threads with INTERRUPT-THREAD.
+;;; Some of those threads that we forcibly terminate might have already been
+;;; halfway through cleanly unwinding, which very possibly (or even likely)
+;;; involves calling free() to release resources. When we interrupt a thread
+;;; in the midst of its voluntary call to free(), no matter how clever free() is,
+;;; we can presume that it will acquire a global lock.
+;;;
+;;; And there you have it - some dead thread owned a global lock that can't become
+;;; available because we just punched that thread in the guts and killed it.
+;;; Hence *all* threads (including non-Lisp) are permanently stuck waiting on
+;;; a lock that is not owned by any running thread.
+;;;
+;;; I tried to remedy this by having the SIGPIPE signal handler silently discard
+;;; a signal if interrupted at a non-Lisp program counter, but that presumes the
+;;; call stack is never Lisp -> C -> Lisp, which first of all means that the handler
+;;; needs to be entirely in C (if called from TERMINATE-THREAD) and secondly requires
+;;; TERMINATE-THREAD to keep trying to send a signal if it thinks the signal was dropped.
+;;; And you can't figure out whether a signal got purposely discard because then you'd
+;;; need to wait an arbitrary amount of time to decide whether it might have been.
+;;; Or invent a synchronized technique to see whether the asynchronous mechanism worked.
+;;;
+;;; So then the solution occurred to me:
+;;;  --> just don't terminate threads by force. Problem solved. <--
+;;;
+;;; Therefore, if you want EXIT to be as near to the C exit() call as possible,
+;;; after doing the lisp exit hooks, just set this global flag to NIL.
+;;; Or leave it as T if you're fine with the old lockup-prone behavior.
+;;;
+;;; And if you think it's up to the Lisp environment to try to make threads stop
+;;; by hook or by crook as a precondition to calling OS-EXIT (with which I disagree),
+;;; the right solution is possibly modeled on pthread_cancel() which has specified
+;;; cancelation points (system and C library calls) which check for cancelation,
+;;; perform whatever cleanups were pushed by pthread_cleanup_push() and then stop.
+;;;
+;;; P.S. To see that #+sb-thruption is no magic fix - suppose you have a stack
+;;; with Lisp -> C -> -> Lisp where C acquired a resource and the currently
+;;; top-of-stack Lisp function took a safepoint trap. It would be fine if all it
+;;; wanted to do was GC, but is not safe in general.
+;;; Hence #+sb-thruption paints a dangerously attractive veneer over an unsafe
+;;; concept, making it more subtly bad instead of very obviously bad.
+;;;
+;;; So this is a bad default. Bad bad bad. But it's backward-compatible.
+(define-load-time-global *forcibly-terminate-threads-on-exit* t)
+
 (defun %exit ()
   ;; If anything goes wrong, we will exit immediately and forcibly.
   (handler-bind ((serious-condition *exit-error-handler*))
     (let ((ok nil)
-          (code *exit-in-process*))
+          (code *exit-in-progress*))
       (if (consp code)
           ;; Another thread called EXIT, and passed the buck to us -- only
           ;; final call left to do.
@@ -89,7 +154,11 @@ means to wait indefinitely.")
           (unwind-protect
                (progn
                  (flush-standard-output-streams)
-                 (sb!thread::%exit-other-threads)
+                 (if *forcibly-terminate-threads-on-exit*
+                     (sb-thread::%exit-other-threads)
+                     (with-deadline (:seconds nil :override t)
+                       (sb-impl::finalizer-thread-stop)
+                       (sb-thread:grab-mutex sb-thread::*make-thread-lock*)))
                  (setf ok t))
             (os-exit code :abort (not ok)))))))
 
@@ -103,7 +172,7 @@ means to wait indefinitely.")
                                             (denominator seconds))
     (values quot
             (* rem
-               #.(if (sb!xc:typep 1000000000 'fixnum)
+               #.(if (sb-xc:typep 1000000000 'fixnum)
                      '(truncate 1000000000 (denominator seconds))
                      ;; Can't truncate a bignum by a fixnum without consing
                      '(* 10 (truncate 100000000 (denominator seconds))))))))
@@ -116,19 +185,19 @@ means to wait indefinitely.")
              (values whole-seconds
                      (truly-the (integer 0 #.(expt 10 9))
                                 (%unary-truncate (* (- seconds (float whole-seconds seconds))
-                                                    (load-time-value 1f9 t))))))))
+                                                    $1f9)))))))
     (declare (inline split-float))
     (typecase seconds
-      ((single-float 0f0 #.(float sb!xc:most-positive-fixnum 1f0))
+      ((single-float $0f0 #.(float most-positive-fixnum $1f0))
        (split-float))
-      ((double-float 0d0 #.(float sb!xc:most-positive-fixnum 1d0))
+      ((double-float $0d0 #.(float most-positive-fixnum $1d0))
        (split-float))
       (ratio
        (split-ratio-for-sleep seconds))
       (t
        (multiple-value-bind (sec frac)
            (truncate seconds)
-         (values sec (truncate frac (load-time-value 1f-9 t))))))))
+         (values sec (truncate frac $1f-9)))))))
 
 (declaim (inline %nanosleep))
 (defun %nanosleep (sec nsec)
@@ -137,31 +206,31 @@ means to wait indefinitely.")
   ;; can actually have a reason to sleep for over 3 years?
   (loop while (> sec (expt 10 8))
      do (decf sec (expt 10 8))
-       (sb!unix:nanosleep (expt 10 8) 0))
-  (sb!unix:nanosleep sec nsec))
+       (sb-unix:nanosleep (expt 10 8) 0))
+  (sb-unix:nanosleep sec nsec))
 
 (declaim (inline %sleep))
-#!-win32
+#-win32
 (defun %sleep (seconds)
   (typecase seconds
     (double-float
-     (sb!unix::nanosleep-double seconds))
+     (sb-unix::nanosleep-double seconds))
     (single-float
-     (sb!unix::nanosleep-float seconds))
+     (sb-unix::nanosleep-float seconds))
     (integer
      (%nanosleep seconds 0))
     (t
      (multiple-value-call #'%nanosleep (split-ratio-for-sleep seconds)))))
 
-#!+(and win32 sb-thread)
+#+(and win32 sb-thread)
 (defun %sleep (seconds)
   (if (integerp seconds)
       (%nanosleep seconds 0)
       (multiple-value-call #'%nanosleep (split-seconds-for-sleep seconds))))
 
-#!+(and win32 (not sb-thread))
+#+(and win32 (not sb-thread))
 (defun %sleep (seconds)
-  (sb!win32:millisleep (truncate (* seconds 1000))))
+  (sb-win32:millisleep (truncate (* seconds 1000))))
 
 (defun sleep (seconds)
   "This function causes execution to be suspended for SECONDS. SECONDS may be
@@ -302,11 +371,12 @@ any non-negative real number."
            ;; Scripts don't need to be stylish or fast, but silence is usually a
            ;; desirable quality...
            (handler-bind (((or style-warning compiler-note) #'muffle-warning)
-                          (stream-error (lambda (e)
-                                          ;; Shell-style.
-                                          (when (member (stream-error-stream e)
-                                                        (list *stdout* *stdin* *stderr*))
-                                            (exit)))))
+                          ((or broken-pipe end-of-file)
+                            (lambda (e)
+                              ;; Shell-style.
+                              (when (member (stream-error-stream e)
+                                            (list *stdout* *stdin* *stderr*))
+                                (exit)))))
              ;; Let's not use the *TTY* for scripts, ok? Also, normally we use
              ;; synonym streams, but in order to have the broken pipe/eof error
              ;; handling right we want to bind them for scripts.
@@ -320,7 +390,7 @@ any non-negative real number."
       (if (eq t script)
           (load-script *stdin*)
           (with-open-file (f (native-pathname script) :element-type :default)
-            (sb!fasl::maybe-skip-shebang-line f)
+            (sb-fasl::maybe-skip-shebang-line f)
             (load-script f))))))
 
 ;; Errors while processing the command line cause the system to EXIT,
@@ -484,7 +554,7 @@ any non-negative real number."
                        (cond (stream
                               (dx-flet ((thunk ()
                                           (load-as-source stream :context kind)))
-                                (sb!fasl::call-with-load-bindings #'thunk stream)))
+                                (sb-fasl::call-with-load-bindings #'thunk stream)))
                              (specified-pathname
                               (cerror "Ignore missing init file"
                                       "The specified ~A file ~A was not found."
@@ -582,22 +652,16 @@ that provides the REPL for the system. Assumes that *STANDARD-INPUT* and
       ;; Each REPL in a multithreaded world should have bindings of
       ;; most CL specials (most critically *PACKAGE*).
       (with-rebound-io-syntax
-          (handler-bind ((step-condition 'invoke-stepper))
-            (loop
-               (/show0 "about to set up restarts in TOPLEVEL-REPL")
-               ;; CLHS recommends that there should always be an
-               ;; ABORT restart; we have this one here, and one per
-               ;; debugger level.
-               (with-simple-restart
-                   (abort "~@<Exit debugger, returning to top level.~@:>")
-                 (catch 'toplevel-catcher
-                   ;; In the event of a control-stack-exhausted-error, we
-                   ;; should have unwound enough stack by the time we get
-                   ;; here that this is now possible.
-                   #!-win32
-                   (sb!kernel::reset-control-stack-guard-page)
-                   (funcall repl-fun noprint)
-                   (critically-unreachable "after REPL")))))))))
+        (loop
+         (/show0 "about to set up restarts in TOPLEVEL-REPL")
+         ;; CLHS recommends that there should always be an
+         ;; ABORT restart; we have this one here, and one per
+         ;; debugger level.
+         (with-simple-restart
+             (abort "~@<Exit debugger, returning to top level.~@:>")
+           (catch 'toplevel-catcher
+             (funcall repl-fun noprint)
+             (critically-unreachable "after REPL"))))))))
 
 ;;; Our default REPL prompt is the minimal traditional one.
 (defun repl-prompt-fun (stream)
@@ -625,7 +689,7 @@ that provides the REPL for the system. Assumes that *STANDARD-INPUT* and
    (unwind-protect
         (progn
           (scrub-control-stack)
-          (sb!thread::get-foreground)
+          (sb-thread::get-foreground)
           (unless noprint
             (flush-standard-output-streams)
             (funcall *repl-prompt-fun* *standard-output*)
@@ -635,7 +699,14 @@ that provides the REPL for the system. Assumes that *STANDARD-INPUT* and
             ;; odd. But maybe there *is* a valid reason in some
             ;; circumstances? perhaps some deadlock issue when being driven
             ;; by another process or something...)
-            (force-output *standard-output*))
+            (force-output *standard-output*)
+            (let ((real (maybe-resolve-synonym-stream *standard-output*)))
+              ;; Because by default *standard-output* is not
+              ;; *terminal-io* but STDOUT the column is not reset
+              ;; after pressing enter. Reduce confusion by resetting
+              ;; the column to 0
+              (when (fd-stream-p real)
+                (setf (fd-stream-output-column real) 0))))
           (let* ((form (funcall *repl-read-form-fun*
                                 *standard-input*
                                 *standard-output*))
@@ -649,4 +720,68 @@ that provides the REPL for the system. Assumes that *STANDARD-INPUT* and
 
 ;;; a convenient way to get into the assembly-level debugger
 (defun %halt ()
-  (%primitive sb!c:halt))
+  (%primitive sb-c:halt))
+
+
+;;;; Examples of why NEVER NEVER NEVER to call TERMINATE-THREAD in production.
+
+#|
+Example 1. Stuck somewhere in free() and in JOIN-THREAD.
+[stuck at https://github.com/google/tcmalloc/blob/master/tcmalloc/cpu_cache.cc#L204]
+  Thread "A":
+    0x98e697c [AbslInternalSpinLockDelay]
+    0x9898287 [tcmalloc::CPUCache::UpdateCapacity(int, unsigned long, unsigned long, bool, tcmalloc::CPUCache::ObjectClass*, unsigned long*)]
+    0x9898b38 [tcmalloc::CPUCache::Overflow(void*, unsigned long, int)]
+    0x9891f39 [(anonymous namespace)::do_free_no_hooks(void*)]
+    0x988cea1 [MallocBlock::ProcessFreeQueue(MallocBlock*, unsigned long, int)]
+    0x26f490a [CFFI::FOREIGN-ARRAY-FREE]
+  Thread "main thread":
+    0x9877f23 [futex_wait]
+    0x32f32c8 [(FLET SB-UNIX::BODY :IN SB-THREAD::FUTEX-WAIT)]
+    0x32f301f [SB-THREAD::%%WAIT-FOR-MUTEX]
+    0x2f97f5d [SB-THREAD::%WAIT-FOR-MUTEX]
+    0x2f287be [SB-THREAD::JOIN-THREAD]
+
+Example 2. Stuck somewhere else in free() and in JOIN-THREAD.
+  Thread "A":
+    0x7f657d1a7cb9 [syscall]
+    0x98e697c [AbslInternalSpinLockDelay]
+    0x995e9e7 [absl::base_internal::SpinLock::SlowLock()]
+    0x988d63b [MallocBlock::CheckAndClear(int)]
+    0x9a8511f [__libc_free]
+    0x26f490a [CFFI::FOREIGN-ARRAY-FREE]
+  Thread "main thread":
+    0x9877f23 [futex_wait]
+    0x32f32c8 [(FLET SB-UNIX::BODY :IN SB-THREAD::FUTEX-WAIT)]
+    0x32f301f [SB-THREAD::%%WAIT-FOR-MUTEX]
+    0x2f97f5d [SB-THREAD::%WAIT-FOR-MUTEX]
+    0x2f287be [SB-THREAD::JOIN-THREAD]
+
+Example 3. main thread stuck in a C library exit handler:
+   0x98e697c [AbslInternalSpinLockDelay]
+   0x995e9e7 [absl::base_internal::SpinLock::SlowLock()]
+   0x98a43cf [tcmalloc::HugePageAwareAllocator::LockAndAlloc(unsigned long, bool*)]
+   0x98a423f [tcmalloc::HugePageAwareAllocator::New(unsigned long)]
+   0x9896ee8 [tcmalloc::CentralFreeList::Populate()]
+   0x9896c69 [tcmalloc::CentralFreeList::RemoveRange(void**, int)]
+   0x9897e74 [tcmalloc::CPUCache::Refill(int, unsigned long)]
+   0x989221b [tcmalloc::dbg_do_malloc(unsigned long, unsigned long*)]
+   0x9a85045 [__libc_malloc]
+   0x7f25a8ad5772 [__run_exit_handlers]
+   0x7f25a8ad57c5 [exit]
+   0x2fa99a3 [SB-SYS::OS-EXIT]
+
+Example 4. main thread stuck in a different C library exit handler:
+   0x992ecfe [absl::synchronization_internal::Waiter::Wait(absl::synchronization_internal::KernelTimeout)]
+   0x98e6865 [AbslInternalPerThreadSemWait]
+   0x992fd4d [absl::Mutex::Block(absl::base_internal::PerThreadSynch*)]
+   0x99324a0 [absl::Mutex::LockSlowLoop(absl::SynchWaitParams*, int)]
+   0x9930b0e [absl::Mutex::LockSlow(absl::MuHowS const*, absl::Condition const*, int)]
+   0x9930346 [absl::Mutex::Lock()]
+   0x97ca6c0 [thread::local::internal::Var::~Var()]
+   0x7fcf3fde0772 [__run_exit_handlers]
+   0x7fcf3fde07c5 [exit]
+   0x2fa99a3 [SB-SYS::OS-EXIT]
+
+There are plenty more where those came from.
+|#
