@@ -12,7 +12,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!WIN32")
+(in-package "SB-WIN32")
 
 ;;; Alien definitions for commonly used Win32 types.  Woe unto whoever
 ;;; tries to untangle this someday for 64-bit Windows.
@@ -25,11 +25,11 @@
 (define-alien-type lispbool (boolean 32))
 
 (define-alien-type system-string
-                   #!-sb-unicode c-string
-                   #!+sb-unicode (c-string :external-format :ucs-2))
+                   #-sb-unicode c-string
+                   #+sb-unicode (c-string :external-format :ucs-2))
 
-(define-alien-type tchar #!-sb-unicode char
-                         #!+sb-unicode (unsigned 16))
+(define-alien-type tchar #-sb-unicode char
+                         #+sb-unicode (unsigned 16))
 
 (defconstant default-environment-length 1024)
 
@@ -74,39 +74,12 @@
 ;;; last-error code is maintained on a per-thread basis.
 (define-alien-routine ("GetLastError" get-last-error) dword)
 
-;;;; File Handles
-
-;;; Historically, SBCL on Windows used CRT (lowio) file descriptors,
-;;; unlike other Lisps. They really help to minimize required effort
-;;; for porting Unix-specific software, at least to the level that it
-;;; mostly works most of the time.
-;;;
-;;; Alastair Bridgewater recommended to switch away from CRT
-;;; descriptors, and Anton Kovalenko thinks it's the time to heed his
-;;; advice. I see that SBCL for Windows needs much more effort in the
-;;; area of OS IO abstractions and the like; using or leaving lowio
-;;; FDs doesn't change the big picture so much.
-;;;
-;;; Lowio layer, in exchange for `semi-automatic almost-portability',
-;;; brings some significant problems, which a grown-up cross-platform
-;;; CL implementation shouldn't have. Therefore, as its benefits
-;;; become negligible, it's a good reason to throw it away.
-;;;
-;;;  -- comment from AK's branch
-
-;;; For a few more releases, let's preserve old functions (now
-;;; implemented as identity) for user code which might have had to peek
-;;; into our internals in past versions when we hadn't been using
-;;; handles yet. -- DFL, 2012
-(defun get-osfhandle (fd) fd)
-(defun open-osfhandle (handle flags) (declare (ignore flags)) handle)
-
 ;;; Get the operating system handle for a C file descriptor.  Returns
 ;;; INVALID-HANDLE on failure.
-(define-alien-routine ("_get_osfhandle" real-get-osfhandle) handle
+(define-alien-routine ("_get_osfhandle" get-osfhandle) handle
   (fd int))
 
-(define-alien-routine ("_close" real-crt-close) int
+(define-alien-routine ("_close" crt-close) int
   (fd int))
 
 ;;; Read data from a file handle into a buffer.  This may be used
@@ -154,9 +127,12 @@
   (length dword)
   (nevents (* dword)))
 
-(define-alien-routine ("socket_input_available" socket-input-available) int
+(define-alien-routine "socket_input_available" int
   (socket handle))
 
+(define-alien-routine "console_handle_p" boolean
+  (handle handle))
+
 ;;; Listen for input on a Windows file handle.  Unlike UNIX, there
 ;;; isn't a unified interface to do this---we have to know what sort
 ;;; of handle we have.  Of course, there's no way to actually
@@ -164,30 +140,21 @@
 ;;; something that works.  Returns true if there could be input
 ;;; available, or false if there is not.
 (defun handle-listen (handle)
-  (with-alien ((avail dword)
-               (buf (array char #.input-record-size)))
-    (when
-        ;; Make use of the fact that console handles are technically no
-        ;; real handles, and unlike those, have these bits set:
-        (= 3 (logand 3 handle))
-      (return-from handle-listen
-        (alien-funcall (extern-alien "win32_tty_listen"
-                                     (function boolean handle))
-                       handle)))
-    (unless (zerop (peek-named-pipe handle nil 0 nil (addr avail) nil))
-      (return-from handle-listen (plusp avail)))
-    (let ((res (socket-input-available handle)))
-      (unless (zerop res)
-        (return-from handle-listen (= res 1))))
-    t))
-
-;;; Listen for input on a C runtime file handle.  Returns true if
-;;; there could be input available, or false if there is not.
-(defun fd-listen (fd)
-  (let ((handle (get-osfhandle fd)))
-    (if handle
-        (handle-listen handle)
-        t)))
+  (cond ((console-handle-p handle)
+         (alien-funcall (extern-alien "win32_tty_listen"
+                                      (function boolean handle))
+                        handle))
+        (t
+         (with-alien ((avail dword))
+           (let* ((res (peek-named-pipe handle nil 0 nil (addr avail) nil))
+                  (code (get-last-error)))
+             (cond
+               ((not (zerop res)) (return-from handle-listen (plusp avail)))
+               ((= code error-broken-pipe) (return-from handle-listen t)))))
+         (let ((res (socket-input-available handle)))
+           (unless (zerop res)
+             (return-from handle-listen (= res 1))))
+         t)))
 
 ;;; Clear all available input from a file handle.
 (defun handle-clear-input (handle)
@@ -202,230 +169,27 @@
      (when (< count 1024)
        (return)))))
 
-;;; Clear all available input from a C runtime file handle.
-(defun fd-clear-input (fd)
-  (let ((handle (get-osfhandle fd)))
-    (when handle
-      (handle-clear-input handle))))
-
 ;;;; System Functions
 
-#!-sb-thread
+#-sb-thread
 (define-alien-routine ("Sleep" millisleep) void
   (milliseconds dword))
 
-#!+sb-thread
-(defun sb!unix:nanosleep (sec nsec)
+#+sb-thread
+(defun sb-unix:nanosleep (sec nsec)
   (let ((*allow-with-interrupts* *interrupts-enabled*))
     (without-interrupts
-      (let ((timer (sb!impl::os-create-wtimer)))
-        (sb!impl::os-set-wtimer timer sec nsec)
+      (let ((timer (sb-impl::os-create-wtimer)))
+        (sb-impl::os-set-wtimer timer sec nsec)
         (unwind-protect
              (do () ((with-local-interrupts
-                       (zerop (sb!impl::os-wait-for-wtimer timer)))))
-          (sb!impl::os-close-wtimer timer))))))
+                       (zerop (sb-impl::os-wait-for-wtimer timer)))))
+          (sb-impl::os-close-wtimer timer))))))
 
 (define-alien-routine ("win32_wait_object_or_signal" wait-object-or-signal)
-    (signed 16)
+    dword
   (handle handle))
 
-#!+sb-unicode
-(progn
-  (defvar *ansi-codepage* nil)
-  (defvar *oem-codepage* nil)
-  (defvar *codepage-to-external-format* (make-hash-table)))
-
-#!+sb-unicode
-(dolist
-    (cp '(;;037       IBM EBCDIC - U.S./Canada
-          (437 :CP437) ;; OEM - United States
-          ;;500       IBM EBCDIC - International
-          ;;708       Arabic - ASMO 708
-          ;;709       Arabic - ASMO 449+, BCON V4
-          ;;710       Arabic - Transparent Arabic
-          ;;720       Arabic - Transparent ASMO
-          ;;737       OEM - Greek (formerly 437G)
-          ;;775       OEM - Baltic
-          (850 :CP850)     ;; OEM - Multilingual Latin I
-          (852 :CP852)     ;; OEM - Latin II
-          (855 :CP855)     ;; OEM - Cyrillic (primarily Russian)
-          (857 :CP857)     ;; OEM - Turkish
-          ;;858       OEM - Multilingual Latin I + Euro symbol
-          (860 :CP860)     ;; OEM - Portuguese
-          (861 :CP861)     ;; OEM - Icelandic
-          (862 :CP862)     ;; OEM - Hebrew
-          (863 :CP863)     ;; OEM - Canadian-French
-          (864 :CP864)     ;; OEM - Arabic
-          (865 :CP865)     ;; OEM - Nordic
-          (866 :CP866)     ;; OEM - Russian
-          (869 :CP869)     ;; OEM - Modern Greek
-          ;;870       IBM EBCDIC - Multilingual/ROECE (Latin-2)
-          (874 :CP874) ;; ANSI/OEM - Thai (same as 28605, ISO 8859-15)
-          ;;875       IBM EBCDIC - Modern Greek
-          (932 :CP932)     ;; ANSI/OEM - Japanese, Shift-JIS
-          ;;936       ANSI/OEM - Simplified Chinese (PRC, Singapore)
-          ;;949       ANSI/OEM - Korean (Unified Hangul Code)
-          ;;950       ANSI/OEM - Traditional Chinese (Taiwan; Hong Kong SAR, PRC)
-          ;;1026      IBM EBCDIC - Turkish (Latin-5)
-          ;;1047      IBM EBCDIC - Latin 1/Open System
-          ;;1140      IBM EBCDIC - U.S./Canada (037 + Euro symbol)
-          ;;1141      IBM EBCDIC - Germany (20273 + Euro symbol)
-          ;;1142      IBM EBCDIC - Denmark/Norway (20277 + Euro symbol)
-          ;;1143      IBM EBCDIC - Finland/Sweden (20278 + Euro symbol)
-          ;;1144      IBM EBCDIC - Italy (20280 + Euro symbol)
-          ;;1145      IBM EBCDIC - Latin America/Spain (20284 + Euro symbol)
-          ;;1146      IBM EBCDIC - United Kingdom (20285 + Euro symbol)
-          ;;1147      IBM EBCDIC - France (20297 + Euro symbol)
-          ;;1148      IBM EBCDIC - International (500 + Euro symbol)
-          ;;1149      IBM EBCDIC - Icelandic (20871 + Euro symbol)
-          (1200 :UCS-2LE)    ;; Unicode UCS-2 Little-Endian (BMP of ISO 10646)
-          (1201 :UCS-2BE)    ;; Unicode UCS-2 Big-Endian
-          (1250 :CP1250)     ;; ANSI - Central European
-          (1251 :CP1251)     ;; ANSI - Cyrillic
-          (1252 :CP1252)     ;; ANSI - Latin I
-          (1253 :CP1253)     ;; ANSI - Greek
-          (1254 :CP1254)     ;; ANSI - Turkish
-          (1255 :CP1255)     ;; ANSI - Hebrew
-          (1256 :CP1256)     ;; ANSI - Arabic
-          (1257 :CP1257)     ;; ANSI - Baltic
-          (1258 :CP1258)     ;; ANSI/OEM - Vietnamese
-          ;;1361      Korean (Johab)
-          ;;10000 MAC - Roman
-          ;;10001     MAC - Japanese
-          ;;10002     MAC - Traditional Chinese (Big5)
-          ;;10003     MAC - Korean
-          ;;10004     MAC - Arabic
-          ;;10005     MAC - Hebrew
-          ;;10006     MAC - Greek I
-          (10007 :X-MAC-CYRILLIC) ;; MAC - Cyrillic
-          ;;10008     MAC - Simplified Chinese (GB 2312)
-          ;;10010     MAC - Romania
-          ;;10017     MAC - Ukraine
-          ;;10021     MAC - Thai
-          ;;10029     MAC - Latin II
-          ;;10079     MAC - Icelandic
-          ;;10081     MAC - Turkish
-          ;;10082     MAC - Croatia
-          ;;12000     Unicode UCS-4 Little-Endian
-          ;;12001     Unicode UCS-4 Big-Endian
-          ;;20000     CNS - Taiwan
-          ;;20001     TCA - Taiwan
-          ;;20002     Eten - Taiwan
-          ;;20003     IBM5550 - Taiwan
-          ;;20004     TeleText - Taiwan
-          ;;20005     Wang - Taiwan
-          ;;20105     IA5 IRV International Alphabet No. 5 (7-bit)
-          ;;20106     IA5 German (7-bit)
-          ;;20107     IA5 Swedish (7-bit)
-          ;;20108     IA5 Norwegian (7-bit)
-          ;;20127     US-ASCII (7-bit)
-          ;;20261     T.61
-          ;;20269     ISO 6937 Non-Spacing Accent
-          ;;20273     IBM EBCDIC - Germany
-          ;;20277     IBM EBCDIC - Denmark/Norway
-          ;;20278     IBM EBCDIC - Finland/Sweden
-          ;;20280     IBM EBCDIC - Italy
-          ;;20284     IBM EBCDIC - Latin America/Spain
-          ;;20285     IBM EBCDIC - United Kingdom
-          ;;20290     IBM EBCDIC - Japanese Katakana Extended
-          ;;20297     IBM EBCDIC - France
-          ;;20420     IBM EBCDIC - Arabic
-          ;;20423     IBM EBCDIC - Greek
-          ;;20424     IBM EBCDIC - Hebrew
-          ;;20833     IBM EBCDIC - Korean Extended
-          ;;20838     IBM EBCDIC - Thai
-          (20866 :KOI8-R) ;; Russian - KOI8-R
-          ;;20871     IBM EBCDIC - Icelandic
-          ;;20880     IBM EBCDIC - Cyrillic (Russian)
-          ;;20905     IBM EBCDIC - Turkish
-          ;;20924     IBM EBCDIC - Latin-1/Open System (1047 + Euro symbol)
-          ;;20932     JIS X 0208-1990 & 0121-1990
-          ;;20936     Simplified Chinese (GB2312)
-          ;;21025     IBM EBCDIC - Cyrillic (Serbian, Bulgarian)
-          ;;21027     (deprecated)
-          (21866 :KOI8-U)      ;; Ukrainian (KOI8-U)
-          (28591 :LATIN-1)     ;; ISO 8859-1 Latin I
-          (28592 :ISO-8859-2)  ;; ISO 8859-2 Central Europe
-          (28593 :ISO-8859-3)  ;; ISO 8859-3 Latin 3
-          (28594 :ISO-8859-4)  ;; ISO 8859-4 Baltic
-          (28595 :ISO-8859-5)  ;; ISO 8859-5 Cyrillic
-          (28596 :ISO-8859-6)  ;; ISO 8859-6 Arabic
-          (28597 :ISO-8859-7)  ;; ISO 8859-7 Greek
-          (28598 :ISO-8859-8)  ;; ISO 8859-8 Hebrew
-          (28599 :ISO-8859-9)  ;; ISO 8859-9 Latin 5
-          (28605 :LATIN-9)     ;; ISO 8859-15 Latin 9
-          ;;29001     Europa 3
-          (38598 :ISO-8859-8) ;; ISO 8859-8 Hebrew
-          ;;50220     ISO 2022 Japanese with no halfwidth Katakana
-          ;;50221     ISO 2022 Japanese with halfwidth Katakana
-          ;;50222     ISO 2022 Japanese JIS X 0201-1989
-          ;;50225     ISO 2022 Korean
-          ;;50227     ISO 2022 Simplified Chinese
-          ;;50229     ISO 2022 Traditional Chinese
-          ;;50930     Japanese (Katakana) Extended
-          ;;50931     US/Canada and Japanese
-          ;;50933     Korean Extended and Korean
-          ;;50935     Simplified Chinese Extended and Simplified Chinese
-          ;;50936     Simplified Chinese
-          ;;50937     US/Canada and Traditional Chinese
-          ;;50939     Japanese (Latin) Extended and Japanese
-          (51932 :EUC-JP) ;; EUC - Japanese
-          ;;51936     EUC - Simplified Chinese
-          ;;51949     EUC - Korean
-          ;;51950     EUC - Traditional Chinese
-          ;;52936     HZ-GB2312 Simplified Chinese
-          ;;54936     Windows XP: GB18030 Simplified Chinese (4 Byte)
-          ;;57002     ISCII Devanagari
-          ;;57003     ISCII Bengali
-          ;;57004     ISCII Tamil
-          ;;57005     ISCII Telugu
-          ;;57006     ISCII Assamese
-          ;;57007     ISCII Oriya
-          ;;57008     ISCII Kannada
-          ;;57009     ISCII Malayalam
-          ;;57010     ISCII Gujarati
-          ;;57011     ISCII Punjabi
-          ;;65000     Unicode UTF-7
-          (65001 :UTF8))) ;; Unicode UTF-8
-  (setf (gethash (car cp) *codepage-to-external-format*) (cadr cp)))
-
-#!+sb-unicode
-;; FIXME: Something odd here: why are these two #+SB-UNICODE, whereas
-;; the console just behave differently?
-(progn
-  (declaim (ftype (function () keyword) ansi-codepage))
-  (defun ansi-codepage ()
-    (or *ansi-codepage*
-        (setq *ansi-codepage*
-              (gethash (alien-funcall (extern-alien "GetACP" (function UINT)))
-                       *codepage-to-external-format*
-                       :latin-1))))
-
-  (declaim (ftype (function () keyword) oem-codepage))
-  (defun oem-codepage ()
-    (or *oem-codepage*
-        (setq *oem-codepage*
-            (gethash (alien-funcall (extern-alien "GetOEMCP" (function UINT)))
-                     *codepage-to-external-format*
-                     :latin-1)))))
-
-;; http://msdn.microsoft.com/library/en-us/dllproc/base/getconsolecp.asp
-(declaim (ftype (function () keyword) console-input-codepage))
-(defun console-input-codepage ()
-  (or #!+sb-unicode
-      (gethash (alien-funcall (extern-alien "GetConsoleCP" (function UINT)))
-               *codepage-to-external-format*)
-      :latin-1))
-
-;; http://msdn.microsoft.com/library/en-us/dllproc/base/getconsoleoutputcp.asp
-(declaim (ftype (function () keyword) console-output-codepage))
-(defun console-output-codepage ()
-  (or #!+sb-unicode
-      (gethash (alien-funcall
-                (extern-alien "GetConsoleOutputCP" (function UINT)))
-               *codepage-to-external-format*)
-      :latin-1))
-
 (define-alien-routine ("LocalFree" local-free) void
   (lptr (* t)))
 
@@ -434,18 +198,8 @@
   `(prog1 (cast ,value ,type)
      (,free-function ,value)))
 
-(eval-when (:compile-toplevel :load-toplevel :execute)
-(defmacro with-funcname ((name description) &body body)
-  `(let
-     ((,name (etypecase ,description
-               (string ,description)
-               (cons (destructuring-bind (s &optional c) ,description
-                       (format nil "~A~A" s
-                               (if c #!-sb-unicode "A" #!+sb-unicode "W" "")))))))
-     ,@body)))
-
 (defmacro make-system-buffer (x)
- `(make-alien char #!+sb-unicode (ash ,x 1) #!-sb-unicode ,x))
+ `(make-alien char #+sb-unicode (ash ,x 1) #-sb-unicode ,x))
 
 (defmacro with-handle ((var initform
                             &key (close-operator 'close-handle))
@@ -459,49 +213,51 @@
              (,close-operator ,var))))))
 
 (define-alien-type pathname-buffer
-    (array char #.(ash (1+ max_path) #!+sb-unicode 1 #!-sb-unicode 0)))
+    (array char #.(ash (1+ max_path) #+sb-unicode 1 #-sb-unicode 0)))
 
 (define-alien-type long-pathname-buffer
-    #!+sb-unicode (array char 65536)
-    #!-sb-unicode pathname-buffer)
+    #+sb-unicode (array char 65536)
+    #-sb-unicode pathname-buffer)
 
 (defmacro decode-system-string (alien)
   `(cast (cast ,alien (* char)) system-string))
 
-;;; FIXME: The various FOO-SYSCALL-BAR macros, and perhaps some other
-;;; macros in this file, are only used in this file, and could be
-;;; implemented using SB!XC:DEFMACRO wrapped in EVAL-WHEN.
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (defun win-funcname (description)
+    (etypecase description
+      (string description)
+      (cons (destructuring-bind (s &optional c) description
+              (format nil "~A~A" s
+                      (if c #-sb-unicode "A" #+sb-unicode "W" "")))))))
 
 (defmacro syscall ((name ret-type &rest arg-types) success-form &rest args)
-  (with-funcname (sname name)
-    `(locally
-       (declare (optimize (sb!c::float-accuracy 0)))
-       (let ((result (alien-funcall
-                       (extern-alien ,sname
-                                     (function ,ret-type ,@arg-types))
-                       ,@args)))
-         (declare (ignorable result))
-         ,success-form))))
+  `(locally
+       (declare (optimize (sb-c::float-accuracy 0)))
+     (let ((result (alien-funcall
+                    (extern-alien ,(win-funcname name)
+                                  (function ,ret-type ,@arg-types))
+                    ,@args)))
+       (declare (ignorable result))
+       ,success-form)))
 
 ;;; This is like SYSCALL, but if it fails, signal an error instead of
 ;;; returning error codes. Should only be used for syscalls that will
 ;;; never really get an error.
 (defmacro syscall* ((name &rest arg-types) success-form &rest args)
-  (with-funcname (sname name)
+  (let ((sname (win-funcname name)))
     `(locally
-       (declare (optimize (sb!c::float-accuracy 0)))
+         (declare (optimize (sb-c::float-accuracy 0)))
        (let ((result (alien-funcall
-                       (extern-alien ,sname (function bool ,@arg-types))
-                       ,@args)))
+                      (extern-alien ,sname (function bool ,@arg-types))
+                      ,@args)))
          (when (zerop result)
            (win32-error ,sname))
          ,success-form))))
 
 (defmacro with-sysfun ((func name ret-type &rest arg-types) &body body)
-  (with-funcname (sname name)
-    `(with-alien ((,func (function ,ret-type ,@arg-types)
-                         :extern ,sname))
-       ,@body)))
+  `(with-alien ((,func (function ,ret-type ,@arg-types)
+                       :extern ,(win-funcname name)))
+     ,@body))
 
 (defmacro void-syscall* ((name &rest arg-types) &rest args)
   `(syscall* (,name ,@arg-types) (values t 0) ,@args))
@@ -538,7 +294,7 @@
 (defun get-folder-pathname (csidl)
   (parse-native-namestring (get-folder-namestring csidl)))
 
-(defun sb!unix:posix-getcwd ()
+(defun sb-unix:posix-getcwd ()
   (with-alien ((apath pathname-buffer))
     (with-sysfun (afunc ("GetCurrentDirectory" t) dword dword (* char))
       (let ((ret (alien-funcall afunc (1+ max_path) (cast apath (* char)))))
@@ -550,21 +306,21 @@
               (cast-and-free apath))
             (decode-system-string apath))))))
 
-(defun sb!unix:unix-mkdir (name mode)
-  (declare (type sb!unix:unix-pathname name)
-           (type sb!unix:unix-file-mode mode)
+(defun sb-unix:unix-mkdir (name mode)
+  (declare (type sb-unix:unix-pathname name)
+           (type sb-unix:unix-file-mode mode)
            (ignore mode))
   (syscall (("CreateDirectory" t) lispbool system-string (* t))
            (values result (if result 0 (get-last-error)))
            name nil))
 
-(defun sb!unix:unix-rename (name1 name2)
-  (declare (type sb!unix:unix-pathname name1 name2))
+(defun sb-unix:unix-rename (name1 name2)
+  (declare (type sb-unix:unix-pathname name1 name2))
   (syscall (("MoveFile" t) lispbool system-string system-string)
            (values result (if result 0 (get-last-error)))
            name1 name2))
 
-(defun sb!unix::posix-getenv (name)
+(defun sb-unix::posix-getenv (name)
   (declare (type simple-string name))
   (with-alien ((aenv (* char) (make-system-buffer default-environment-length)))
     (with-sysfun (afunc ("GetEnvironmentVariable" t)
@@ -589,7 +345,7 @@
 ;;;; Process time information
 
 (defconstant 100ns-per-internal-time-unit
-  (/ 10000000 sb!xc:internal-time-units-per-second))
+  (/ 10000000 internal-time-units-per-second))
 
 ;; FILETIME
 ;; The FILETIME structure is a 64-bit value representing the number of
@@ -622,8 +378,6 @@
                (addr ,kernel-time)
                (addr ,user-time))))
 
-(declaim (inline system-internal-real-time))
-
 (let ((epoch 0))
   (declare (unsigned-byte epoch))
   ;; FIXME: For optimization ideas see the unix implementation.
@@ -637,6 +391,7 @@
                   (addr system-time)))
        epoch)))
 
+(declaim (inline system-internal-run-time))
 (defun system-internal-run-time ()
   (with-process-times (creation-time exit-time kernel-time user-time)
     (values (floor (+ user-time kernel-time) 100ns-per-internal-time-unit))))
@@ -671,11 +426,9 @@
 ;;            (addr epoch)
 ;;            (addr filetime)))
 (defconstant +unix-epoch-filetime+ 116444736000000000)
-(defconstant +filetime-unit+ (* 100ns-per-internal-time-unit
-                                internal-time-units-per-second))
+(defconstant +filetime-unit+ 10000000)
 (defconstant +common-lisp-epoch-filetime-seconds+ 9435484800)
 
-#!-sb-fluid
 (declaim (inline get-time-of-day))
 (defun get-time-of-day ()
   "Return the number of seconds and microseconds since the beginning of the
@@ -684,8 +437,7 @@ UNIX epoch: January 1st 1970."
     (syscall (("GetSystemTimeAsFileTime") void (* filetime))
              (multiple-value-bind (sec 100ns)
                  (floor (- system-time +unix-epoch-filetime+)
-                        (* 100ns-per-internal-time-unit
-                           internal-time-units-per-second))
+                        +filetime-unit+)
                (values sec (floor 100ns 10)))
              (addr system-time))))
 
@@ -813,7 +565,7 @@ absense."
 ;;
 ;; DFL: Merged this function because it seems useful to me.  But
 ;; shouldn't we then define it on actual POSIX, too?
-(defun (setf sb!unix::posix-getenv) (new-value name)
+(defun (setf sb-unix::posix-getenv) (new-value name)
   (if (setenv name new-value)
       new-value
       (posix-getenv name)))
@@ -836,7 +588,7 @@ absense."
     (dwMinorVersion dword)
     (dwBuildNumber dword)
     (dwPlatformId dword)
-    (szCSDVersion (array char #!-sb-unicode 128 #!+sb-unicode 256))))
+    (szCSDVersion (array char #-sb-unicode 128 #+sb-unicode 256))))
 
 (defun get-version-ex ()
   (with-alien ((info (struct OSVERSIONINFO)))
@@ -876,15 +628,19 @@ absense."
   (whence dword))
 
 (defun lseeki64 (handle offset whence)
-  (multiple-value-bind (moved to-place)
-      (set-file-pointer-ex handle offset whence)
-    (if moved
-        (values to-place 0)
-        (values -1 (get-last-error)))))
+  (let ((type (get-file-type handle)))
+    (if (or (= type file-type-char)
+            (= type file-type-disk))
+        (multiple-value-bind (moved to-place)
+            (set-file-pointer-ex handle offset whence)
+          (if moved
+              (values to-place 0)
+              (values -1 (get-last-error))))
+        (values -1 0))))
 
 ;; File mapping support routines
-(define-alien-routine (#!+sb-unicode "CreateFileMappingW"
-                       #!-sb-unicode "CreateFileMappingA"
+(define-alien-routine (#+sb-unicode "CreateFileMappingW"
+                       #-sb-unicode "CreateFileMappingA"
                        create-file-mapping)
     handle
   (handle handle)
@@ -930,11 +686,11 @@ absense."
 (defconstant file-share-write #x02)
 
 ;; CreateFile (the real file-opening workhorse).
-(define-alien-routine (#!+sb-unicode "CreateFileW"
-                       #!-sb-unicode "CreateFileA"
+(define-alien-routine (#+sb-unicode "CreateFileW"
+                       #-sb-unicode "CreateFileA"
                        create-file)
     handle
-  (name (c-string #!+sb-unicode #!+sb-unicode :external-format :ucs-2))
+  (name (c-string #+sb-unicode #+sb-unicode :external-format :ucs-2))
   (desired-access dword)
   (share-mode dword)
   (security-attributes (* t))
@@ -949,100 +705,91 @@ absense."
 
 ;; GetFileAttribute is like a tiny subset of fstat(),
 ;; enough to distinguish directories from anything else.
-(define-alien-routine (#!+sb-unicode "GetFileAttributesW"
-                       #!-sb-unicode "GetFileAttributesA"
+(define-alien-routine (#+sb-unicode "GetFileAttributesW"
+                       #-sb-unicode "GetFileAttributesA"
                        get-file-attributes)
     dword
-  (name (c-string #!+sb-unicode #!+sb-unicode :external-format :ucs-2)))
+  (name (c-string #+sb-unicode #+sb-unicode :external-format :ucs-2)))
 
 (define-alien-routine ("CloseHandle" close-handle) bool
   (handle handle))
 
-(define-alien-routine ("_open_osfhandle" real-open-osfhandle)
+(define-alien-routine ("_open_osfhandle" open-osfhandle)
     int
   (handle handle)
   (flags int))
 
-;; Intended to be an imitation of sb!unix:unix-open based on
+;; Intended to be an imitation of sb-unix:unix-open based on
 ;; CreateFile, as complete as possibly.
 ;; FILE_FLAG_OVERLAPPED is a must for decent I/O.
 
-(defun unixlike-open (path flags mode &optional revertable)
-  (declare (type sb!unix:unix-pathname path)
-           (type fixnum flags)
-           (type sb!unix:unix-file-mode mode)
-           (ignorable mode))
+(defun unixlike-open (path flags &key revertable
+                                      overlapped)
+  (declare (type sb-unix:unix-pathname path)
+           (type fixnum flags))
   (let* ((disposition-flags
-          (logior
-           (if (zerop (logand sb!unix:o_creat flags)) 0 #b100)
-           (if (zerop (logand sb!unix:o_excl flags)) 0 #b010)
-           (if (zerop (logand sb!unix:o_trunc flags)) 0 #b001)))
+           (logior
+            (if (logtest sb-unix:o_creat flags) #b100 0)
+            (if (logtest sb-unix:o_excl flags) #b010 0)
+            (if (logtest sb-unix:o_trunc flags) #b001 0)))
          (create-disposition
-          ;; there are 8 combinations of creat|excl|trunc, some of
-          ;; them are equivalent. Case stmt below maps them to 5
-          ;; dispositions (see CreateFile manual).
-          (case disposition-flags
-            ((#b110 #b111) file-create-new)
-            ((#b001 #b011) file-truncate-existing)
-            ((#b000 #b010) file-open-existing)
-            (#b100 file-open-always)
-            (#b101 file-create-always))))
+           ;; there are 8 combinations of creat|excl|trunc, some of
+           ;; them are equivalent. Case stmt below maps them to 5
+           ;; dispositions (see CreateFile manual).
+           (case disposition-flags
+             ((#b110 #b111) file-create-new)
+             ((#b001 #b011) file-truncate-existing)
+             ((#b000 #b010) file-open-existing)
+             (#b100 file-open-always)
+             (#b101 file-create-always))))
     (let ((handle
-           (create-file path
-                        (logior
-                         (if revertable #x10000 0)
-                         (if (plusp (logand sb!unix:o_append flags))
-                             access-file-append-data
-                             0)
-                         (ecase (logand 3 flags)
-                           (0 FILE_GENERIC_READ)
-                           (1 FILE_GENERIC_WRITE)
-                           ((2 3) (logior FILE_GENERIC_READ
-                                          FILE_GENERIC_WRITE))))
-                        (logior FILE_SHARE_READ
-                                FILE_SHARE_WRITE)
-                        nil
-                        create-disposition
-                        (logior
-                         file-attribute-normal
-                         file-flag-overlapped
-                         file-flag-sequential-scan)
-                        0)))
-      (if (eql handle invalid-handle)
-          (values nil
-                  (let ((error-code (get-last-error)))
-                    (case error-code
-                      (#.error_file_not_found
-                       sb!unix:enoent)
-                      ((#.error_already_exists #.error_file_exists)
-                       sb!unix:eexist)
-                      (otherwise error-code))))
-          (progn
-            ;; FIXME: seeking to the end is not enough for real APPEND
-            ;; semantics, but it's better than nothing.
-            ;;   -- AK
-            ;;
-            ;; On the other hand, the CL spec implies the "better than
-            ;; nothing" seek-once semantics implemented here, and our
-            ;; POSIX backend is incorrect in implementing :APPEND as
-            ;; O_APPEND.  Other CL implementations get this right across
-            ;; platforms.
-            ;;
-            ;; Of course, it would be nice if we had :IF-EXISTS
-            ;; :ATOMICALLY-APPEND separately as an extension, and in
-            ;; that case, we will have to worry about supporting it
-            ;; here after all.
-            ;;
-            ;; I've tested this only very briefly (on XP and Windows 7),
-            ;; but my impression is that WriteFile (without documenting
-            ;; it?) is like ZwWriteFile, i.e. if we pass in -1 as the
-            ;; offset in our overlapped structure, WriteFile seeks to the
-            ;; end for us.  Should we depend on that?  How do we communicate
-            ;; our desire to do so to the runtime?
-            ;;   -- DFL
-            ;;
-            (set-file-pointer-ex handle 0 (if (plusp (logand sb!unix::o_append flags)) 2 0))
-            (values handle 0))))))
+            (create-file path
+                         (logior
+                          (if revertable #x10000 0)
+                          (if (logtest sb-unix:o_append flags)
+                              access-file-append-data
+                              0)
+                          (ecase (logand 3 flags)
+                            (0 FILE_GENERIC_READ)
+                            (1 FILE_GENERIC_WRITE)
+                            ((2 3) (logior FILE_GENERIC_READ
+                                           FILE_GENERIC_WRITE))))
+                         (logior FILE_SHARE_READ
+                                 FILE_SHARE_WRITE)
+                         nil
+                         create-disposition
+                         (if overlapped
+                             file-flag-overlapped
+                             file-attribute-normal)
+                         0)))
+      (cond ((eql handle invalid-handle)
+             (values nil (get-last-error)))
+            (t
+             ;; FIXME: seeking to the end is not enough for real APPEND
+             ;; semantics, but it's better than nothing.
+             ;;   -- AK
+             ;;
+             ;; On the other hand, the CL spec implies the "better than
+             ;; nothing" seek-once semantics implemented here, and our
+             ;; POSIX backend is incorrect in implementing :APPEND as
+             ;; O_APPEND.  Other CL implementations get this right across
+             ;; platforms.
+             ;;
+             ;; Of course, it would be nice if we had :IF-EXISTS
+             ;; :ATOMICALLY-APPEND separately as an extension, and in
+             ;; that case, we will have to worry about supporting it
+             ;; here after all.
+             ;;
+             ;; I've tested this only very briefly (on XP and Windows 7),
+             ;; but my impression is that WriteFile (without documenting
+             ;; it?) is like ZwWriteFile, i.e. if we pass in -1 as the
+             ;; offset in our overlapped structure, WriteFile seeks to the
+             ;; end for us.  Should we depend on that?  How do we communicate
+             ;; our desire to do so to the runtime?
+             ;;   -- DFL
+             ;;
+             (set-file-pointer-ex handle 0 (if (logtest sb-unix::o_append flags) 2 0))
+             (values handle 0))))))
 
 (define-alien-routine ("closesocket" close-socket) int (handle handle))
 (define-alien-routine ("shutdown" shutdown-socket) int (handle handle)
@@ -1129,10 +876,10 @@ absense."
 (defun duplicate-and-unwrap-fd (fd &key inheritp)
   (let ((me (get-current-process)))
     (multiple-value-bind (duplicated handle)
-        (duplicate-handle me (real-get-osfhandle fd)
+        (duplicate-handle me (get-osfhandle fd)
                           me 0 inheritp +duplicate-same-access+)
       (if duplicated
-          (prog1 handle (real-crt-close fd))
+          (prog1 handle (crt-close fd))
           (win32-error 'duplicate-and-unwrap-fd)))))
 
 (define-alien-routine ("CreatePipe" create-pipe) lispbool
@@ -1143,7 +890,7 @@ absense."
 
 (defun windows-pipe ()
   (multiple-value-bind (created read-handle write-handle)
-      (create-pipe nil 256)
+      (create-pipe nil 0)
     (if created (values read-handle write-handle)
         (win32-error 'create-pipe))))
 
@@ -1164,7 +911,7 @@ absense."
       allow
       (win32-error '(setf inheritable-handle-p))))
 
-(defun sb!unix:unix-dup (fd)
+(defun sb-unix:unix-dup (fd)
   (let ((me (get-current-process)))
     (multiple-value-bind (duplicated handle)
         (duplicate-handle me fd me 0 t +duplicate-same-access+)
@@ -1174,11 +921,11 @@ absense."
 
 (defun call-with-crt-fd (thunk handle &optional (flags 0))
   (multiple-value-bind (duplicate errno)
-      (sb!unix:unix-dup handle)
+      (sb-unix:unix-dup handle)
     (if duplicate
-        (let ((fd (real-open-osfhandle duplicate flags)))
+        (let ((fd (open-osfhandle duplicate flags)))
           (unwind-protect (funcall thunk fd)
-            (real-crt-close fd)))
+            (crt-close fd)))
         (values nil errno))))
 
 ;;; random seeding
@@ -1188,8 +935,8 @@ absense."
   (length dword)
   (buffer (* t)))
 
-(define-alien-routine (#!-sb-unicode "CryptAcquireContextA"
-                       #!+sb-unicode "CryptAcquireContextW"
+(define-alien-routine (#-sb-unicode "CryptAcquireContextA"
+                       #+sb-unicode "CryptAcquireContextW"
                        %crypt-acquire-context) lispbool
   (handle handle :out)
   (container system-string)
