@@ -129,6 +129,15 @@
   (let ((float (make-double-float (lvar-value hi) (lvar-value lo))))
     (if (float-nan-p float) (give-up-ir1-transform) float)))
 
+;;; I'd like to transition all the 64-bit backends to use the single-arg
+;;; %MAKE-DOUBLE-FLOAT constructor instead of the 2-arg MAKE-DOUBLE-FLOAT.
+;;; So we need a transform to fold constant calls for either.
+#+64-bit
+(deftransform %make-double-float ((bits) ((constant-arg t)))
+  "Conditional constant folding"
+  (let ((float (%make-double-float (lvar-value bits))))
+    (if (float-nan-p float) (give-up-ir1-transform) float)))
+
 ;;; On the face of it, these transforms are ridiculous because if we're going
 ;;; to express (MINUSP X) as (MINUSP (foo-FLOAT-BITS X)), then why not _always_
 ;;; transform MINUSP of a float into an integer comparison instead of a
@@ -216,49 +225,12 @@
 ;;; Given a number X, create a form suitable as a bound for an
 ;;; interval. Make the bound open if OPEN-P is T. NIL remains NIL.
 ;;; FIXME: as this is a constructor, shouldn't it be named MAKE-BOUND?
-#-sb-fluid (declaim (inline set-bound))
+(declaim (inline set-bound))
 (defun set-bound (x open-p)
   (if (and x open-p) (list x) x))
 
-;;; What is the CROSS-FLOAT-INFINITY-KLUDGE?
-;;;
-;;; SBCL's own implementation of floating point supports floating
-;;; point infinities. Some of the old CMU CL :PROPAGATE-FLOAT-TYPE and
-;;; :PROPAGATE-FUN-TYPE code, like the DEFOPTIMIZERs below, uses this
-;;; floating point support. Thus, we have to avoid running it on the
-;;; cross-compilation host, since we're not guaranteed that the
-;;; cross-compilation host will support floating point infinities.
-;;;
-;;; If we wanted to live dangerously, we could conditionalize the code
-;;; with #+(OR SBCL SB-XC) instead. That way, if the cross-compilation
-;;; host happened to be SBCL, we'd be able to run the infinity-using
-;;; code. Pro:
-;;;   * SBCL itself gets built with more complete optimization.
-;;; Con:
-;;;   * You get a different SBCL depending on what your cross-compilation
-;;;     host is.
-;;; So far the pros and cons seem seem to be mostly academic, since
-;;; AFAIK (WHN 2001-08-28) the propagate-foo-type optimizations aren't
-;;; actually important in compiling SBCL itself. If this changes, then
-;;; we have to decide:
-;;;   * Go for simplicity, leaving things as they are.
-;;;   * Go for performance at the expense of conceptual clarity,
-;;;     using #+(OR SBCL SB-XC) and otherwise leaving the build
-;;;     process as is.
-;;;   * Go for performance at the expense of build time, using
-;;;     #+(OR SBCL SB-XC) and also making SBCL do not just
-;;;     make-host-1.sh and make-host-2.sh, but a third step
-;;;     make-host-3.sh where it builds itself under itself. (Such a
-;;;     3-step build process could also help with other things, e.g.
-;;;     using specialized arrays to represent debug information.)
-;;;   * Rewrite the code so that it doesn't depend on unportable
-;;;     floating point infinities.
-
 ;;; optimizers for SCALE-FLOAT. If the float has bounds, new bounds
 ;;; are computed for the result, if possible.
-#-sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
-(progn
-
 (defun scale-float-derive-type-aux (f ex same-arg)
   (declare (ignore same-arg))
   (flet ((scale-bound (x n)
@@ -267,8 +239,8 @@
            ;; zeros.
            (set-bound
             (handler-case
-             (scale-float (type-bound-number x) n)
-             (floating-point-overflow ()
+                (scale-float (type-bound-number x) n)
+              (floating-point-overflow ()
                 nil))
             (consp x))))
     (when (and (numeric-type-p f) (numeric-type-p ex))
@@ -279,13 +251,13 @@
             (new-lo nil)
             (new-hi nil))
         (when f-hi
-          (if (< (float-sign (type-bound-number f-hi)) $0.0)
+          (if (sb-xc:< (float-sign (type-bound-number f-hi)) $0.0)
               (when ex-lo
                 (setf new-hi (scale-bound f-hi ex-lo)))
               (when ex-hi
                 (setf new-hi (scale-bound f-hi ex-hi)))))
         (when f-lo
-          (if (< (float-sign (type-bound-number f-lo)) $0.0)
+          (if (sb-xc:< (float-sign (type-bound-number f-lo)) $0.0)
               (when ex-hi
                 (setf new-lo (scale-bound f-lo ex-hi)))
               (when ex-lo
@@ -314,13 +286,13 @@
               ;; When converting a number to a float, the limits are
               ;; the same.
               (let* ((lo (bound-func (lambda (x)
-                                       (if (< x ,most-negative)
+                                       (if (sb-xc:< x ,most-negative)
                                            ,most-negative
                                            (coerce x ',type)))
                                      (numeric-type-low num)
                                      nil))
                      (hi (bound-func (lambda (x)
-                                       (if (< ,most-positive x )
+                                       (if (sb-xc:< ,most-positive x )
                                            ,most-positive
                                            (coerce x ',type)))
                                      (numeric-type-high num)
@@ -333,10 +305,9 @@
                 (type-error ()
                   nil)))))))
   (frob %single-float single-float
-        sb-xc:most-negative-single-float sb-xc:most-positive-single-float)
+        most-negative-single-float most-positive-single-float)
   (frob %double-float double-float
-        sb-xc:most-negative-double-float sb-xc:most-positive-double-float))
-) ; PROGN
+        most-negative-double-float most-positive-double-float))
 
 ;;;; float contagion
 
@@ -478,35 +449,6 @@
   (frob =)
   (frob = t))
 
-;;;; irrational derive-type methods
-
-;;; Derive the result to be float for argument types in the
-;;; appropriate domain.
-#+sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
-(dolist (stuff '((asin (real $-1.0 $1.0))
-                 (acos (real $-1.0 $1.0))
-                 (acosh (real $1.0))
-                 (atanh (real $-1.0 $1.0))
-                 (sqrt (real $0.0))))
-  (destructuring-bind (name type) stuff
-    (let ((type (specifier-type type)))
-      (setf (fun-info-derive-type (fun-info-or-lose name))
-            (lambda (call)
-              (declare (type combination call))
-              (when (csubtypep (lvar-type
-                                (first (combination-args call)))
-                               type)
-                (specifier-type 'float)))))))
-
-#+sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
-(defoptimizer (log derive-type) ((x &optional y))
-  (when (and (csubtypep (lvar-type x)
-                        (specifier-type '(real $0.0)))
-             (or (null y)
-                 (csubtypep (lvar-type y)
-                            (specifier-type '(real $0.0)))))
-    (specifier-type 'float)))
-
 ;;;; irrational transforms
 
 (macrolet ((def (name prim rtype)
@@ -606,7 +548,7 @@
 
 (deftransform phase ((x) ((float)) float)
   '(if (minusp (float-sign x))
-       (float sb-xc:pi x)
+       (float pi x)
        (float 0 x)))
 
 ;;; The number is of type REAL.
@@ -614,9 +556,6 @@
   (and (numeric-type-p type)
        (eq (numeric-type-complexp type) :real)))
 
-#-sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
-(progn
-
 ;;;; optimizers for elementary functions
 ;;;;
 ;;;; These optimizers compute the output range of the elementary
@@ -646,8 +585,6 @@
     (specifier-type `(or (,float-type ,(or lo '*) ,(or hi '*))
                          (complex ,float-type)))))
 
-) ; PROGN
-
 (eval-when (:compile-toplevel :execute)
   ;; So the problem with this hack is that it's actually broken.  If
   ;; the host does not have long floats, then setting *R-D-F-F* to
@@ -658,7 +595,6 @@
 ;;; Test whether the numeric-type ARG is within the domain specified by
 ;;; DOMAIN-LOW and DOMAIN-HIGH, consider negative and positive zero to
 ;;; be distinct.
-#-sb-xc-host  ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
 (defun domain-subtypep (arg domain-low domain-high)
   (declare (type numeric-type arg)
            (type (or real null) domain-low domain-high))
@@ -690,11 +626,11 @@
            (fp-pos-zero-p (f)           ; Is F +0.0?
              (and (floatp f) (zerop f) (= (float-sign-bit f) 0))))
       (and (or (null domain-low)
-               (and arg-lo (>= arg-lo-val domain-low)
+               (and arg-lo (sb-xc:>= arg-lo-val domain-low)
                     (not (and (fp-pos-zero-p domain-low)
                               (fp-neg-zero-p arg-lo)))))
            (or (null domain-high)
-               (and arg-hi (<= arg-hi-val domain-high)
+               (and arg-hi (sb-xc:<= arg-hi-val domain-high)
                     (not (and (fp-neg-zero-p domain-high)
                               (fp-pos-zero-p arg-hi)))))))))
 
@@ -707,9 +643,7 @@
 (defstruct (interval (:constructor %make-interval (low high))
                      (:copier nil))
   low high)
-
-#-sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
-(progn
+(declaim (freeze-type interval))
 
 ;;; Handle monotonic functions of a single variable whose domain is
 ;;; possibly part of the real line. ARG is the variable, FUN is the
@@ -799,10 +733,10 @@
 
   ;; These functions are only defined for part of the real line. The
   ;; condition selects the desired part of the line.
-  (frob asin $-1d0 $1d0 (sb-xc:- (sb-xc:/ sb-xc:pi 2)) (sb-xc:/ sb-xc:pi 2))
+  (frob asin $-1d0 $1d0 (sb-xc:- (sb-xc:/ pi 2)) (sb-xc:/ pi 2))
   ;; Acos is monotonic decreasing, so we need to swap the function
   ;; values at the lower and upper bounds of the input domain.
-  (frob acos $-1d0 $1d0 0 sb-xc:pi :increasingp nil)
+  (frob acos $-1d0 $1d0 0 pi :increasingp nil)
   (frob acosh $1d0 nil nil nil)
   (frob atanh $-1d0 $1d0 -1 1)
   ;; Kahan says that (sqrt -0.0) is -0.0, so use a specifier that
@@ -813,13 +747,17 @@
 ;;; y) = (exp (* y (log x))). However, computations done this way
 ;;; have too much roundoff. Thus we have to do it the hard way.
 (defun safe-expt (x y)
-  #+sb-xc-host (when (< (abs y) 10000) (expt x y))
-  #-sb-xc-host
-  (handler-case
-      (when (< (abs y) 10000)
-        (expt x y))
-    (error ()
-      nil)))
+  (when (and (numberp x) (numberp y))
+    (handler-case
+        (when (< (abs y) 10000)
+          (expt x y))
+      ;; Currently we can hide unanticipated errors (such as failure to use SB-XC: math
+      ;; when cross-compiling) as well as the anticipated potential problem of overflow.
+      ;; So don't handle anything when cross-compiling.
+      ;; FIXME: I think this should not handle ERROR, but just FLOATING-POINT-OVERFLOW.
+      (#+sb-xc-host nil
+       #-sb-xc-host error ()
+        nil))))
 
 ;;; Handle the case when x >= 1.
 (defun interval-expt-> (x y)
@@ -1071,7 +1009,7 @@
       (one-arg-derive-type x #'log-derive-type-aux-1 #'log)))
 
 (defun atan-derive-type-aux-1 (y)
-  (elfun-derive-type-simple y #'atan nil nil (sb-xc:- (sb-xc:/ sb-xc:pi 2)) (sb-xc:/ sb-xc:pi 2)))
+  (elfun-derive-type-simple y #'atan nil nil (sb-xc:- (sb-xc:/ pi 2)) (sb-xc:/ pi 2)))
 
 (defun atan-derive-type-aux-2 (y x same-arg)
   (declare (ignore same-arg))
@@ -1088,8 +1026,8 @@
              (make-numeric-type :class 'float
                                 :format format
                                 :complexp :real
-                                :low (coerce (sb-xc:- sb-xc:pi) bound-format)
-                                :high (coerce sb-xc:pi bound-format))))
+                                :low (coerce (sb-xc:- pi) bound-format)
+                                :high (coerce pi bound-format))))
           (t
            ;; The result is a float or a complex number
            (float-or-complex-float-type result-type)))))
@@ -1116,7 +1054,7 @@
                    (t (numeric-type-format arg))))
          (bound-type (or format 'float)))
     (cond ((numeric-type-real-p arg)
-           (case (interval-range-info (numeric-type->interval arg) $0.0)
+           (case (interval-range-info> (numeric-type->interval arg) $0.0)
              (+
               ;; The number is positive, so the phase is 0.
               (make-numeric-type :class 'float
@@ -1129,8 +1067,8 @@
               (make-numeric-type :class 'float
                                  :format format
                                  :complexp :real
-                                 :low (coerce sb-xc:pi bound-type)
-                                 :high (coerce sb-xc:pi bound-type)))
+                                 :low (coerce pi bound-type)
+                                 :high (coerce pi bound-type)))
              (t
               ;; We can't tell. The result is 0 or pi. Use a union
               ;; type for this.
@@ -1143,22 +1081,20 @@
                (make-numeric-type :class 'float
                                   :format format
                                   :complexp :real
-                                  :low (coerce sb-xc:pi bound-type)
-                                  :high (coerce sb-xc:pi bound-type))))))
+                                  :low (coerce pi bound-type)
+                                  :high (coerce pi bound-type))))))
           (t
            ;; We have a complex number. The answer is the range -pi
            ;; to pi. (-pi is included because we have -0.)
            (make-numeric-type :class 'float
                               :format format
                               :complexp :real
-                              :low (coerce (sb-xc:- sb-xc:pi) bound-type)
-                              :high (coerce sb-xc:pi bound-type))))))
+                              :low (coerce (sb-xc:- pi) bound-type)
+                              :high (coerce pi bound-type))))))
 
 (defoptimizer (phase derive-type) ((num))
   (one-arg-derive-type num #'phase-derive-type-aux #'phase))
 
-) ; PROGN
-
 (deftransform realpart ((x) ((complex rational)) *)
   '(%realpart x))
 (deftransform imagpart ((x) ((complex rational)) *)
@@ -1259,7 +1195,6 @@
                                             :complex)))))
       (specifier-type 'complex)))
 
-#-sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
 (defoptimizer (complex derive-type) ((re &optional im))
   (if im
       (two-arg-derive-type re im #'complex-derive-type-aux-2 #'complex)
@@ -1413,10 +1348,8 @@
 ;;; possible answer. This gets around the problem of doing range
 ;;; reduction correctly but still provides useful results when the
 ;;; inputs are union types.
-#-sb-xc-host ; (See CROSS-FLOAT-INFINITY-KLUDGE.)
-(progn
 (defun trig-derive-type-aux (arg domain fun
-                                 &optional def-lo def-hi (increasingp t))
+                             &optional def-lo def-hi (increasingp t))
   (etypecase arg
     (numeric-type
      (flet ((floatify-format ()
@@ -1462,7 +1395,7 @@
      ;; Derive the bounds if the arg is in [-pi/2, pi/2].
      (trig-derive-type-aux
       arg
-      (specifier-type `(float ,(sb-xc:- (sb-xc:/ sb-xc:pi 2)) ,(sb-xc:/ sb-xc:pi 2)))
+      (specifier-type `(float ,(sb-xc:- (sb-xc:/ pi 2)) ,(sb-xc:/ pi 2)))
       #'sin
       -1 1))
    #'sin))
@@ -1473,7 +1406,7 @@
    (lambda (arg)
      ;; Derive the bounds if the arg is in [0, pi].
      (trig-derive-type-aux arg
-                           (specifier-type `(float $0d0 ,sb-xc:pi))
+                           (specifier-type `(float $0d0 ,pi))
                            #'cos
                            -1 1
                            nil))
@@ -1485,8 +1418,8 @@
    (lambda (arg)
      ;; Derive the bounds if the arg is in [-pi/2, pi/2].
      (trig-derive-type-aux arg
-                           (specifier-type `(float ,(sb-xc:- (sb-xc:/ sb-xc:pi 2))
-                                                   ,(sb-xc:/ sb-xc:pi 2)))
+                           (specifier-type `(float ,(sb-xc:- (sb-xc:/ pi 2))
+                                                   ,(sb-xc:/ pi 2)))
                            #'tan
                            nil nil))
    #'tan))
@@ -1510,7 +1443,7 @@
                   (high (numeric-type-high arg)))
               (let ((new-low (most-negative-bound low high))
                     (new-high (most-positive-bound low high)))
-              (modified-numeric-type arg :low new-low :high new-high))))))
+                (modified-numeric-type arg :low new-low :high new-high))))))
     #'conjugate))
 
 (defoptimizer (cis derive-type) ((num))
@@ -1520,7 +1453,6 @@
        `(complex ,(or (numeric-type-format arg) 'float))))
     #'cis))
 
-) ; PROGN
 
 ;;;; TRUNCATE, FLOOR, CEILING, and ROUND
 
@@ -1653,8 +1585,8 @@
 ;;; For 32-bit big-endian machines, the bit patterns were those of subnormals.
 ;;; So thank goodness for that - it allowed detection of the problem.
 (defun test-ctype-involving-double-float ()
-  (specifier-type '(double-float #.sb-xc:pi)))
-(assert (sb-xc:= (numeric-type-low (test-ctype-involving-double-float)) sb-xc:pi))
+  (specifier-type '(double-float #.pi)))
+(assert (sb-xc:= (numeric-type-low (test-ctype-involving-double-float)) pi))
 
 ;;; Dummy functions to test that complex number are dumped correctly in genesis.
 (defun try-folding-complex-single ()
