@@ -31,9 +31,8 @@
   ;; constraint on the high end is that it must not exceed
   ;; WORD-SHIFT (defined below) due to the use of unboxed
   ;; word-aligned byte pointers as boxed values in various
-  ;; places.  FIXME: This should possibly be exposed for
-  ;; configuration via customize-target-features.
-  #+64-bit 1
+  ;; places.
+  #+64-bit (or #+ppc64 3 #-ppc64 1)
   ;; On 32-bit targets, this may be as low as 2 (for 30-bit
   ;; fixnums) and as high as 2 (for 30-bit fixnums).  The
   ;; constraint on the low end is simple overcrowding of the
@@ -62,10 +61,10 @@
 ;;; a mask to extract the type from a data block header word
 (defconstant widetag-mask (1- (ash 1 n-widetag-bits)))
 
-(defconstant sb-xc:most-positive-fixnum
+(defconstant most-positive-fixnum
     (1- (ash 1 n-positive-fixnum-bits))
   "the fixnum closest in value to positive infinity")
-(defconstant sb-xc:most-negative-fixnum
+(defconstant most-negative-fixnum
     (ash -1 n-positive-fixnum-bits)
   "the fixnum closest in value to negative infinity")
 
@@ -77,7 +76,7 @@
   ;; leaving one bit for a GC mark bit.
   (ldb (byte (- n-word-bits n-widetag-bits 1) 0) -1))
 
-(defconstant sb-xc:char-code-limit #-sb-unicode 256 #+sb-unicode #x110000
+(defconstant char-code-limit #-sb-unicode 256 #+sb-unicode #x110000
   "the upper exclusive bound on values produced by CHAR-CODE")
 
 (defconstant base-char-code-limit #-sb-unicode 256 #+sb-unicode 128)
@@ -95,13 +94,13 @@
 (defconstant sb-kernel::internal-time-bits 61)
 
 (defconstant most-positive-exactly-single-float-fixnum
-  (min (expt 2 single-float-digits) sb-xc:most-positive-fixnum))
+  (min (expt 2 single-float-digits) most-positive-fixnum))
 (defconstant most-negative-exactly-single-float-fixnum
-  (max (- (expt 2 single-float-digits)) sb-xc:most-negative-fixnum))
+  (max (- (expt 2 single-float-digits)) most-negative-fixnum))
 (defconstant most-positive-exactly-double-float-fixnum
-  (min (expt 2 double-float-digits) sb-xc:most-positive-fixnum))
+  (min (expt 2 double-float-digits) most-positive-fixnum))
 (defconstant most-negative-exactly-double-float-fixnum
-  (max (- (expt 2 double-float-digits)) sb-xc:most-negative-fixnum))
+  (max (- (expt 2 double-float-digits)) most-negative-fixnum))
 
 ;;;; Point where continuous area starting at dynamic-space-start bumps into
 ;;;; next space. Computed for genesis/constants.h, not used in Lisp.
@@ -122,22 +121,27 @@
 ;; To get a layout, you must call %INSTANCE-LAYOUT - don't assume index 0.
 (defconstant instance-data-start (+ #-compact-instance-header 1))
 
-;; The largest number that may appear in the header-data for an instance,
-;; and some other mostly-boxed objects, such as FDEFNs.
-;; This constraint exists because for objects managed by the immobile GC,
-;; their generation number is stored in the header, so we have to know
-;; how much to mask off to obtain the payload size.
-;; Objects whose payload gets capped to this limit are considered
-;; "short_boxed" objects in the sizetab[] array in 'gc-common'.
-;; Additionally there are "tiny_boxed" objects, the payload length of
-;; which can be expressed in 8 bits.
+;;; The largest number that may appear in the header-data for closures
+;;; and funcallable instances.
+;;; This constraint exists because for objects managed by the immobile GC,
+;;; their generation number is stored in the header, so we have to know
+;;; how much to mask off to obtain the payload size.
+;;; Objects whose payload gets capped to this limit are considered
+;;; "short_boxed" objects in the sizetab[] array in 'gc-common'.
+;;; Additionally there are "tiny_boxed" objects, the payload length of
+;;; which can be expressed in 8 bits.
 (defconstant short-header-max-words #x7fff)
 
+;;; Amount to righ-shift an instance header to get the length.
+;;; Similar consideration as above with regard to use of generation# byte.
+(defconstant instance-length-shift 10)
+(defconstant instance-length-mask #x3FFF)
+
 ;;; Is X a fixnum in the target Lisp?
 #+sb-xc-host
 (defun fixnump (x)
   (and (integerp x)
-       (<= sb-xc:most-negative-fixnum x sb-xc:most-positive-fixnum)))
+       (<= most-negative-fixnum x most-positive-fixnum)))
 
 ;;; Helper macro for defining FIXUP-CODE-OBJECT so that its body
 ;;; can be the same between the host and target.
@@ -159,6 +163,7 @@
                     (sb-fasl::descriptor-gspace code)))) ; use CODE, not SAP
               (sap-ref-8 (sap offset) `(sb-fasl::bvref-8 ,sap ,offset))
               (sap-ref-32 (sap offset) `(sb-fasl::bvref-32 ,sap ,offset))
+              (sap-ref-64 (sap offset) `(sb-fasl::bvref-64 ,sap ,offset))
               (sap-ref-word (sap offset) `(sb-fasl::bvref-word ,sap ,offset)))
      ,@body))
 
@@ -166,3 +171,5 @@
 ;;; The offset from the fault address reported to the runtime to the
 ;;; END of the global safepoint page.
 (defconstant gc-safepoint-trap-offset n-word-bytes)
+
+#+sb-xc-host (deftype sb-xc:fixnum () `(signed-byte ,n-fixnum-bits))
