@@ -41,7 +41,7 @@
 
 (define-move-fun (load-constant 5) (vop x y)
   ((constant) (descriptor-reg any-reg))
-  (let ((offset (- (ash (tn-offset x) word-shift) other-pointer-lowtag)))
+  (let ((offset (- (tn-byte-offset x) other-pointer-lowtag)))
     (cond
       ((typep offset '(signed-byte 11))
        (inst ldw offset code-tn y))
@@ -212,7 +212,7 @@
     ;; the bignum-construction, doing the shift in the delay slot.
     (inst comb := temp zero-tn done)
     (inst sll x n-fixnum-tag-bits y)
-    (load-constant vop (emit-constant (1+ sb-xc:most-positive-fixnum))
+    (load-constant vop (emit-constant (1+ most-positive-fixnum))
                    y)
     DONE))
 
@@ -226,7 +226,7 @@
     ;; the bignum-construction, doing the shift in the delay slot.
     (inst comb := temp zero-tn done)
     (inst sll x n-fixnum-tag-bits y)
-    (load-constant vop (emit-constant (1- sb-xc:most-negative-fixnum))
+    (load-constant vop (emit-constant (1- most-negative-fixnum))
                    y)
     DONE))
 
