@@ -217,7 +217,10 @@
 (defvar *functional-escape-info*)
 
 (defun functional-may-escape-p (functional)
-  (binding* ((table (or *functional-escape-info*
+  (binding* ((functional (if (lambda-p functional)
+                             (lambda-home functional)
+                             functional))
+             (table (or *functional-escape-info*
                         ;; Many components have no escapes, so we
                         ;; allocate it lazily.
                         (setf *functional-escape-info*
@@ -262,7 +265,7 @@
              (policy exit (zerop check-tag-existence))
              ;; Dynamic extent is a promise things won't escape --
              ;; and an explicit request to avoid heap consing.
-             (member (lambda-extent exit-lambda) '(:always-dynamic :maybe-dynamic))
+             (member (lambda-extent exit-lambda) '(truly-dynamic-extent dynamic-extent))
              ;; If the exit lambda cannot escape, then we should be safe.
              ;; ...since the escape analysis is kinda new, and not particularly
              ;; exhaustively tested, let alone proven, disable it for SAFETY 3.
@@ -302,7 +305,7 @@
          (info (make-nlx-info cleanup exit))
          (new-block (insert-cleanup-code (list exit-block) next-block
                                          entry
-                                         `(%nlx-entry ',info)
+                                         `(%nlx-entry ,(opaquely-quote info))
                                          cleanup))
          (component (block-component new-block)))
     (unlink-blocks exit-block new-block)
@@ -360,7 +363,7 @@
       (mapc (lambda (x)
               (setf (node-derived-type x) *wild-type*))
             (leaf-refs exit-fun))
-      (substitute-leaf (find-constant info) exit-fun))
+      (substitute-leaf (find-constant (opaquely-quote info)) exit-fun))
     (when lvar
       (let ((node (block-last (nlx-info-target info))))
         (unless (node-lvar node)
@@ -463,18 +466,18 @@
                        (basic-combination-args node))))
           (ecase (cleanup-kind cleanup)
             (:special-bind
-             (code `(%special-unbind ',(leaf-source-name (lvar-value (car args))))))
+             (code `(%special-unbind ',(lvar-value (car args)))))
             (:catch
-             (code `(%catch-breakup)))
+             (code `(%catch-breakup ,(opaquely-quote (car (cleanup-info cleanup))))))
             (:unwind-protect
-              (code `(%unwind-protect-breakup))
-              (let ((fun (ref-leaf (lvar-uses (second args)))))
+             (code `(%unwind-protect-breakup ,(opaquely-quote (car (cleanup-info cleanup)))))
+             (let ((fun (ref-leaf (lvar-uses (second args)))))
                 (when (functional-p fun)
                   (reanalyze-funs fun)
                   (code `(%funcall ,fun)))))
             ((:block :tagbody)
              (dolist (nlx (cleanup-info cleanup))
-               (code `(%lexical-exit-breakup ',nlx))))
+               (code `(%lexical-exit-breakup ,(opaquely-quote nlx)))))
             (:dynamic-extent
              (when (cleanup-info cleanup)
                (code `(%cleanup-point))))
@@ -482,7 +485,7 @@
              (code `(%primitive set-nsp ,(ref-leaf node))))))))
     (flet ((coalesce-unbinds (code)
              code
-              #!+(vop-named sb-c:unbind-n)
+              #+(vop-named sb-c:unbind-n)
               (loop with cleanup
                     while code
                     do (setf cleanup (pop code))
