@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;; Make a TN for the argument count passing location for a
 ;;; non-local entry.
@@ -42,7 +42,7 @@
   (:args (catch :scs (descriptor-reg))
          (nfp :scs (descriptor-reg))
          (nsp :scs (descriptor-reg)))
-  #!+sb-thread (:temporary (:scs (any-reg)) temp)
+  #+sb-thread (:temporary (:scs (any-reg)) temp)
   (:vop-var vop)
   (:generator 10
     (store-tl-symbol-value catch *current-catch-block* temp)
@@ -84,7 +84,7 @@
   (:temporary (:scs (descriptor-reg)) temp)
   (:temporary (:scs (non-descriptor-reg)) ndescr)
   (:generator 22
-    (inst addi block cfp-tn (* (tn-offset tn) n-word-bytes))
+    (inst addi block cfp-tn (tn-byte-offset tn))
     (load-tl-symbol-value temp *current-unwind-protect-block*)
     (storew temp block unwind-block-uwp-slot)
     (storew cfp-tn block unwind-block-cfp-slot)
@@ -105,7 +105,7 @@
   (:temporary (:scs (descriptor-reg) :target block :to (:result 0)) result)
   (:temporary (:scs (non-descriptor-reg)) ndescr)
   (:generator 44
-    (inst addi result cfp-tn (* (tn-offset tn) n-word-bytes))
+    (inst addi result cfp-tn (tn-byte-offset tn))
     (load-tl-symbol-value temp *current-unwind-protect-block*)
     (storew temp result catch-block-uwp-slot)
     (storew cfp-tn result catch-block-cfp-slot)
@@ -121,33 +121,33 @@
     (move block result)))
 
 
-;;; Just set the current unwind-protect to TN's address.  This instantiates an
+;;; Just set the current unwind-protect to UWP.  This instantiates an
 ;;; unwind block as an unwind-protect.
 ;;;
 (define-vop (set-unwind-protect)
-  (:args (tn))
-  (:temporary (:scs (descriptor-reg)) new-uwp)
-  #!+sb-thread (:temporary (:scs (any-reg)) temp)
+  (:args (uwp :scs (any-reg)))
+  #+sb-thread (:temporary (:scs (any-reg)) temp)
   (:generator 7
-    (inst addi new-uwp cfp-tn (* (tn-offset tn) n-word-bytes))
-    (store-tl-symbol-value new-uwp *current-unwind-protect-block* temp)))
+    (store-tl-symbol-value uwp *current-unwind-protect-block* temp)))
 
 
-(define-vop (unlink-catch-block)
+(define-vop (%catch-breakup)
+  (:args (current-block))
+  (:ignore current-block)
   (:temporary (:scs (any-reg)) block)
-  #!+sb-thread (:temporary (:scs (any-reg)) temp)
+  #+sb-thread (:temporary (:scs (any-reg)) temp)
   (:policy :fast-safe)
-  (:translate %catch-breakup)
   (:generator 17
     (load-tl-symbol-value block *current-catch-block*)
     (loadw block block catch-block-previous-catch-slot)
     (store-tl-symbol-value block *current-catch-block* temp)))
 
-(define-vop (unlink-unwind-protect)
+(define-vop (%unwind-protect-breakup)
+  (:args (current-block))
+  (:ignore current-block)
   (:temporary (:scs (any-reg)) block)
-  #!+sb-thread (:temporary (:scs (any-reg)) temp)
+  #+sb-thread (:temporary (:scs (any-reg)) temp)
   (:policy :fast-safe)
-  (:translate %unwind-protect-breakup)
   (:generator 17
     (load-tl-symbol-value block *current-unwind-protect-block*)
     (loadw block block unwind-block-uwp-slot)
