@@ -55,136 +55,76 @@
 ;;; Assume that any constant operand is the second arg...
 
 (define-vop (fast-fixnum-binop fast-safe-arith-op)
-  (:args (x :target r :scs (any-reg zero))
-         (y :target r :scs (any-reg zero)))
+  (:args (x :target r :scs (any-reg))
+         (y :target r :scs (any-reg)))
   (:arg-types tagged-num tagged-num)
   (:results (r :scs (any-reg)))
   (:result-types tagged-num)
   (:note "inline fixnum arithmetic"))
 
 (define-vop (fast-unsigned-binop fast-safe-arith-op)
-  (:args (x :target r :scs (unsigned-reg zero))
-         (y :target r :scs (unsigned-reg zero)))
+  (:args (x :target r :scs (unsigned-reg))
+         (y :target r :scs (unsigned-reg)))
   (:arg-types unsigned-num unsigned-num)
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
-  (:note "inline (unsigned-byte 32) arithmetic"))
+  (:note "inline word arithmetic"))
 
 (define-vop (fast-signed-binop fast-safe-arith-op)
-  (:args (x :target r :scs (signed-reg zero))
-         (y :target r :scs (signed-reg zero)))
+  (:args (x :target r :scs (signed-reg))
+         (y :target r :scs (signed-reg)))
   (:arg-types signed-num signed-num)
   (:results (r :scs (signed-reg)))
   (:result-types signed-num)
-  (:note "inline (signed-byte 32) arithmetic"))
+  (:note "inline signed-word arithmetic"))
 
 (define-vop (fast-fixnum-binop-c fast-safe-arith-op)
-  (:args (x :target r :scs (any-reg zero)))
+  (:args (x :target r :scs (any-reg)))
   (:info y)
-  (:arg-types tagged-num
-              (:constant (and (signed-byte 14) (not (integer 0 0)))))
-  (:results (r :scs (any-reg)))
-  (:result-types tagged-num)
-  (:note "inline fixnum arithmetic"))
-
-(define-vop (fast-fixnum-binop30-c fast-safe-arith-op)
-  (:args (x :target r :scs (any-reg zero)))
-  (:info y)
-  (:arg-types tagged-num
-              (:constant (and (signed-byte 30) (not (integer 0 0)))))
+  (:arg-types tagged-num (:constant fixnum))
   (:results (r :scs (any-reg)))
   (:result-types tagged-num)
   (:note "inline fixnum arithmetic"))
 
 (define-vop (fast-fixnum-logop-c fast-safe-arith-op)
-  (:args (x :target r :scs (any-reg zero)))
+  (:args (x :target r :scs (any-reg)))
   (:info y)
-  (:arg-types tagged-num
-              (:constant (and (unsigned-byte 14) (not (integer 0 0)))))
-  (:results (r :scs (any-reg)))
-  (:result-types tagged-num)
-  (:note "inline fixnum logical op"))
-
-(define-vop (fast-fixnum-logop30-c fast-safe-arith-op)
-  (:args (x :target r :scs (any-reg zero)))
-  (:info y)
-  (:arg-types tagged-num
-              (:constant (and (unsigned-byte 16) (not (integer 0 0)))))
+  (:arg-types tagged-num (:constant fixnum))
   (:results (r :scs (any-reg)))
   (:result-types tagged-num)
   (:note "inline fixnum logical op"))
 
 (define-vop (fast-unsigned-binop-c fast-safe-arith-op)
-  (:args (x :target r :scs (unsigned-reg zero)))
+  (:args (x :target r :scs (unsigned-reg)))
   (:info y)
-  (:arg-types unsigned-num
-              (:constant (and (signed-byte 16) (not (integer 0 0)))))
+  (:arg-types unsigned-num (:constant word))
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
-  (:note "inline (unsigned-byte 32) arithmetic"))
-
-(define-vop (fast-unsigned-binop32-c fast-safe-arith-op)
-  (:args (x :target r :scs (unsigned-reg zero)))
-  (:info y)
-  (:arg-types unsigned-num
-              (:constant (and (unsigned-byte 32) (not (integer 0 0)))))
-  (:results (r :scs (unsigned-reg)))
-  (:result-types unsigned-num)
-  (:note "inline (unsigned-byte 32) arithmetic"))
-
-(define-vop (fast-signed-binop32-c fast-safe-arith-op)
-  (:args (x :target r :scs (signed-reg zero)))
-  (:info y)
-  (:arg-types signed-num
-              (:constant (and (signed-byte 32) (not (integer 0 0)))))
-  (:results (r :scs (signed-reg)))
-  (:result-types signed-num)
-  (:note "inline (signed-byte 32) arithmetic"))
+  (:note "inline word arithmetic"))
 
 (define-vop (fast-unsigned-logop-c fast-safe-arith-op)
-  (:args (x :target r :scs (unsigned-reg zero)))
+  (:args (x :target r :scs (unsigned-reg)))
   (:info y)
-  (:arg-types unsigned-num
-              (:constant (and (unsigned-byte 16) (not (integer 0 0)))))
+  (:arg-types unsigned-num (:constant word))
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
-  (:note "inline (unsigned-byte 32) logical op"))
-
-(define-vop (fast-unsigned-logop32-c fast-safe-arith-op)
-  (:args (x :target r :scs (unsigned-reg zero)))
-  (:info y)
-  (:arg-types unsigned-num
-              (:constant (and (unsigned-byte 32) (not (integer 0 0)))))
-  (:results (r :scs (unsigned-reg)))
-  (:result-types unsigned-num)
-  (:note "inline (unsigned-byte 32) logical op"))
-
-(define-vop (fast-signed-logop32-c fast-safe-arith-op)
-  (:args (x :target r :scs (signed-reg zero)))
-  (:info y)
-  (:arg-types signed-num
-              (:constant (and (unsigned-byte 32) (not (integer 0 0)))))
-  (:results (r :scs (signed-reg)))
-  (:result-types signed-num)
-  (:note "inline (signed-byte 32) logical op"))
+  (:note "inline word logical op"))
 
 (define-vop (fast-signed-binop-c fast-safe-arith-op)
-  (:args (x :target r :scs (signed-reg zero)))
+  (:args (x :target r :scs (signed-reg)))
   (:info y)
-  (:arg-types signed-num
-              (:constant (and (signed-byte 16) (not (integer 0 0)))))
+  (:arg-types signed-num (:constant signed-word))
   (:results (r :scs (signed-reg)))
   (:result-types signed-num)
-  (:note "inline (signed-byte 32) arithmetic"))
+  (:note "inline signed-word arithmetic"))
 
 (define-vop (fast-signed-logop-c fast-safe-arith-op)
-  (:args (x :target r :scs (signed-reg zero)))
+  (:args (x :target r :scs (signed-reg)))
   (:info y)
-  (:arg-types signed-num
-              (:constant (and (unsigned-byte 16) (not (integer 0 0)))))
+  (:arg-types signed-num (:constant signed-word))
   (:results (r :scs (signed-reg)))
   (:result-types signed-num)
-  (:note "inline (signed-byte 32) logical op"))
+  (:note "inline signed-word logical op"))
 
 (defmacro !define-var-binop (translate untagged-penalty op
                              &optional arg-swap restore-fixnum-mask)
@@ -201,7 +141,7 @@
          ;; FIXME: remind me what convention we used for 64bitizing
          ;; stuff?  -- CSR, 2003-08-27
          ,@(when restore-fixnum-mask
-             `((inst clrrdi r temp (1- n-lowtag-bits))))))
+             `((inst clrrdi r temp n-fixnum-tag-bits)))))
      (define-vop (,(symbolicate "FAST-" translate "/SIGNED=>SIGNED")
                   fast-signed-binop)
        (:translate ,translate)
@@ -217,181 +157,6 @@
              `(inst ,op r y x)
              `(inst ,op r x y))))))
 
-;;; FIXME: the code has really only been checked for adds; we could do
-;;; subtracts, too, but my brain is not up to the task of figuring out
-;;; signs and borrows.
-(defmacro !define-const-binop (translate untagged-penalty op &optional (shifted-op nil))
-  `(progn
-     (define-vop (,(symbolicate 'fast- translate '-c/fixnum=>fixnum)
-                  ,(if shifted-op
-                       'fast-fixnum-binop30-c
-                       'fast-fixnum-binop-c))
-       (:translate ,translate)
-       ,@(when shifted-op
-          `((:temporary (:sc any-reg :target r) temp)))
-       (:generator 1
-        ,(if shifted-op
-             `(let* ((y (fixnumize y))
-                     (high-half (ldb (byte 16 16) y))
-                     (low-half (ldb (byte 16 0) y)))
-               ;; Compare %LR in insts.lisp.
-               (cond
-                 ((and (logbitp 15 low-half) (= high-half #xffff))
-                  ;; Let sign-extension do the work for us, but make sure
-                  ;; to turn LOW-HALF into a signed integer.
-                  (inst ,op r x (dpb low-half (byte 16 0) -1)))
-                 ((and (not (logbitp 15 low-half)) (zerop high-half))
-                  (inst ,op r x low-half))
-                 ((zerop low-half)
-                  (inst ,shifted-op r x (if (logbitp 15 high-half)
-                                            (dpb high-half (byte 16 0) -1)
-                                            high-half)))
-                 (t
-                  ;; Check to see whether compensating for the sign bit
-                  ;; of LOW-HALF is necessary.
-                  (let ((high-half (let ((top (if (logbitp 15 low-half)
-                                                  (ldb (byte 16 0)
-                                                       (1+ high-half))
-                                                  high-half)))
-                                     (if (logbitp 15 top)
-                                         (dpb top (byte 16 0) -1)
-                                         top))))
-                    (inst ,shifted-op temp x high-half)
-                    (inst ,op r temp low-half)))))
-             `(inst ,op r x (fixnumize y)))))
-     (define-vop (,(symbolicate 'fast- translate '-c/signed=>signed)
-                  ,(if shifted-op
-                       'fast-signed-binop32-c
-                       'fast-signed-binop-c))
-       (:translate ,translate)
-       ,@(when shifted-op
-          `((:temporary (:sc non-descriptor-reg :target r) temp)))
-       (:generator ,untagged-penalty
-        ,(if shifted-op
-             `(let ((high-half (ldb (byte 16 16) y))
-                    (low-half (ldb (byte 16 0) y)))
-               ;; Compare %LR in insts.lisp.
-               (cond
-                 ((and (logbitp 15 low-half) (= high-half #xffff))
-                  ;; Let sign-extension do the work for us, but make sure
-                  ;; to turn LOW-HALF into a signed integer.
-                  (inst ,op r x (dpb low-half (byte 16 0) -1)))
-                 ((and (not (logbitp 15 low-half)) (zerop high-half))
-                  (inst ,op r x low-half))
-                 ((zerop low-half)
-                  (inst ,shifted-op r x (if (logbitp 15 high-half)
-                                            (dpb high-half (byte 16 0) -1)
-                                            high-half)))
-                 (t
-                  ;; Check to see whether compensating for the sign bit
-                  ;; of LOW-HALF is necessary.
-                  (let ((high-half (let ((top (if (logbitp 15 low-half)
-                                                  (ldb (byte 16 0)
-                                                       (1+ high-half))
-                                                  high-half)))
-                                     (if (logbitp 15 top)
-                                         (dpb top (byte 16 0) -1)
-                                         top))))
-                    (inst ,shifted-op temp x high-half)
-                    (inst ,op r temp low-half)))))
-             `(inst ,op r x y))))
-     (define-vop (,(symbolicate 'fast- translate '-c/unsigned=>unsigned)
-                  ,(if shifted-op
-                       'fast-unsigned-binop32-c
-                       'fast-unsigned-binop-c))
-       (:translate ,translate)
-       ,@(when shifted-op
-          `((:temporary (:sc non-descriptor-reg :target r) temp)))
-       (:generator ,untagged-penalty
-        ,(if shifted-op
-             `(let ((high-half (ldb (byte 16 16) y))
-                    (low-half (ldb (byte 16 0) y)))
-               ;; Compare %LR in insts.lisp.
-               (cond
-                 ((and (logbitp 15 low-half) (= high-half #xffff))
-                  ;; Let sign-extension do the work for us, but make sure
-                  ;; to turn LOW-HALF into a signed integer.
-                  (inst ,op r x (dpb low-half (byte 16 0) -1)))
-                 ((and (not (logbitp 15 low-half)) (zerop high-half))
-                  (inst ,op r x low-half))
-                 ((zerop low-half)
-                  (inst ,shifted-op r x (if (logbitp 15 high-half)
-                                            (dpb high-half (byte 16 0) -1)
-                                            high-half)))
-                 (t
-                  ;; Check to see whether compensating for the sign bit
-                  ;; of LOW-HALF is necessary.
-                  (let ((high-half (let ((top (if (logbitp 15 low-half)
-                                                  (ldb (byte 16 0)
-                                                       (1+ high-half))
-                                                  high-half)))
-                                     (if (logbitp 15 top)
-                                         (dpb top (byte 16 0) -1)
-                                         top))))
-                    (inst ,shifted-op temp x high-half)
-                    (inst ,op r temp low-half)))))
-             `(inst ,op r x y))))))
-
-;;; For logical operations, we don't have to worry about signed bit
-;;; propagation from the lower half of a 32-bit operand.
-(defmacro !define-const-logop (translate untagged-penalty op &optional (shifted-op nil))
-  `(progn
-     (define-vop (,(symbolicate 'fast- translate '-c/fixnum=>fixnum)
-                  ,(if shifted-op
-                       'fast-fixnum-logop30-c
-                       'fast-fixnum-logop-c))
-       (:translate ,translate)
-       ,@(when shifted-op
-          `((:temporary (:sc any-reg :target r) temp)))
-       (:generator 1
-        ,(if shifted-op
-             `(let* ((y (fixnumize y))
-                     (high-half (ldb (byte 16 16) y))
-                     (low-half (ldb (byte 16 0) y)))
-               (cond
-                 ((zerop high-half) (inst ,op r x low-half))
-                 ((zerop low-half) (inst ,shifted-op r x high-half))
-                 (t
-                  (inst ,shifted-op temp x high-half)
-                  (inst ,op r temp low-half))))
-             `(inst ,op r x (fixnumize y)))))
-     (define-vop (,(symbolicate 'fast- translate '-c/signed=>signed)
-                  ,(if shifted-op
-                       'fast-signed-logop32-c
-                       'fast-signed-logop-c))
-       (:translate ,translate)
-       ,@(when shifted-op
-          `((:temporary (:sc non-descriptor-reg :target r) temp)))
-       (:generator ,untagged-penalty
-        ,(if shifted-op
-             `(let ((high-half (ldb (byte 16 16) y))
-                    (low-half (ldb (byte 16 0) y)))
-               (cond
-                 ((zerop high-half) (inst ,op r x low-half))
-                 ((zerop low-half) (inst ,shifted-op r x high-half))
-                 (t
-                  (inst ,shifted-op temp x high-half)
-                  (inst ,op r temp low-half))))
-             `(inst ,op r x y))))
-     (define-vop (,(symbolicate 'fast- translate '-c/unsigned=>unsigned)
-                  ,(if shifted-op
-                       'fast-unsigned-logop32-c
-                       'fast-unsigned-logop-c))
-       (:translate ,translate)
-       ,@(when shifted-op
-          `((:temporary (:sc non-descriptor-reg :target r) temp)))
-       (:generator ,untagged-penalty
-        ,(if shifted-op
-             `(let ((high-half (ldb (byte 16 16) y))
-                    (low-half (ldb (byte 16 0) y)))
-               (cond
-                 ((zerop high-half) (inst ,op r x low-half))
-                 ((zerop low-half) (inst ,shifted-op r x high-half))
-                 (t
-                  (inst ,shifted-op temp x high-half)
-                  (inst ,op r temp low-half))))
-             `(inst ,op r x y))))))
-
 (!define-var-binop + 4 add)
 (!define-var-binop - 4 sub)
 (!define-var-binop logand 2 and)
@@ -405,17 +170,62 @@
 (!define-var-binop lognand 2 nand nil t)
 (!define-var-binop lognor 2 nor nil t)
 
-(!define-const-binop + 4 addi addis)
-(!define-const-binop - 4 subi)
-;;; Implementing a 32-bit immediate version of LOGAND wouldn't be any
-;;; better than loading the 32-bit constant via LR and then performing
-;;; an /AND/.  So don't bother.  (It would be better in some cases, such
-;;; as when one half of the word is zeros--we save a register--but we
-;;; would have specified one temporary register in the VOP, so we lose
-;;; any possible advantage.)
-(!define-const-logop logand 2 andi.)
-(!define-const-logop logior 2 ori oris)
-(!define-const-logop logxor 2 xori xoris)
+(defun generate-fast-+-c (r x y)
+  (cond ((typep y '(signed-byte 16))
+         (inst addi r x y))
+        ;; See if this can be done as an addis + addi.
+        ;; If bit 15 is on, 1 is added to the high part to undo the
+        ;; effect of sign-extension from the low. The post-adjustment
+        ;; high part needs to fit in (signed-byte 16)
+        ((typep (+ (ash y -16) (ldb (byte 1 15) y)) '(signed-byte 16))
+         (inst addis r x (+ (ash y -16) (ldb (byte 1 15) y)))
+         (let ((low (ldb (byte 16 0) y)))
+           (unless (zerop low)
+             (inst addi r r low))))
+        (t
+         (inst lr temp-reg-tn y)
+         (inst add r x temp-reg-tn))))
+
+(macrolet ((define-const-binop (translate untagged-penalty)
+             `(progn
+                (define-vop (,(symbolicate 'fast- translate '-c/fixnum=>fixnum) fast-fixnum-binop-c)
+                  (:translate ,translate)
+                  (:generator 1 (generate-fast-+-c r x (fixnumize (,translate y)))))
+                (define-vop (,(symbolicate 'fast- translate '-c/signed=>signed) fast-signed-binop-c)
+                  (:translate ,translate)
+                  (:generator ,untagged-penalty (generate-fast-+-c r x (,translate y))))
+                (define-vop (,(symbolicate 'fast- translate '-c/unsigned=>unsigned) fast-unsigned-binop-c)
+                  (:translate ,translate)
+                  (:generator ,untagged-penalty (generate-fast-+-c r x (,translate y)))))))
+  (define-const-binop + 4)
+  (define-const-binop - 4))
+
+(macrolet ((define-const-logop (translate untagged-penalty op shifted-op general-op)
+  `(flet ((emit (r x y)
+            (cond ((typep y '(unsigned-byte 16))
+                   (inst ,op r x y))
+                  ((and (typep (ash y -16) '(unsigned-byte 16)) ; effectively (unsigned-byte 32)
+                        ;; logical AND can't be split into two instructions
+                        ,@(if (eq translate 'logand) '((zerop (ldb (byte 16 0) y)))))
+                   (inst ,shifted-op r x (ash y -16))
+                   (when (ldb-test (byte 16 0) y)
+                     (inst ,op r r (ldb (byte 16 0) y)))) ; not sign-extended
+                  (t ; everything else: just load the constant from memory
+                   (inst lr temp-reg-tn y)
+                   (inst ,general-op r x temp-reg-tn)))))
+     (define-vop (,(symbolicate 'fast- translate '-c/fixnum=>fixnum) fast-fixnum-logop-c)
+       (:translate ,translate)
+       (:generator 1 (emit r x (fixnumize y))))
+     (define-vop (,(symbolicate 'fast- translate '-c/signed=>signed) fast-signed-logop-c)
+       (:translate ,translate)
+       (:generator ,untagged-penalty (emit r x y)))
+     (define-vop (,(symbolicate 'fast- translate '-c/unsigned=>unsigned) fast-unsigned-logop-c)
+       (:translate ,translate)
+       (:generator ,untagged-penalty (emit r x y))))))
+
+  (define-const-logop logand 2 andi. andis. and)
+  (define-const-logop logior 2 ori oris or)
+  (define-const-logop logxor 2 xori xoris xor))
 
 (define-vop (fast-*/fixnum=>fixnum fast-fixnum-binop)
   (:temporary (:scs (non-descriptor-reg)) temp)
@@ -424,41 +234,31 @@
     (inst sradi temp y n-fixnum-tag-bits)
     (inst mulld r x temp)))
 
-(define-vop (fast-*-c/fixnum=>fixnum fast-fixnum-binop-c)
-  (:translate *)
-  (:arg-types tagged-num
-              (:constant (and (signed-byte 16) (not (integer 0 0)))))
-  (:generator 1
-    (inst mulli r x y)))
-
-(define-vop (fast-*-bigc/fixnum=>fixnum fast-fixnum-binop-c)
-  (:translate *)
-  (:arg-types tagged-num
-              (:constant (and fixnum (not (signed-byte 16)))))
-  (:temporary (:scs (non-descriptor-reg)) temp)
-  (:generator 1
-    (inst lr temp y)
-    (inst mulld r x temp)))
-
 (define-vop (fast-*/signed=>signed fast-signed-binop)
   (:translate *)
   (:generator 4
     (inst mulld r x y)))
 
-(define-vop (fast-*-c/signed=>signed fast-signed-binop-c)
-  (:translate *)
-  (:generator 3
-    (inst mulli r x y)))
-
 (define-vop (fast-*/unsigned=>unsigned fast-unsigned-binop)
   (:translate *)
   (:generator 4
     (inst mulld r x y)))
 
-(define-vop (fast-*-c/unsigned=>unsigned fast-unsigned-binop-c)
-  (:translate *)
-  (:generator 3
-    (inst mulli r x y)))
+(macrolet ((generate ()
+             '(cond ((typep y '(signed-byte 16))
+                     (inst mulli r x y))
+                    (t
+                     (inst lr temp-reg-tn y)
+                     (inst mulld r x temp-reg-tn)))))
+  (define-vop (fast-*-c/fixnum=>fixnum fast-fixnum-binop-c)
+    (:translate *)
+    (:generator 1 (generate)))
+  (define-vop (fast-*-c/signed=>signed fast-signed-binop-c)
+    (:translate *)
+    (:generator 3 (generate)))
+  (define-vop (fast-*-c/unsigned=>unsigned fast-unsigned-binop-c)
+    (:translate *)
+    (:generator 3 (generate))))
 
 ;;; Shifting
 
@@ -503,7 +303,7 @@
       (inst cmpdi ndesc 63)
       (inst srd result number ndesc)
       (inst ble done)
-      (move result zero-tn)
+      (inst li result 0)
       (inst b done)
 
       POSITIVE
@@ -523,10 +323,10 @@
   (:policy :fast-safe)
   (:generator 4
     (cond
-      ((and (minusp amount) (< amount -63)) (move result zero-tn))
+      ((and (minusp amount) (< amount -63)) (inst li result 0))
       ((minusp amount) (inst srdi result number (- amount)))
       ;; possible because this is used in the modular version too
-      ((> amount 63) (move result zero-tn))
+      ((> amount 63) (inst li result 0))
       (t (inst sldi result number amount)))))
 
 (define-vop (fast-ash/signed=>signed)
@@ -602,12 +402,10 @@
   (:policy :fast-safe)
   (:args (arg :scs (unsigned-reg)))
   (:arg-types unsigned-num)
-  (:results (res :scs (any-reg)))
-  (:result-types positive-fixnum)
-  (:temporary (:scs (non-descriptor-reg)) temp)
+  (:results (res :scs (unsigned-reg)))
+  (:result-types unsigned-num)
   (:generator 2
-    (inst popcntd temp arg)
-    (inst slwi res arg n-fixnum-tag-bits)))
+    (inst popcntd res arg)))
 
 ;;;; %LDB
 
@@ -703,30 +501,50 @@
     (sb-c::give-up-ir1-transform))
   '(%primitive fast-ash-left-mod64/unsigned=>unsigned integer count))
 
-(macrolet
-    ((define-modular-backend (fun &optional constantp)
-       (let ((mfun-name (symbolicate fun '-mod64))
-             (modvop (symbolicate 'fast- fun '-mod64/unsigned=>unsigned))
-             (modcvop (symbolicate 'fast- fun 'mod64-c/unsigned=>unsigned))
-             (vop (symbolicate 'fast- fun '/unsigned=>unsigned))
-             (cvop (symbolicate 'fast- fun '-c/unsigned=>unsigned)))
-         `(progn
-            (define-modular-fun ,mfun-name (x y) ,fun :untagged nil 64)
-            (define-vop (,modvop ,vop)
-              (:translate ,mfun-name))
-            ,@(when constantp
-                `((define-vop (,modcvop ,cvop)
-                    (:translate ,mfun-name))))))))
-  (define-modular-backend + t)
-  (define-modular-backend - t)
-  (define-modular-backend * t)
-  (define-modular-backend logeqv)
-  (define-modular-backend lognand)
-  (define-modular-backend lognor)
-  (define-modular-backend logandc1)
-  (define-modular-backend logandc2)
-  (define-modular-backend logorc1)
-  (define-modular-backend logorc2))
+(defmacro define-mod-binop ((name prototype) function)
+  `(define-vop (,name ,prototype)
+     (:args (x :target r :scs (unsigned-reg signed-reg))
+            (y :scs (unsigned-reg signed-reg)))
+     (:arg-types untagged-num untagged-num)
+     (:results (r :scs (unsigned-reg signed-reg) :from (:argument 0)))
+     (:result-types unsigned-num)
+     (:translate ,function)))
+
+(defmacro define-mod-binop-c ((name prototype) function)
+  `(define-vop (,name ,prototype)
+     (:args (x :target r :scs (unsigned-reg signed-reg)))
+     (:info y)
+     (:arg-types untagged-num (:constant (and (signed-byte 30) (not (integer 0 0)))))
+     (:results (r :scs (unsigned-reg signed-reg) :from (:argument 0)))
+     (:result-types unsigned-num)
+     (:translate ,function)))
+
+(macrolet ((def (name -c-p)
+             (let ((fun64   (symbolicate name "-MOD64"))
+                   (funfx   (symbolicate name "-MODFX"))
+                   (vopu    (symbolicate "FAST-" name "/UNSIGNED=>UNSIGNED"))
+                   (vopcu   (symbolicate "FAST-" name "-C/UNSIGNED=>UNSIGNED"))
+                   (vopf    (symbolicate "FAST-" name "/FIXNUM=>FIXNUM"))
+                   (vopcf   (symbolicate "FAST-" name "-C/FIXNUM=>FIXNUM"))
+                   (vop64u  (symbolicate "FAST-" name "-MOD64/WORD=>UNSIGNED"))
+                   (vop64f  (symbolicate "FAST-" name "-MOD64/FIXNUM=>FIXNUM"))
+                   (vop64cu (symbolicate "FAST-" name "-MOD64-C/WORD=>UNSIGNED"))
+                   (vopfxf  (symbolicate "FAST-" name "-MODFX/FIXNUM=>FIXNUM"))
+                   (vopfxcf (symbolicate "FAST-" name "-MODFX-C/FIXNUM=>FIXNUM")))
+               `(progn
+                  (define-modular-fun ,fun64 (x y) ,name :untagged nil 64)
+                  (define-modular-fun ,funfx (x y) ,name :tagged t
+                    #.(- n-word-bits n-fixnum-tag-bits))
+                  (define-mod-binop (,vop64u ,vopu) ,fun64)
+                  (define-vop (,vop64f ,vopf) (:translate ,fun64))
+                  (define-vop (,vopfxf ,vopf) (:translate ,funfx))
+                  ,@(when -c-p
+                      `((define-mod-binop-c (,vop64cu ,vopcu) ,fun64)
+                        (define-vop (,vopfxcf ,vopcf) (:translate ,funfx))))))))
+  (def + t)
+  (def - t)
+  (def * nil))
+
 
 ;;;; Binary conditional VOPs:
 
@@ -736,35 +554,35 @@
   (:policy :fast-safe))
 
 (define-vop (fast-conditional/fixnum fast-conditional)
-  (:args (x :scs (any-reg zero))
-         (y :scs (any-reg zero)))
+  (:args (x :scs (any-reg))
+         (y :scs (any-reg)))
   (:arg-types tagged-num tagged-num)
   (:note "inline fixnum comparison"))
 
 (define-vop (fast-conditional-c/fixnum fast-conditional/fixnum)
-  (:args (x :scs (any-reg zero)))
-  (:arg-types tagged-num (:constant (signed-byte 14)))
+  (:args (x :scs (any-reg)))
+  (:arg-types tagged-num (:constant (signed-byte #.(- 16 n-fixnum-tag-bits))))
   (:info target not-p y))
 
 (define-vop (fast-conditional/signed fast-conditional)
-  (:args (x :scs (signed-reg zero))
-         (y :scs (signed-reg zero)))
+  (:args (x :scs (signed-reg))
+         (y :scs (signed-reg)))
   (:arg-types signed-num signed-num)
   (:note "inline (signed-byte 64) comparison"))
 
 (define-vop (fast-conditional-c/signed fast-conditional/signed)
-  (:args (x :scs (signed-reg zero)))
+  (:args (x :scs (signed-reg)))
   (:arg-types signed-num (:constant (signed-byte 16)))
   (:info target not-p y))
 
 (define-vop (fast-conditional/unsigned fast-conditional)
-  (:args (x :scs (unsigned-reg zero))
-         (y :scs (unsigned-reg zero)))
+  (:args (x :scs (unsigned-reg))
+         (y :scs (unsigned-reg)))
   (:arg-types unsigned-num unsigned-num)
   (:note "inline (unsigned-byte 64) comparison"))
 
 (define-vop (fast-conditional-c/unsigned fast-conditional/unsigned)
-  (:args (x :scs (unsigned-reg zero)))
+  (:args (x :scs (unsigned-reg)))
   (:arg-types unsigned-num (:constant (unsigned-byte 16)))
   (:info target not-p y))
 
@@ -832,6 +650,52 @@
         (inst andi. test x (ash 1 y))
         (inst andis. test x (ash 1 (- y 16))))
     (inst b? (if not-p :eq :ne) target)))
+(define-vop (mask-signed-field-word/c)
+  (:translate sb-c::mask-signed-field)
+  (:policy :fast-safe)
+  (:args (x :scs (signed-reg unsigned-reg) :target r))
+  (:arg-types (:constant (integer 0 64)) untagged-num)
+  (:results (r :scs (signed-reg)))
+  (:result-types signed-num)
+  (:info width)
+  (:generator 3
+    (aver (/= width 0))
+    (cond ((= width 64)
+           (move r x))
+          (t
+           (inst sldi r x (- n-word-bits width))
+           (inst sradi r r (- n-word-bits width))))))
+
+(define-vop (mask-signed-field-bignum/c)
+  (:translate sb-c::mask-signed-field)
+  (:policy :fast-safe)
+  (:args (x :scs (descriptor-reg) :target r))
+  (:arg-types (:constant (integer 0 64)) bignum)
+  (:results (r :scs (signed-reg)))
+  (:result-types signed-num)
+  (:info width)
+  (:generator 4
+    (aver (/= width 0))
+    (loadw r x bignum-digits-offset other-pointer-lowtag)
+    (inst sldi r r (- n-word-bits width))
+    (inst sradi r r (- n-word-bits width))))
+
+(define-vop (mask-signed-field-fixnum)
+  (:translate sb-c::mask-signed-field)
+  (:policy :fast-safe)
+  (:args (x :scs (descriptor-reg) :target r))
+  (:arg-types (:constant (eql #.n-fixnum-bits)) t)
+  (:results (r :scs (any-reg)))
+  (:result-types fixnum)
+  (:info width)
+  (:ignore width)
+  (:generator 5
+    (move r x)
+    (inst andi. temp-reg-tn r fixnum-tag-mask)
+    (inst beq DONE)
+    (loadw temp-reg-tn r bignum-digits-offset other-pointer-lowtag)
+    (inst sldi r temp-reg-tn (- n-word-bits n-fixnum-bits))
+    DONE))
 
 (define-vop (fast-if-</fixnum fast-conditional/fixnum)
   (:translate <)
@@ -941,8 +805,8 @@
 ;;;
 
 (define-vop (fast-eql/fixnum fast-conditional)
-  (:args (x :scs (any-reg descriptor-reg zero))
-         (y :scs (any-reg zero)))
+  (:args (x :scs (any-reg descriptor-reg))
+         (y :scs (any-reg)))
   (:arg-types tagged-num tagged-num)
   (:note "inline fixnum comparison")
   (:translate eql)
@@ -955,7 +819,7 @@
   (:variant-cost 7))
 
 (define-vop (fast-eql-c/fixnum fast-conditional/fixnum)
-  (:args (x :scs (any-reg descriptor-reg zero)))
+  (:args (x :scs (any-reg descriptor-reg)))
   (:arg-types tagged-num (:constant (signed-byte #.(- 16 n-fixnum-tag-bits))))
   (:info target not-p y)
   (:translate eql)
@@ -1017,7 +881,7 @@
   (:variant bignum-digits-offset other-pointer-lowtag)
   (:translate sb-bignum:%bignum-set)
   (:args (object :scs (descriptor-reg))
-         (index :scs (any-reg immediate zero))
+         (index :scs (any-reg immediate))
          (value :scs (unsigned-reg)))
   (:arg-types t positive-fixnum unsigned-num)
   (:results (result :scs (unsigned-reg)))
@@ -1049,9 +913,10 @@
             (carry :scs (unsigned-reg)))
   (:result-types unsigned-num positive-fixnum)
   (:generator 3
-    (inst addic temp c -1)
-    (inst adde result a b)
-    (inst addze carry zero-tn)))
+    (inst addic temp c -1) ; set CA flag if 'c' is nonzero
+    (inst adde result a b) ; add a + b + CA and carry out
+    (inst li temp 0)
+    (inst addze carry temp)))
 
 (define-vop (sub-w/borrow)
   (:translate sb-bignum:%subtract-with-borrow)
@@ -1067,7 +932,8 @@
   (:generator 4
     (inst addic temp c -1)
     (inst sube result a b)
-    (inst addze borrow zero-tn)))
+    (inst li temp 0)
+    (inst addze borrow temp)))
 
 (define-vop (bignum-mult-and-add-3-arg)
   (:translate sb-bignum:%multiply-and-add)
@@ -1083,9 +949,8 @@
             (lo :scs (unsigned-reg)))
   (:result-types unsigned-num unsigned-num)
   (:generator 40
-   (inst tw :t x x)
-    (inst mulhwu hi-temp x y)
-    (inst mullw lo-temp x y)
+    (inst mulhdu hi-temp x y)
+    (inst mulld lo-temp x y)
     (inst addc lo lo-temp carry-in)
     (inst addze hi hi-temp)))
 
@@ -1104,9 +969,8 @@
             (lo :scs (unsigned-reg)))
   (:result-types unsigned-num unsigned-num)
   (:generator 40
-   (inst tw :t x x)
-    (inst mulhwu hi-temp x y)
-    (inst mullw lo-temp x y)
+    (inst mulhdu hi-temp x y)
+    (inst mulld lo-temp x y)
     (inst addc lo-temp lo-temp carry-in)
     (inst addze hi-temp hi-temp)
     (inst addc lo lo-temp prev)
@@ -1122,9 +986,8 @@
             (lo :scs (unsigned-reg) :from (:eval 0)))
   (:result-types unsigned-num unsigned-num)
   (:generator 40
-   (inst tw :t x x)
-    (inst mullw lo x y)
-    (inst mulhwu hi x y)))
+    (inst mulld lo x y)
+    (inst mulhdu hi x y)))
 
 (define-vop (mulhi)
   (:translate %multiply-high)
@@ -1163,61 +1026,46 @@
   (:generator 1
     (inst sradi digit fixnum n-fixnum-tag-bits)))
 
+;;; Algorithm from page 74 of of Power ISA version 2.07
+;;; under "Programming Note" for the divweu instruction
+;;; with modifications as described at the divdeu instruction.
 (define-vop (bignum-floor)
   (:translate sb-bignum:%bigfloor)
   (:policy :fast-safe)
-  (:args (num-high :scs (unsigned-reg) :target rem)
-         (num-low :scs (unsigned-reg) :target rem-low)
-         (denom :scs (unsigned-reg) :to (:eval 1)))
+  ;; I tried to tighten up the TN lifetimes for better packing
+  ;; (in fewer physical registers), but I got it wrong and everything broke.
+  ;; We have enough non-descriptor registers that this isn't a problem.
+  (:args (Dh :scs (unsigned-reg)) ; dividend high
+         (Dl :scs (unsigned-reg)) ; dividend low
+         (Dv :scs (unsigned-reg) :to :save)) ; divisor
   (:arg-types unsigned-num unsigned-num unsigned-num)
-  (:temporary (:scs (unsigned-reg) :from (:argument 1)) rem-low)
-  (:temporary (:scs (unsigned-reg) :from (:eval 0)) temp)
-  (:results (quo :scs (unsigned-reg) :from (:eval 0))
-            (rem :scs (unsigned-reg) :from (:argument 0)))
+  (:results (Q :scs (unsigned-reg))  ; quotient
+            (R :scs (unsigned-reg))) ; remainder
   (:result-types unsigned-num unsigned-num)
-  (:generator 325 ; number of inst assuming targeting works.
-    ;; FIXME: can't we use the CPU's divider ?
-    (inst tw :t quo quo)
-    (move rem num-high)
-    (move rem-low num-low)
-    (flet ((maybe-subtract (&optional (guess temp))
-             (inst subi temp guess 1)
-             (inst and temp temp denom)
-             (inst sub rem rem temp))
-           (sltu (res x y)
-             (inst subfc res y x)
-             (inst subfe res res res)
-             (inst neg res res)))
-      (sltu quo rem denom)
-      (maybe-subtract quo)
-      (dotimes (i 32)
-        (inst slwi rem rem 1)
-        (inst srwi temp rem-low 31)
-        (inst or rem rem temp)
-        (inst slwi rem-low rem-low 1)
-        (sltu temp rem denom)
-        (inst slwi quo quo 1)
-        (inst or quo quo temp)
-        (maybe-subtract)))
-    (inst not quo quo)))
-
-#|
-
-(define-vop (bignum-floor)
-  (:translate sb-bignum:%bigfloor)
-  (:policy :fast-safe)
-  (:args (div-high :scs (unsigned-reg) :target rem)
-         (div-low :scs (unsigned-reg) :target quo)
-         (divisor :scs (unsigned-reg)))
-  (:arg-types unsigned-num unsigned-num unsigned-num)
-  (:results (quo :scs (unsigned-reg) :from (:argument 1))
-            (rem :scs (unsigned-reg) :from (:argument 0)))
-  (:result-types unsigned-num unsigned-num)
-  (:generator 300
-    (inst mtmq div-low)
-    (inst div quo div-high divisor)
-    (inst mfmq rem)))
-|#
+  (:temporary (:scs (non-descriptor-reg)) q1)
+  (:temporary (:scs (non-descriptor-reg)) q2)
+  (:temporary (:scs (non-descriptor-reg)) -r1)
+  (:temporary (:scs (non-descriptor-reg)) r2)
+  (:generator 15
+    ;; I don't know how we know when utilizing this vop
+    ;; that overflow won't occur. But the code for x86[-64]
+    ;; does not check for overflow either.
+    (inst divdeu q1 Dh Dv)   ; q1
+    (inst divdu  q2 Dl Dv)   ; q2
+    (inst mulld -r1 q1 Dv)   ; -r1 = q1 * Dv
+    (inst mulld r2 q2 Dv)    ; temp = q2 * Dv
+    (inst subf r2 r2 Dl)     ; r2 = Dl - (q2 * Dv)
+    ;; move to result TNs
+    (inst add Q q1 q2)       ; Q = q1 + q2
+    (inst subf R -r1 r2)     ; R = r1 + r2
+    (inst cmpld R r2)        ; R < r2 ?
+    (inst blt ADJ)           ; must adjust Q and R if yes
+    (inst cmpld R Dv)        ; R >= Dv ?
+    (inst blt DONE)          ; must adjust Q and R if yes
+    ADJ
+    (inst addi Q Q 1)        ; Q = Q + 1
+    (inst subf R Dv R)       ; R = R - Dv
+    DONE))
 
 (define-vop (signify-digit)
   (:translate sb-bignum:%fixnum-digit-with-correct-sign)
