@@ -173,11 +173,33 @@
 ;;; here, the values are read by an automatically generated reader method.
 (defmethod add-direct-subclass ((class class) (subclass class))
   (with-slots (direct-subclasses) class
-    (pushnew subclass direct-subclasses :test #'eq)
+    (with-world-lock ()
+      (pushnew subclass direct-subclasses :test #'eq)
+      (let ((layout (class-wrapper subclass)))
+        (when layout
+          (let* ((classoid (layout-classoid layout)))
+            (dovector (super-layout (layout-inherits layout))
+              (let* ((super (layout-classoid super-layout))
+                     (subclasses (or (classoid-subclasses super)
+                                     (or (classoid-subclasses super)
+                                         (setf (classoid-subclasses super)
+                                               (make-hash-table :hash-function #'type-hash-value
+                                                                :test 'eq
+                                                                :synchronized t))))))
+                (when subclasses
+                  (setf (gethash classoid subclasses) layout))))))))
     subclass))
 (defmethod remove-direct-subclass ((class class) (subclass class))
   (with-slots (direct-subclasses) class
-    (setq direct-subclasses (remove subclass direct-subclasses))
+    (with-world-lock ()
+      (setq direct-subclasses (remove subclass direct-subclasses))
+      ;; Remove from classoid subclasses as well.
+      (let ((classoid (class-classoid subclass)))
+        (dovector (super-layout (layout-inherits (classoid-layout classoid)))
+          (let* ((super (layout-classoid super-layout))
+                 (subclasses (classoid-subclasses super)))
+            (when subclasses
+              (remhash classoid subclasses))))))
     subclass))
 
 ;;; Maintaining the direct-methods and direct-generic-functions backpointers.
@@ -291,7 +313,14 @@
   (let ((table (specializer-method-table self))
         (object (specializer-object self)))
     (if create
-        (sb-impl::puthash-if-absent object table (lambda () (cons nil nil)))
+        (sb-thread::with-system-mutex ((hash-table-lock table))
+          (ensure-gethash object table (cons nil nil)))
+        ;; FIXME: in the CREATE case we're locking, because we might be inserting,
+        ;; and must ensure mutual exclusion with other writers. Fine,
+        ;; but SBCL's hash-tables aren't *reader* *safe* with any writer, so
+        ;; how can we know that this branch is safe? Honestly I have no idea.
+        ;; I believe that the lock should scope be widened,
+        ;; surrounding either branch of the IF.
         (gethash object table))))
 
 (defun map-specializers (function)
@@ -529,10 +558,10 @@
             (let ((layout (make-wrapper 0 class)))
               (setf (slot-value class 'wrapper) layout)
               (let ((cpl (compute-preliminary-cpl class)))
-                (setf (layout-inherits layout)
-                      (order-layout-inherits
-                       (map 'simple-vector #'class-wrapper
-                            (reverse (rest cpl))))))
+                (set-layout-inherits layout
+                                     (order-layout-inherits
+                                      (map 'simple-vector #'class-wrapper
+                                           (reverse (rest cpl))))))
               (register-layout layout :invalidate t))))
         (mapc #'make-preliminary-layout (class-direct-subclasses class))))))
 
@@ -1194,6 +1223,9 @@
                      cell)))))
       (unless (slot-definition-class eslotd)
         (setf (slot-definition-class eslotd) class))
+      (let ((info (slot-definition-info eslotd)))
+        (setf (slot-info-allocation info) (slot-definition-allocation eslotd)
+              (slot-info-location info) (slot-definition-location eslotd)))
       (initialize-internal-slot-functions eslotd))))
 
 (defmethod compute-slots :around ((class standard-class))
@@ -1223,18 +1255,18 @@
 
 (defmethod compute-effective-slot-definition-initargs
     ((class slot-class) direct-slotds)
-  (let* ((name nil)
-         (initfunction nil)
-         (initform nil)
-         (initargs nil)
-         (allocation nil)
-         (allocation-class nil)
-         (type t)
-         (documentation nil)
-         (documentationp nil)
-         (namep  nil)
-         (initp  nil)
-         (allocp nil))
+  (let ((name nil)
+        (initfunction nil)
+        (initform nil)
+        (initargs nil)
+        (allocation nil)
+        (allocation-class nil)
+        (type t)
+        (documentation nil)
+        (documentationp nil)
+        (namep  nil)
+        (initp  nil)
+        (allocp nil))
 
     (dolist (slotd direct-slotds)
       (when slotd
@@ -1254,7 +1286,7 @@
           (setq allocation (slot-definition-allocation slotd)
                 allocation-class (slot-definition-class slotd)
                 allocp t))
-        (setq initargs (append (slot-definition-initargs slotd) initargs))
+        (setq initargs (union (slot-definition-initargs slotd) initargs))
         (let ((slotd-type (slot-definition-type slotd)))
           (setq type (cond
                        ((eq type t) slotd-type)
@@ -1335,16 +1367,21 @@
                              'source source-location)))
 
 (defmethod remove-reader-method ((class slot-class) generic-function)
-  (let ((method (get-method generic-function () (list class) nil)))
+  (let ((method
+         (and (= (length (arg-info-metatypes (gf-arg-info generic-function))) 1)
+              (get-method generic-function () (list class) nil))))
     (when method (remove-method generic-function method))))
 
 (defmethod remove-writer-method ((class slot-class) generic-function)
   (let ((method
-          (get-method generic-function () (list *the-class-t* class) nil)))
+         (and (= (length (arg-info-metatypes (gf-arg-info generic-function))) 2)
+              (get-method generic-function () (list *the-class-t* class) nil))))
     (when method (remove-method generic-function method))))
 
 (defmethod remove-boundp-method ((class slot-class) generic-function)
-  (let ((method (get-method generic-function () (list class) nil)))
+  (let ((method
+         (and (= (length (arg-info-metatypes (gf-arg-info generic-function))) 1)
+              (get-method generic-function () (list class) nil))))
     (when method (remove-method generic-function method))))
 
 ;;; MAKE-READER-METHOD-FUNCTION and MAKE-WRITER-METHOD-FUNCTION
@@ -1484,13 +1521,17 @@
                                 old-class new-class)
   (do () ((typep value slot-type))
     (restart-case
-        (bad-type value slot-type
+        (error 'simple-type-error
+               :datum value
+               :expected-type slot-type
+               :format-control
                   (sb-format:tokens
                   "~@<Error during ~A. Current value in slot ~
                    ~/sb-ext:print-symbol-with-prefix/ of an instance ~
                    of ~S is ~S, which does not match the new slot type ~
                    ~S in class ~S.~:@>")
-                  context slot-name old-class value slot-type new-class)
+               :format-arguments
+               (list context slot-name old-class value slot-type new-class))
       (use-value (new-value)
         :interactive read-evaluated-form
         :report (lambda (stream)
@@ -1504,13 +1545,13 @@
 
 (defun %set-slot-value-checking-type (context slots slot value
                                       safe old-class new-class)
-  (setf (clos-slots-ref slots (slot-definition-location slot))
-        (if (and safe (not (unbound-marker-p value)))
-            (let ((name (slot-definition-name slot))
-                  (type (slot-definition-type slot)))
-              (%ensure-slot-value-type context name type value
-                                       old-class new-class))
-            value)))
+  (let ((value (if (and safe (not (unbound-marker-p value)))
+                   (let ((name (slot-definition-name slot))
+                         (type (slot-definition-type slot)))
+                     (%ensure-slot-value-type context name type value
+                                              old-class new-class))
+                   value)))
+    (setf (clos-slots-ref slots (slot-definition-location slot)) value)))
 
 (defvar *in-obsolete-instance-trap* nil)
 
@@ -1525,11 +1566,12 @@
 
 (defun %obsolete-instance-trap (owrapper nwrapper instance)
   (cond
-    ((layout-for-std-class-p owrapper)
-     (binding* ((class (wrapper-class* nwrapper))
-                (copy (allocate-instance class)) ;??? allocate-instance ???
+    ((layout-for-pcl-obj-p owrapper)
+     (binding* ((class (wrapper-class nwrapper))
                 (oslots (get-slots instance))
-                (nslots (get-slots copy))
+                (nwrapper (class-wrapper class))
+                (nslots (make-array (layout-length nwrapper)
+                                    :initial-element +slot-unbound+))
                 (added ())
                 (discarded ())
                 (plist ())
@@ -1565,7 +1607,16 @@
          ;; Go through all the old local slots.
          (dolist (old old-instance-slots)
            (let* ((name (slot-definition-name old))
-                  (value (clos-slots-ref oslots (slot-definition-location old))))
+                  (cell (find-slot-cell owrapper name))
+                  (location (car cell))
+                  (value (cond
+                           ((fixnump location)
+                            (clos-slots-ref oslots location))
+                           ((not location)
+                            (let ((location (slot-info-location (cdr cell))))
+                              (aver (integerp location))
+                              (clos-slots-ref oslots (slot-info-location (cdr cell)))))
+                           (t (bug "non-FIXNUM non-NULL location in cell: ~S" cell)))))
              (unless (unbound-marker-p value)
                (let ((new (assq name layout)))
                  (cond (new
@@ -1597,7 +1648,11 @@
          (dolist (cell layout)
            (push (car cell) added)))
 
-       (%swap-wrappers-and-slots instance copy)
+       (if (std-instance-p instance)
+           (setf (%instance-layout instance) nwrapper
+                 (std-instance-slots instance) nslots)
+           (setf (%fun-layout instance) nwrapper
+                 (%fsc-instance-slots instance) nslots))
 
        (update-instance-for-redefined-class
         instance added discarded plist)
@@ -1609,31 +1664,30 @@
        (error 'obsolete-structure :datum instance)))))
 
 
-(defun %change-class (instance new-class initargs)
-  (declare (notinline allocate-instance))
-  (binding* ((old-class (class-of instance))
-             (copy (allocate-instance new-class))
-             (new-wrapper (get-wrapper copy))
-             (old-wrapper (class-wrapper old-class))
+(defun %change-class (copy instance new-class initargs)
+  (binding* ((new-wrapper (class-wrapper (ensure-class-finalized new-class)))
+             (new-slots (make-array (layout-length new-wrapper)
+                                    :initial-element +slot-unbound+))
+             (old-wrapper (layout-of instance))
+             (old-class (wrapper-class old-wrapper))
              (old-slots (get-slots instance))
-             (new-slots (get-slots copy))
              (safe (safe-p new-class))
-             (new-wrapper-slots (layout-slot-list new-wrapper))
-             (old-wrapper-slots (layout-slot-list old-wrapper)))
-    (labels ((find-instance-slot (name slots)
-               (loop for slot in slots
-                     when (and (eq (slot-definition-allocation slot) :instance)
-                               (eq (slot-definition-name slot) name))
-                     return slot))
-             (initarg-for-slot-p (slot)
+             (new-wrapper-slots (layout-slot-list new-wrapper)))
+    (if (functionp copy)
+        (setf (%fun-layout copy) new-wrapper
+              (%fsc-instance-slots copy) new-slots)
+        (setf (%instance-layout copy) new-wrapper
+              (%std-instance-slots copy) new-slots))
+    (flet ((initarg-for-slot-p (slot)
+             (when initargs
                (dolist (slot-initarg (slot-definition-initargs slot))
                  (unless (unbound-marker-p
                           (getf initargs slot-initarg +slot-unbound+))
-                   (return t))))
-             (set-value (value slotd)
-               (%set-slot-value-checking-type
-                'change-class new-slots slotd value safe
-                old-class new-class)))
+                   (return t)))))
+           (set-value (value slotd)
+             (%set-slot-value-checking-type
+              'change-class new-slots slotd value safe
+              old-class new-class)))
 
       ;; "The values of local slots specified by both the class CTO
       ;; and CFROM are retained. If such a local slot was unbound, it
@@ -1641,26 +1695,34 @@
       (dolist (new new-wrapper-slots)
         (when (and (not (initarg-for-slot-p new))
                    (eq (slot-definition-allocation new) :instance))
-          (binding* ((old (find-instance-slot (slot-definition-name new) old-wrapper-slots)
-                          :exit-if-null)
-                     (value (clos-slots-ref old-slots (slot-definition-location old))))
-            (set-value value new))))
-
-      ;; "The values of slots specified as shared in the class CFROM and
-      ;; as local in the class CTO are retained."
-      (dolist (old old-wrapper-slots)
-        (when (eq (slot-definition-allocation old) :class)
-         (binding* ((slot-and-val (slot-definition-location old))
-                    (new (find-instance-slot (car slot-and-val) new-wrapper-slots)
-                         :exit-if-null))
-           (set-value (cdr slot-and-val) new)))))
+          (binding* ((cell (find-slot-cell old-wrapper (slot-definition-name new))
+                           :exit-if-null)
+                     (location (car cell))
+                     (value (cond
+                              ((fixnump location)
+                               (clos-slots-ref old-slots location))
+                              ((not location)
+                               (let ((info (cdr cell)))
+                                 (case (slot-info-allocation info)
+                                   (:instance
+                                    (clos-slots-ref old-slots (slot-info-location info)))
+                                   (:class (cdr (slot-info-location info))))))
+                              (t
+                               (cdr location)))))
+            (set-value value new)))))
 
     ;; Make the copy point to the old instance's storage, and make the
     ;; old instance point to the new storage.
-    (%swap-wrappers-and-slots instance copy)
-
+    ;; All uses of %CHANGE-CLASS are under the world lock, but that doesn't
+    ;; preclude user code operating on the old slots + new layout or v.v.
+    ;; Users need to synchronize their own access when changing class.
+    (cond ((functionp instance)
+           (rotatef (%fun-layout instance) (%fun-layout copy))
+           (rotatef (%fsc-instance-slots instance) (%fsc-instance-slots copy)))
+          (t
+           (rotatef (%instance-layout instance) (%instance-layout copy))
+           (rotatef (%std-instance-slots instance) (%std-instance-slots copy))))
     (apply #'update-instance-for-different-class copy instance initargs)
-
     instance))
 
 (defun check-new-class-not-metaobject (new-class)
@@ -1676,11 +1738,23 @@
       (check-metaobject method)
       (check-metaobject slot-definition))))
 
+;;; "The first argument to update-instance-for-different-class, /previous/,
+;;; is that copy; it holds the old slot values temporarily. This argument has
+;;; dynamic extent within change-class; if it is referenced in any way once
+;;; update-instance-for-different-class returns, the results are undefined."
+;;; The full ALLOCATE-INSTANCE protocol can not possibly support dynamic-extent
+;;; allocation (at least, for SBCL; maybe for others it can).
+;;; Calling ALLOCATE-INSTANCE from CHANGE-CLASS doesn't seem to be mandatory.
+;;; At least 4 other Lisp implementations I tested don't call it.
+(macrolet ((with-temporary-instance ((var) &body body)
+             `(dx-let ((,var (%make-instance (1+ sb-vm:instance-data-start))))
+                ,@body)))
 (defmethod change-class ((instance standard-object) (new-class standard-class)
                          &rest initargs)
   (with-world-lock ()
     (check-new-class-not-metaobject new-class)
-    (%change-class instance new-class initargs)))
+    (with-temporary-instance (temp)
+      (%change-class temp instance new-class initargs))))
 
 (defmethod change-class ((instance forward-referenced-class)
                          (new-class standard-class) &rest initargs)
@@ -1693,7 +1767,8 @@
                 (:amop :initialization class))))
       (when (eq class (find-class 'class))
         (return nil)))
-    (%change-class instance new-class initargs)))
+    (with-temporary-instance (temp)
+      (%change-class temp instance new-class initargs)))))
 
 (defmethod change-class ((instance t)
                          (new-class forward-referenced-class) &rest initargs)
@@ -1703,12 +1778,21 @@
    '((:amop :generic-function ensure-class-using-class)
      (:amop :initialization class))))
 
+(macrolet ((with-temporary-funinstance ((var) &body body)
+             `(dx-let ((,var (%make-funcallable-instance
+                              (+ sb-vm:instance-data-start 2))))
+                (dx-flet ((signal-error ()
+                            (declare (optimize (sb-c::verify-arg-count 0)))
+                            (error-no-implementation-function ,var)))
+                  (setf (%funcallable-instance-fun ,var) #'signal-error)
+                  ,@body))))
 (defmethod change-class ((instance funcallable-standard-object)
                          (new-class funcallable-standard-class)
                          &rest initargs)
   (with-world-lock ()
     (check-new-class-not-metaobject new-class)
-    (%change-class instance new-class initargs)))
+    (with-temporary-funinstance (temp)
+      (%change-class temp instance new-class initargs)))))
 
 (defmethod change-class ((instance standard-object)
                          (new-class funcallable-standard-class)
@@ -1746,8 +1830,7 @@
              `(defmethod ,name ,args
                 (declare (ignore initargs))
                 (error 'metaobject-initialization-violation
-                       :format-control ,(coerce (format nil "~@<~A~@:>" control)
-                                                'base-string)
+                       :format-control ,(format nil "~~@<~A~~@:>" control)
                        :format-arguments (list (class-name class))
                        :references '((:amop :initialization "Class"))))))
   (def initialize-instance ((class system-class) &rest initargs)
