@@ -32,81 +32,60 @@
 #include "genesis/static-symbols.h"
 #include "genesis/primitive-objects.h"
 #include "thread.h"
-#include "gc-internal.h"
+#include "gc.h"
 #include "code.h"
 #include "var-io.h"
+#include "gc-internal.h"
+#include "forwarding-ptr.h"
+#include "lispstring.h"
 
 #ifdef LISP_FEATURE_OS_PROVIDES_DLADDR
 # include <dlfcn.h>
 #endif
 
-static void
-sbcl_putwc(wchar_t c, FILE *file)
-{
-#ifdef LISP_FEATURE_OS_PROVIDES_PUTWC
-    putwc(c, file);
-#else
-    if (c < 256) {
-        fputc(c, file);
-    } else {
-        fputc('?', file);
-    }
-#endif
-}
-
-unsigned int decode_elsewhere_pc(lispobj packed_integer)
+static int decode_locs(lispobj packed_integer, int *offset, int *elsewhere)
 {
     struct varint_unpacker unpacker;
-    int value;
     varint_unpacker_init(&unpacker, packed_integer);
-    varint_unpack(&unpacker, &value);
-    return value;
+    return varint_unpack(&unpacker, offset) && varint_unpack(&unpacker, elsewhere);
 }
 
 struct compiled_debug_fun *
 debug_function_from_pc (struct code* code, void *pc)
 {
-    uword_t offset = (char*)pc - code_text_start(code);
-    struct compiled_debug_fun *df;
     struct compiled_debug_info *di;
-    struct vector *v;
-    int i, len;
 
-    if (!instancep(code->debug_info))
+    if (instancep(code->debug_info))
+        di = (void*)native_pointer(code->debug_info);
+    else if (listp(code->debug_info) && instancep(CONS(code->debug_info)->car))
+        di = (void*)native_pointer(CONS(code->debug_info)->car);
+    else
         return NULL;
 
-    di = (struct compiled_debug_info *) native_pointer(code->debug_info);
-
     if (!instancep(di->fun_map))
         return NULL;
 
-    v = VECTOR(di->fun_map);
-
-    len = fixnum_value(v->length);
-
-    if (!instancep(v->data[0]))
+    struct compiled_debug_fun *df = (struct compiled_debug_fun*)native_pointer(di->fun_map);
+    int begin, end, elsewhere_begin, elsewhere_end;
+    if (!decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
         return NULL;
-
-    df = (struct compiled_debug_fun *) native_pointer(v->data[0]);
-
-    if (len == 1)
-        return df;
-
-    for (i = 1;; i += 2) {
-        unsigned next_pc;
-
-        if (i == len)
-            return ((struct compiled_debug_fun *) native_pointer(v->data[i - 1]));
-
-        if (offset >= (uword_t)decode_elsewhere_pc(df->encoded_locs)) {
-            struct compiled_debug_fun *p
-                = ((struct compiled_debug_fun *) native_pointer(v->data[i + 1]));
-            next_pc = decode_elsewhere_pc(p->encoded_locs);
-        } else
-            next_pc = fixnum_value(v->data[i]);
-
-        if (offset < next_pc)
-            return ((struct compiled_debug_fun *) native_pointer(v->data[i - 1]));
+    sword_t offset = (char*)pc - code_text_start(code);
+    while (df) {
+        struct compiled_debug_fun *next;
+        if (df->next != NIL) {
+            next = (struct compiled_debug_fun*) native_pointer(df->next);
+            if (!decode_locs(next->encoded_locs, &end, &elsewhere_end))
+                return NULL;
+        } else {
+            next = 0;
+            end = elsewhere_end = code_text_size(code);
+        }
+        if ((begin <= offset && offset < end) ||
+            (elsewhere_begin <= offset && offset < elsewhere_end))
+            return df;
+        begin = end;
+        elsewhere_begin = elsewhere_end;
+        df = next;
     }
 
     return NULL;
@@ -115,34 +94,31 @@ debug_function_from_pc (struct code* code, void *pc)
 static void
 print_string (struct vector *vector, FILE *f)
 {
-  int tag = widetag_of(&vector->header);
+    if (!string_widetag_p(widetag_of(&vector->header))) {
+        fprintf(f, "<??? type %d>", widetag_of(&vector->header));
+        return;
+    }
+    int i;
+    int n = fixnum_value(vector->length);
+    for (i = 0; i < n; i++) {
+        unsigned int c = schar(vector, i);
+        if (c > 0xFFFF) fprintf(f,"\\U%08x", c);
+        // without knowing whether the terminal can accept
+        // character codes 128 through 255, it's conservative
+        // to just output unicode escapes.
+        else if (c > 0x7F) fprintf(f,"\\u%04x", c);
+        else {
+            if (c == '\\' || c == '"') putc('\\', f);
+            putc(c, f);
+        }
+    }
+}
 
-#define doit(TYPE)                              \
-  do {                                          \
-    int i;                                      \
-    int n = fixnum_value(vector->length);       \
-    TYPE *data = (TYPE *) vector->data;         \
-    for (i = 0; i < n; i++) {                   \
-      wchar_t c = (wchar_t) data[i];            \
-      if (c == '\\' || c == '"')                \
-        putc('\\', f);                          \
-      sbcl_putwc(c, f);                         \
-    }                                           \
-  } while (0)
-
-  switch (tag) {
-  case SIMPLE_BASE_STRING_WIDETAG:
-    doit(unsigned char);
-    break;
-#ifdef SIMPLE_CHARACTER_STRING_WIDETAG
-  case SIMPLE_CHARACTER_STRING_WIDETAG:
-    doit(unsigned int);
-    break;
-#endif
-  default:
-    fprintf(f, "<??? type %d>", tag);
-  }
-#undef doit
+lispobj debug_print(lispobj string)
+{
+    print_string(VECTOR(string), stderr);
+    putc('\n', stderr);
+    return 0;
 }
 
 static int string_equal (struct vector *vector, char *string)
@@ -155,6 +131,7 @@ static int string_equal (struct vector *vector, char *string)
 static void
 print_entry_name (lispobj name, FILE *f)
 {
+    name = follow_maybe_fp(name);
     if (listp(name)) {
         putc('(', f);
         while (name != NIL) {
@@ -164,19 +141,20 @@ print_entry_name (lispobj name, FILE *f)
                 return;
             }
             print_entry_name(CONS(name)->car, f);
-            name = CONS(name)->cdr;
+            name = follow_maybe_fp(CONS(name)->cdr);
             if (name != NIL)
                 putc(' ', f);
         }
         putc(')', f);
     } else if (lowtag_of(name) == OTHER_POINTER_LOWTAG) {
-        lispobj *object = native_pointer(name);
-        if (widetag_of(object) == SYMBOL_WIDETAG) {
-            struct symbol *symbol = (struct symbol *) object;
+        struct symbol *symbol = SYMBOL(name);
+        int widetag = header_widetag(symbol->header);
+        switch (widetag) {
+        case SYMBOL_WIDETAG:
             if (symbol->package != NIL) {
                 struct package *pkg
-                    = (struct package *) native_pointer(symbol->package);
-                struct vector *pkg_name = VECTOR(pkg->_name);
+                    = (struct package *) native_pointer(follow_maybe_fp(symbol->package));
+                struct vector *pkg_name = VECTOR(follow_maybe_fp(pkg->_name));
                 if (string_equal(pkg_name, "COMMON-LISP"))
                     ;
                 else if (string_equal(pkg_name, "COMMON-LISP-USER")) {
@@ -189,18 +167,21 @@ print_entry_name (lispobj name, FILE *f)
                     fputs("::", f);
                 }
             }
-            print_string(VECTOR(symbol->name), f);
-        } else if (widetag_of(object) == SIMPLE_BASE_STRING_WIDETAG
+            print_string(VECTOR(follow_maybe_fp(symbol->name)), f);
+            break;
+        case SIMPLE_BASE_STRING_WIDETAG:
 #ifdef SIMPLE_CHARACTER_STRING_WIDETAG
-                   || widetag_of(object) == SIMPLE_CHARACTER_STRING_WIDETAG
+        case SIMPLE_CHARACTER_STRING_WIDETAG:
 #endif
-            ) {
             putc('"', f);
-            print_string((struct vector*)object, f);
+            print_string((struct vector*)symbol, f);
             putc('"', f);
-        } else {
-            fprintf(f, "<??? type %d>", widetag_of(object));
+            break;
+        default:
+            fprintf(f, "<??? type %d>", widetag);
         }
+    } else if (fixnump(name)) {
+        fprintf(f, "%d", (int)fixnum_value(name));
     } else {
         fprintf(f, "<??? lowtag %d>", (int) lowtag_of(name));
     }
@@ -215,7 +196,7 @@ print_entry_points (struct code *code, FILE *f)
             fprintf(f, "%p: bogus function entry", fun);
             return;
         }
-        print_entry_name(fun->name, f);
+        print_entry_name(code->constants[CODE_SLOTS_PER_SIMPLE_FUN*index], f);
         if ((index + 1) < n_funs) fprintf(f, ", ");
     });
 }
@@ -253,7 +234,9 @@ struct call_info {
     int pc; /* Note: this is the trace file offset, not the actual pc. */
 };
 
-#define HEADER_LENGTH(header) ((header)>>8)
+// simple-fun headers have a pointer to layout-of-function in the
+// upper bytes if words are 8 bytes, so mask off those bytes.
+#define HEADER_LENGTH(header) (((header)>>8) & FUN_HEADER_NWORDS_MASK)
 
 static int previous_info(struct call_info *info);
 
@@ -373,19 +356,14 @@ previous_info(struct call_info *info)
         }
     } else if (fixnump(lra)) {
         info->code = (struct code*)native_pointer(this_frame->code);
+        // FIXME: is this right? fixnumish LRAs are based off the object base address
+        // and not the code text start?
         info->pc = (uword_t)(info->code + lra);
         info->lra = NIL;
     } else {
         info->code = code_pointer(lra);
-
         if (info->code != NULL)
-            info->pc = (uword_t)native_pointer(info->lra) -
-                (uword_t)info->code -
-#ifndef LISP_FEATURE_ALPHA
-                (HEADER_LENGTH(info->code->header) * sizeof(lispobj));
-#else
-        (HEADER_LENGTH(((struct code *)info->code)->header) * sizeof(lispobj));
-#endif
+            info->pc = (char*)native_pointer(info->lra) - code_text_start(info->code);
         else
             info->pc = 0;
     }
@@ -424,7 +402,7 @@ lisp_backtrace(int nframes)
             printf(" <no LRA>");
 
         if (info.pc)
-            printf(" pc = %p", (void*)(long)info.pc);
+            printf(" pc_ofs = %p", (void*)(long)info.pc);
         putchar('\n');
 
     } while (i++ < nframes && previous_info(&info));
@@ -479,7 +457,7 @@ ra_pointer_p (void *ra)
   return ((uword_t) ra) > 4096 && !stack_pointer_p (ra);
 }
 
-static int
+static int NO_SANITIZE_MEMORY
 x86_call_context (void *fp, void **ra, void **ocfp)
 {
   void *c_ocfp;
@@ -530,11 +508,11 @@ describe_thread_state(void)
     printf("Pending handler = %p\n", data->pending_handler);
 }
 
-static void print_backtrace_frame(void *pc, void *fp, int i, FILE *f) {
+static void print_backtrace_frame(char *pc, void *fp, int i, FILE *f) {
     lispobj *p;
     fprintf(f, "%4d: ", i);
 
-    p = component_ptr_from_pc((lispobj *) pc);
+    p = component_ptr_from_pc(pc);
 
     if (p) {
         struct code *cp = (struct code *) p;
@@ -561,7 +539,7 @@ static void print_backtrace_frame(void *pc, void *fp, int i, FILE *f) {
  * backtraces from gdb with call backtrace_from_fp(...). Useful for
  * example when debugging threading deadlocks.
  */
-void
+void NO_SANITIZE_MEMORY
 log_backtrace_from_fp(void *fp, int nframes, int start, FILE *f)
 {
   int i = start;
@@ -611,3 +589,133 @@ lisp_backtrace(int nframes)
     }
 }
 #endif
+
+// Find the simple_fun that contains 'pc' in 'code'
+int simple_fun_index_from_pc(struct code* code, char *pc)
+{
+    char *instruction_area = code_text_start(code);
+    unsigned int* offsets = code_fun_table(code) - 1;
+    int index;
+    for (index = code_n_funs(code) - 1; index >= 0; --index) {
+        char *base = instruction_area + offsets[-index];
+        if (pc >= base) return index;
+    }
+    return -1;
+}
+
+static boolean __attribute__((unused)) print_lisp_fun_name(char* pc)
+{
+  struct code* code;
+  if (gc_managed_addr_p((uword_t)pc) &&
+      (code = (void*)component_ptr_from_pc(pc)) != 0) {
+      struct compiled_debug_fun* df = debug_function_from_pc(code, pc);
+      if (df) {
+          fprintf(stderr, " %p [", pc);
+          print_entry_name(df->name, stderr);
+          fprintf(stderr, "]\n");
+          return 1;
+      }
+  }
+  return 0;
+}
+
+#ifdef LISP_FEATURE_LIBUNWIND_BACKTRACE
+#define UNW_LOCAL_ONLY
+#include <libunwind.h>
+#include "genesis/thread-instance.h"
+#include "genesis/mutex.h"
+static __attribute__((unused))int backtrace_completion_pipe[2] = {-1,-1};
+void libunwind_backtrace(struct thread *th, os_context_t *context)
+{
+    char procname[100];
+    fprintf(stderr, "Lisp thread @ %p, tid %d", th, (int)th->os_kernel_tid);
+#ifdef LISP_FEATURE_SB_THREAD
+    // the TLS area is not used if #-sb-thread. And if so, it must be "main thred"
+    struct thread_instance* lispthread = (void*)native_pointer(th->lisp_thread);
+    if (lispthread->name != NIL) {
+        fprintf(stderr, " (\"");
+        print_string(VECTOR(lispthread->name), stderr);
+        fprintf(stderr, "\")");
+    }
+    putc('\n', stderr);
+    if (lispthread->waiting_for != NIL) {
+        fprintf(stderr, "waiting for %p", (void*)lispthread->waiting_for);
+        if (instancep(lispthread->waiting_for)) {
+            // THREAD-WAITING-FOR can be a mutex or a waitqueue (if not a cons).
+            // Accessing it as if it's a mutex works because both a waitqueue
+            // and a mutex have a name at the same slot offset (if #+sb-futex).
+            // So to reiterate the comment from linux-os.c -
+            // "Use this only if you know what you're doing"
+            struct mutex* lispmutex = (void*)native_pointer(lispthread->waiting_for);
+            if (lispmutex->name != NIL) {
+                fprintf(stderr, " (MUTEX:\"");
+                print_string(VECTOR(lispmutex->name), stderr);
+                fprintf(stderr, "\")");
+            }
+        }
+        putc('\n', stderr);
+    }
+#endif
+    unw_cursor_t cursor;
+    // "unw_init_local() is thread-safe as well as safe to use from a signal handler."
+    // "unw_get_proc_name() is thread-safe. If cursor cp is in the local address-space,
+    //  this routine is also safe to use from a signal handler."
+    if (context) {
+        unw_init_local(&cursor, context);
+    } else {
+        unw_context_t here;
+        unw_getcontext(&here);
+        unw_init_local(&cursor, &here);
+    }
+    do {
+        uword_t offset;
+        char *pc;
+        unw_get_reg(&cursor, UNW_TDEP_IP, (uword_t*)&pc);
+        if (print_lisp_fun_name(pc)) {
+            // printed
+        } else if (!unw_get_proc_name(&cursor, procname, sizeof procname, &offset)) {
+            fprintf(stderr, " %p [%s]\n", pc, procname);
+        } else {
+            fprintf(stderr, " %p ?\n", pc);
+        }
+    } while (unw_step(&cursor));
+}
+void backtrace_lisp_threads(int __attribute__((unused)) signal,
+                                   siginfo_t __attribute__((unused)) *info,
+                                   os_context_t *context)
+{
+#ifdef LISP_FEATURE_SB_THREAD
+    if (backtrace_completion_pipe[1] >= 0) {
+        libunwind_backtrace(current_thread, context);
+        write(backtrace_completion_pipe[1], context /* any random byte */, 1);
+        return;
+    }
+    struct thread *th;
+    int nthreads = 0;
+    for_each_thread(th) { ++nthreads; }
+    fprintf(stderr, "Caught backtrace-all signal in tid %d, %d threads\n",
+            (int)arch_os_get_current_thread()->os_kernel_tid, nthreads);
+    // Would be nice if we could forcibly stop all the other threads,
+    // but pthread_mutex_trylock is not safe to use in a signal handler.
+    if (nthreads > 1) {
+        pipe(backtrace_completion_pipe);
+    }
+    for_each_thread(th) {
+        if (th == arch_os_get_current_thread())
+            libunwind_backtrace(th, context);
+        else {
+            char junk;
+            pthread_kill(th->os_thread, SIGXCPU);
+            read(backtrace_completion_pipe[0], &junk, 1);
+        }
+    }
+    if (nthreads > 1) {
+        close(backtrace_completion_pipe[1]);
+        close(backtrace_completion_pipe[0]);
+        backtrace_completion_pipe[0] = backtrace_completion_pipe[1] = -1;
+    }
+#else
+    libunwind_backtrace(arch_os_get_current_thread(), context);
+#endif
+}
+#endif
