@@ -35,8 +35,6 @@
 #include "genesis/static-symbols.h"
 #include "genesis/fdefn.h"
 
-#include <sys/socket.h>
-#include <sys/utsname.h>
 #include <errno.h>
 
 #include <sys/types.h>
@@ -64,13 +62,98 @@ int personality (unsigned long);
 #include <sys/personality.h>
 #endif
 
-size_t os_vm_page_size;
-
 #if defined(LISP_FEATURE_SB_THREAD) && defined(LISP_FEATURE_SB_FUTEX) && !defined(LISP_FEATURE_SB_PTHREAD_FUTEX)
 #include <sys/syscall.h>
 #include <unistd.h>
 #include <errno.h>
 
+#ifdef MUTEX_EVENTRECORDING
+#include "genesis/mutex.h"
+#define MAXEVENTS 200
+static struct {
+    struct thread* th;
+    struct timespec ts;
+    char *label;
+    char *mutex_name;
+    sword_t timeout;
+} events[MAXEVENTS];
+static int record_mutex_events;
+static int eventcount;
+
+void lisp_mutex_event(char *string) {
+    if (record_mutex_events) {
+        int id = __sync_fetch_and_add(&eventcount, 1);
+        if (id >= MAXEVENTS) lose("event buffer overflow");
+        clock_gettime(CLOCK_REALTIME, &events[id].ts);
+        events[id].th = current_thread;
+        events[id].label = string;
+        events[id].mutex_name = 0;
+        events[id].timeout = -1;
+    }
+}
+void lisp_mutex_event1(char *string, char *string2) {
+    if (record_mutex_events) {
+        int id = __sync_fetch_and_add(&eventcount, 1);
+        if (id >= MAXEVENTS) lose("event buffer overflow");
+        clock_gettime(CLOCK_REALTIME, &events[id].ts);
+        events[id].th = current_thread;
+        events[id].label = string;
+        events[id].mutex_name = string2;
+        events[id].timeout = -1;
+    }
+}
+void lisp_mutex_event2(char *string, char *string2, uword_t usec) {
+    if (record_mutex_events) {
+        int id = __sync_fetch_and_add(&eventcount, 1);
+        if (id >= MAXEVENTS) lose("event buffer overflow");
+        clock_gettime(CLOCK_REALTIME, &events[id].ts);
+        events[id].th = current_thread;
+        events[id].label = string;
+        events[id].mutex_name = string2;
+        events[id].timeout = usec;
+    }
+}
+void lisp_mutex_start_eventrecording() {
+    eventcount = 0;
+    record_mutex_events = 1;
+}
+void lisp_mutex_done_eventrecording() {
+    record_mutex_events = 0;
+    int i;
+    fprintf(stderr, "event log:\n");
+    struct timespec basetime = events[0].ts;
+    for(i=0; i<eventcount;++i) {
+        struct thread *th = events[i].th;
+        struct thread_instance *ti = (void*)native_pointer(th->lisp_thread);
+        struct timespec rel_time = events[i].ts;
+        rel_time.tv_sec -= basetime.tv_sec;
+        rel_time.tv_nsec -= basetime.tv_nsec;
+        if (rel_time.tv_nsec<0) rel_time.tv_nsec += 1000 * 1000 * 1000, rel_time.tv_sec--;
+        lispobj threadname = ti->name;
+        if (events[i].timeout >= 0) // must also have mutex_name in this case
+            fprintf(stderr, "[%d.%09ld] %s: %s '%s' timeout %ld\n",
+                    (int)rel_time.tv_sec, rel_time.tv_nsec,
+                    (char*)VECTOR(threadname)->data,
+                    events[i].label, events[i].mutex_name, events[i].timeout);
+        else if (events[i].mutex_name)
+            fprintf(stderr, "[%d.%09ld] %s: %s '%s'\n",
+                    (int)rel_time.tv_sec, rel_time.tv_nsec,
+                    (char*)VECTOR(threadname)->data,
+                    events[i].label, events[i].mutex_name);
+        else
+            fprintf(stderr, "[%d.%09ld] %s: %s\n",
+                    (int)rel_time.tv_sec, rel_time.tv_nsec,
+                    (char*)VECTOR(threadname)->data,
+                    events[i].label);
+    }
+    fprintf(stderr, "-----\n");
+}
+#else
+#define lisp_mutex_event(x)
+#define lisp_mutex_event1(x,y)
+#define lisp_mutex_event2(x,y,z)
+#endif
+
 /* values taken from the kernel's linux/futex.h.  This header file
    doesn't exist in userspace, which is our excuse for not grovelling
    them automatically */
@@ -81,8 +164,6 @@ size_t os_vm_page_size;
  * futexes when available.*/
 #define FUTEX_WAIT_PRIVATE (0+128)
 #define FUTEX_WAKE_PRIVATE (1+128)
-#define FUTEX_FD (2)
-#define FUTEX_REQUEUE (3)
 
 /* Not static so that Lisp may query it. */
 boolean futex_private_supported_p;
@@ -128,16 +209,11 @@ futex_init()
 #include "genesis/vector.h"
 char* futex_name(int *lock_word)
 {
-    // If there is a Lisp string at lock_word-1 or -2, return that.
-    // Otherwise return NULL.
-    lispobj name = *(lock_word - 1);
-    struct vector* v = (struct vector*)native_pointer(name);
-    if (lowtag_of(name) == OTHER_POINTER_LOWTAG && widetag_of(&v->header) == SIMPLE_BASE_STRING_WIDETAG)
-        return (char*)(v->data);
-    name = *(lock_word - 2);
-    v = (struct vector*)native_pointer(name);
-    if (lowtag_of(name) == OTHER_POINTER_LOWTAG && widetag_of(&v->header) == SIMPLE_BASE_STRING_WIDETAG)
-        return (char*)(v->data);
+    // If there is a Lisp string at lock_word+1, return that, otherwise NULL.
+    lispobj name = ((lispobj*)lock_word)[1];
+    if (lowtag_of(name) == OTHER_POINTER_LOWTAG &&
+        header_widetag(VECTOR(name)->header) == SIMPLE_BASE_STRING_WIDETAG)
+        return (char*)VECTOR(name)->data;
     return 0;
 }
 
@@ -147,14 +223,21 @@ futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
   struct timespec timeout;
   int t;
 
+#ifdef MUTEX_EVENTRECORDING
+    struct mutex* m = (void*)((char*)lock_word - offsetof(struct mutex,state));
+    char *name = m->name != NIL ? (char*)VECTOR(m->name)->data : "(unnamed)";
+#endif
   if (sec<0) {
+      lisp_mutex_event1("start futex wait", name);
       t = sys_futex(lock_word, futex_wait_op(), oldval, 0);
   }
   else {
       timeout.tv_sec = sec;
       timeout.tv_nsec = usec * 1000;
+      lisp_mutex_event2("start futex timedwait", name, usec);
       t = sys_futex(lock_word, futex_wait_op(), oldval, &timeout);
   }
+  lisp_mutex_event1("back from sys_futex", name);
   if (t==0)
       return 0;
   else if (errno==ETIMEDOUT)
@@ -169,67 +252,22 @@ futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
 int
 futex_wake(int *lock_word, int n)
 {
+#ifdef MUTEX_EVENTRECORDING
+    struct mutex* m = (void*)((char*)lock_word - offsetof(struct mutex,state));
+    char *name = m->name != NIL ? (char*)VECTOR(m->name)->data : "(unnamed)";
+    lisp_mutex_event1("waking futex", name);
+#endif
     return sys_futex(lock_word, futex_wake_op(),n,0);
 }
 #endif
 
 
-
-#ifdef LISP_FEATURE_SB_THREAD
-int
-isnptl (void)
-{
-  size_t n = confstr (_CS_GNU_LIBPTHREAD_VERSION, NULL, 0);
-  if (n > 0) {
-      char *buf = alloca (n);
-      confstr (_CS_GNU_LIBPTHREAD_VERSION, buf, n);
-      if (strstr (buf, "NPTL")) {
-          return 1;
-      }
-  }
-  return 0;
-}
-#endif
-
-static void getuname(int *major_version, int* minor_version, int *patch_version)
-{
-    /* Conduct various version checks: do we have enough mmap(), is
-     * this a sparc running 2.2, can we do threads? */
-    struct utsname name;
-    char *p;
-    uname(&name);
-
-    p=name.release;
-    *major_version = atoi(p);
-    *minor_version = *patch_version = 0;
-    p=strchr(p,'.');
-    if (p != NULL) {
-            *minor_version = atoi(++p);
-            p=strchr(p,'.');
-            if (p != NULL)
-                    *patch_version = atoi(++p);
-    }
-}
-
 void os_init(char __attribute__((unused)) *argv[],
              char __attribute__((unused)) *envp[])
 {
-#ifdef LISP_FEATURE_SB_THREAD
-#if defined(LISP_FEATURE_SB_FUTEX) && !defined(LISP_FEATURE_SB_PTHREAD_FUTEX)
+#if defined(LISP_FEATURE_SB_THREAD) && defined(LISP_FEATURE_SB_FUTEX) && !defined(LISP_FEATURE_SB_PTHREAD_FUTEX)
     futex_init();
 #endif
-    if(! isnptl()) {
-       lose("This version of SBCL only works correctly with the NPTL threading\n"
-            "library. Please use a newer glibc, use an older SBCL, or stop using\n"
-            "LD_ASSUME_KERNEL");
-    }
-#endif
-
-    /* Don't use getpagesize(), since it's not constant across Linux
-     * kernel versions on some architectures (for example PPC). FIXME:
-     * possibly the same should be done on other architectures too.
-     */
-    os_vm_page_size = BACKEND_PAGE_BYTES;
 
 #ifdef LISP_FEATURE_X86
     /* Use SSE detector.  Recent versions of Linux enable SSE support
@@ -248,19 +286,6 @@ void os_init(char __attribute__((unused)) *argv[],
 
 int os_preinit(char *argv[], char *envp[])
 {
-    int major_version, minor_version, patch_version;
-    getuname(&major_version, &minor_version, &patch_version);
-    if (major_version<2) {
-        lose("linux kernel version too old: major version=%d (can't run in version < 2.0.0)",
-             major_version);
-    }
-    // Rev b44ca02cb963 conditionally enabled a workaround for a kernel 2.6 bug.
-    // Rev d4d6c4b16a36 moved the workaround check from compile-time to runtime.
-    // 14 years later we're still performing that check, which annoyingly adds
-    // signal-related junk to startup which must be ignored when debugging.
-    // Let's assume that major version 3 and later are fine.
-    extern void set_sigaction_nodefer_works();
-    if (major_version>=3) set_sigaction_nodefer_works();
 
 #if ALLOW_PERSONALITY_CHANGE
     if (getenv("SBCL_IS_RESTARTING")) {
@@ -279,53 +304,43 @@ int os_preinit(char *argv[], char *envp[])
         return 1; // indicate that we already allocated hardwired spaces
 
 #if ALLOW_PERSONALITY_CHANGE
-    /* KLUDGE: Disable memory randomization on new Linux kernels
-     * by setting a personality flag and re-executing. (We need
-     * to re-execute, since the memory maps that can conflict with
-     * the SBCL spaces have already been done at this point).
+    /* KLUDGE: Disable memory randomization by setting a personality
+     * flag and re-executing. (We need to re-execute, since the memory
+     * maps that can conflict with the SBCL spaces have already been
+     * done at this point).
      */
-    if ((major_version == 2
-         /* Some old kernels will apparently lose unsupported personality flags
-          * on exec() */
-         && ((minor_version == 6 && patch_version >= 11)
-             || (minor_version > 6)
-             /* This is what RHEL 3 reports */
-             || (minor_version == 4 && patch_version > 20)))
-        || major_version >= 3)
-    {
-        int pers = personality(0xffffffffUL);
-        if (!(pers & ADDR_NO_RANDOMIZE)) {
-            int retval = personality(pers | ADDR_NO_RANDOMIZE);
-            /* Allegedly some Linux kernels (the reported case was
-             * "hardened Linux 2.6.7") won't set the new personality,
-             * but nor will they return -1 for an error. So as a
-             * workaround query the new personality...
-             */
-            int newpers = personality(0xffffffffUL);
-            /* ... and don't re-execute if either the setting resulted
-             * in an error or if the value didn't change. Otherwise
-             * this might result in an infinite loop.
-             */
+    int pers = personality(0xffffffffUL);
+    if (!(pers & ADDR_NO_RANDOMIZE)) {
+        int retval = personality(pers | ADDR_NO_RANDOMIZE);
+        /* Allegedly some Linux kernels (the reported case was
+         * "hardened Linux 2.6.7") won't set the new personality,
+         * but nor will they return -1 for an error. So as a
+         * workaround query the new personality...
+         */
+        int newpers = personality(0xffffffffUL);
+        /* ... and don't re-execute if either the setting resulted
+         * in an error or if the value didn't change. Otherwise
+         * this might result in an infinite loop.
+         */
 
-            if (!getenv("SBCL_IS_RESTARTING") &&
-                retval != -1 && newpers != pers) {
-                /* Use /proc/self/exe instead of trying to figure out
-                 * the executable path from PATH and argv[0], since
-                 * that's unreliable. We follow the symlink instead of
-                 * executing the file directly in order to prevent top
-                 * from displaying the name of the process as "exe". */
-                char runtime[PATH_MAX+1];
-                int i = readlink("/proc/self/exe", runtime, PATH_MAX);
-                if (i != -1) {
-                    environ = envp;
-                    setenv("SBCL_IS_RESTARTING", "T", 1);
-                    runtime[i] = '\0';
-                    execv(runtime, argv);
-                }
+        if (!getenv("SBCL_IS_RESTARTING") &&
+            retval != -1 && newpers != pers) {
+            /* Use /proc/self/exe instead of trying to figure out
+             * the executable path from PATH and argv[0], since
+             * that's unreliable. We follow the symlink instead of
+             * executing the file directly in order to prevent top
+             * from displaying the name of the process as "exe". */
+            char runtime[PATH_MAX+1];
+            int i = readlink("/proc/self/exe", runtime, PATH_MAX);
+            if (i != -1) {
+                environ = envp;
+                setenv("SBCL_IS_RESTARTING", "T", 1);
+                runtime[i] = '\0';
+                execv(runtime, argv);
             }
-            /* Either changing the personality or execve() failed.
-             * Just get on with life and hope for the best. */
         }
+        /* Either changing the personality or execve() failed.
+         * Just get on with life and hope for the best. */
     }
 #endif
     return 0;
@@ -403,8 +418,7 @@ void
 os_install_interrupt_handlers(void)
 {
     if (INSTALL_SIG_MEMORY_FAULT_HANDLER) {
-    undoably_install_low_level_interrupt_handler(SIG_MEMORY_FAULT,
-                                                 sigsegv_handler);
+    ll_install_handler(SIG_MEMORY_FAULT, sigsegv_handler);
     }
 
     /* OAOOM c.f. sunos-os.c.
@@ -412,17 +426,15 @@ os_install_interrupt_handlers(void)
 #ifdef LISP_FEATURE_SB_THREAD
 # ifdef LISP_FEATURE_SB_SAFEPOINT
 #  ifdef LISP_FEATURE_SB_THRUPTION
-    undoably_install_low_level_interrupt_handler(SIGPIPE, thruption_handler);
+    ll_install_handler(SIGPIPE, thruption_handler);
 #  endif
 # else
-    undoably_install_low_level_interrupt_handler(SIG_STOP_FOR_GC,
-                                                 sig_stop_for_gc_handler);
+    ll_install_handler(SIG_STOP_FOR_GC, sig_stop_for_gc_handler);
 # endif
 #endif
 }
 
-char *
-os_get_runtime_executable_path(int __attribute__((unused)) external)
+char *os_get_runtime_executable_path()
 {
     char path[PATH_MAX + 1];
     int size;
