@@ -37,6 +37,8 @@
 #include "genesis/layout.h"
 #include "genesis/defstruct-description.h"
 #include "genesis/hash-table.h"
+#include "code.h"
+#include "getallocptr.h"
 
 /* We don't ever do purification with GENCGC as of 1.0.5.*. There was
  * a lot of hairy and fragile ifdeffage in here to support purify on
@@ -62,15 +64,13 @@ static lispobj *read_only_free, *static_free;
 static lispobj *pscav(lispobj *addr, long nwords, boolean constant);
 
 #define LATERBLOCKSIZE 1020
-#define LATERMAXCOUNT 10
 
 static struct
 later {
     struct later *next;
-    union {
-        lispobj *ptr;
-        long count;
-    } u[LATERBLOCKSIZE];
+    // slightly denser packing vs a struct of a lispobj* and an int
+    lispobj *ptr[LATERBLOCKSIZE];
+    int count[LATERBLOCKSIZE];
 } *later_blocks = NULL;
 static long later_count = 0;
 
@@ -99,13 +99,13 @@ newspace_alloc(long nwords, int constantp)
     gc_assert((nwords & 1) == 0);
     if(constantp) {
         if(read_only_free + nwords >= (lispobj *)READ_ONLY_SPACE_END) {
-            lose("Ran out of read-only space while purifying!\n");
+            lose("Ran out of read-only space while purifying!");
         }
         ret=read_only_free;
         read_only_free+=nwords;
     } else {
         if(static_free + nwords >= (lispobj *)STATIC_SPACE_END) {
-            lose("Ran out of static space while purifying!\n");
+            lose("Ran out of static space while purifying!");
         }
         ret=static_free;
         static_free+=nwords;
@@ -114,33 +114,22 @@ newspace_alloc(long nwords, int constantp)
 }
 
 
+/* Enqueue <where,count> into later_blocks */
 static void
 pscav_later(lispobj *where, long count)
 {
     struct later *new;
 
-    if (count > LATERMAXCOUNT) {
-        while (count > LATERMAXCOUNT) {
-            pscav_later(where, LATERMAXCOUNT);
-            count -= LATERMAXCOUNT;
-            where += LATERMAXCOUNT;
-        }
+    if (later_blocks == NULL || later_count == LATERBLOCKSIZE) {
+        new  = (struct later *)malloc(sizeof(struct later));
+        new->next = later_blocks;
+        later_blocks = new;
+        later_count = 0;
     }
-    else {
-        if (later_blocks == NULL || later_count == LATERBLOCKSIZE ||
-            (later_count == LATERBLOCKSIZE-1 && count > 1)) {
-            new  = (struct later *)malloc(sizeof(struct later));
-            new->next = later_blocks;
-            if (later_blocks && later_count < LATERBLOCKSIZE)
-                later_blocks->u[later_count].ptr = NULL;
-            later_blocks = new;
-            later_count = 0;
-        }
 
-        if (count != 1)
-            later_blocks->u[later_count++].count = count;
-        later_blocks->u[later_count++].ptr = where;
-    }
+    later_blocks->count[later_count] = count;
+    later_blocks->ptr[later_count] = where;
+    ++later_count;
 }
 
 static lispobj
@@ -148,11 +137,11 @@ ptrans_boxed(lispobj thing, lispobj header, boolean constant)
 {
     /* Allocate it */
     lispobj *old = native_pointer(thing);
-    long nwords = sizetab[widetag_of(header)](old);
+    long nwords = sizetab[header_widetag(header)](old);
     lispobj *new = newspace_alloc(nwords,constant);
 
     /* Copy it. */
-    bcopy(old, new, nwords * sizeof(lispobj));
+    memcpy(new, old, nwords * sizeof(lispobj));
 
     /* Deposit forwarding pointer. */
     lispobj result = make_lispobj(new, lowtag_of(thing));
@@ -188,11 +177,11 @@ ptrans_fdefn(lispobj thing, lispobj header)
 {
     /* Allocate it */
     lispobj *old = native_pointer(thing);
-    long nwords = sizetab[widetag_of(header)](old);
+    long nwords = sizetab[header_widetag(header)](old);
     lispobj *new = newspace_alloc(nwords, 0);    /* inconstant */
 
     /* Copy it. */
-    bcopy(old, new, nwords * sizeof(lispobj));
+    memcpy(new, old, nwords * sizeof(lispobj));
 
     /* Deposit forwarding pointer. */
     lispobj result = make_lispobj(new, lowtag_of(thing));
@@ -213,11 +202,11 @@ ptrans_unboxed(lispobj thing, lispobj header)
 {
     /* Allocate it */
     lispobj *old = native_pointer(thing);
-    long nwords = sizetab[widetag_of(header)](old);
+    long nwords = sizetab[header_widetag(header)](old);
     lispobj *new = newspace_alloc(nwords, 1);     /* always constant */
 
     /* copy it. */
-    bcopy(old, new, nwords * sizeof(lispobj));
+    memcpy(new, old, nwords * sizeof(lispobj));
 
     /* Deposit forwarding pointer. */
     lispobj result = make_lispobj(new, lowtag_of(thing));
@@ -230,10 +219,10 @@ static lispobj
 ptrans_vector(lispobj thing, boolean boxed, boolean constant)
 {
     struct vector *vector = VECTOR(thing);
-    long nwords = sizetab[widetag_of(vector->header)]((lispobj*)vector);
+    long nwords = sizetab[header_widetag(vector->header)]((lispobj*)vector);
 
     lispobj *new = newspace_alloc(nwords, (constant || !boxed));
-    bcopy(vector, new, nwords * sizeof(lispobj));
+    memcpy(new, vector, nwords * sizeof(lispobj));
 
     lispobj result = make_lispobj(new, lowtag_of(thing));
     vector->header = result;
@@ -248,42 +237,45 @@ static lispobj
 ptrans_code(lispobj thing)
 {
     struct code *code = (struct code *)native_pointer(thing);
-    long nwords =
-        ALIGN_UP(code_header_words(code->header) + code_unboxed_nwords(code->code_size),
-                 2);
+    long nwords = code_total_nwords(code);
 
     struct code *new = (struct code *)newspace_alloc(nwords,1); /* constant */
 
-    bcopy(code, new, nwords * sizeof(lispobj));
+    memcpy(new, code, nwords * sizeof(lispobj));
 
     lispobj result = make_lispobj(new, OTHER_POINTER_LOWTAG);
-
-    /* Put in forwarding pointers for all the functions. */
     uword_t displacement = result - thing;
+
+#if defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64
+    // Fixup absolute jump tables. These aren't recorded in code->fixups
+    // because we don't need to denote an arbitrary set of places in the code.
+    // The count alone suffices. A GC immediately after creating the code
+    // could cause us to observe some 0 words here. Those should be ignored.
+    lispobj* jump_table = code_jumptable_start(new);
+    int count = jumptable_count(jump_table);
+    int i;
+    for (i = 1; i < count; ++i)
+        if (jump_table[i]) jump_table[i] += displacement;
+#endif
+    /* Put in forwarding pointers for all the functions. */
     for_each_simple_fun(i, newfunc, new, 1, {
         lispobj* old = (lispobj*)LOW_WORD((char*)newfunc - displacement);
         *old = make_lispobj(newfunc, FUN_POINTER_LOWTAG);
+        pscav(&newfunc->self, 1, 1); // and fix the self-pointer now
     });
 
+    int n_funs = code_n_funs(code);
     /* Stick in a forwarding pointer for the code object. */
     /* This smashes the header, so do it only after reading n_funs */
     *(lispobj *)code = result;
 
-    /* Arrange to scavenge the debug info later. */
-    pscav_later(&new->debug_info, 1);
-
-    /* Scavenge the constants. */
-    pscav(new->constants,
-          code_header_words(new->header) - (offsetof(struct code, constants) >> WORD_SHIFT),
-          1);
-
-    /* Scavenge all the functions. */
-    for_each_simple_fun(i, func, new, 1, {
-        gc_assert(!dynamic_pointer_p((lispobj)func));
-        pscav(&func->self, 1, 1);
-        pscav_later(&func->name, 4);
-    })
-
+    int n_later_words = 1 + n_funs * CODE_SLOTS_PER_SIMPLE_FUN;
+    /* Scavenge the constants excluding the ones that will be done later. */
+    lispobj* from = &new->debug_info + n_later_words;
+    lispobj* end = (lispobj*)new + code_header_words(new);
+    pscav(from, end - from, 1);
+    /* Arrange to scavenge the debug info and simple-fun metadata later. */
+    pscav_later(&new->debug_info, n_later_words);
     return result;
 }
 
@@ -296,15 +288,14 @@ ptrans_func(lispobj thing, lispobj header)
      * Otherwise we have to do something strange, 'cause it is buried
      * inside a code object. */
 
-    if (widetag_of(header) == SIMPLE_FUN_WIDETAG) {
+    if (header_widetag(header) == SIMPLE_FUN_WIDETAG) {
 
         /* We can only end up here if the code object has not been
          * scavenged, because if it had been scavenged, forwarding pointers
          * would have been left behind for all the entry points. */
 
         struct simple_fun *function = (struct simple_fun *)native_pointer(thing);
-        lispobj code = make_lispobj(native_pointer(thing) - HeaderValue(function->header),
-                                    OTHER_POINTER_LOWTAG);
+        lispobj code = fun_code_tagged((lispobj*)function);
 
         /* This will cause the function's header to be replaced with a
          * forwarding pointer. */
@@ -316,16 +307,16 @@ ptrans_func(lispobj thing, lispobj header)
     } else {
         /* It's some kind of closure-like thing. */
         lispobj *old = native_pointer(thing);
-        long nwords = sizetab[widetag_of(header)](old);
+        long nwords = sizetab[header_widetag(header)](old);
 
         /* Allocate the new one.  FINs *must* not go in read_only
          * space.  Closures can; they never change */
 
         lispobj *new = newspace_alloc
-            (nwords,(widetag_of(header)!=FUNCALLABLE_INSTANCE_WIDETAG));
+            (nwords,(header_widetag(header)!=FUNCALLABLE_INSTANCE_WIDETAG));
 
         /* Copy it. */
-        bcopy(old, new, nwords * sizeof(lispobj));
+        memcpy(new, old, nwords * sizeof(lispobj));
 
         /* Deposit forwarding pointer. */
         lispobj result = make_lispobj(new, lowtag_of(thing));
@@ -391,7 +382,7 @@ ptrans_list(lispobj thing, boolean constant)
 static lispobj
 ptrans_otherptr(lispobj thing, lispobj header, boolean constant)
 {
-    int widetag = widetag_of(header);
+    int widetag = header_widetag(header);
     switch (widetag) {
         /* FIXME: this needs a reindent */
       case BIGNUM_WIDETAG:
@@ -447,7 +438,7 @@ ptrans_otherptr(lispobj thing, lispobj header, boolean constant)
         if (other_immediate_lowtag_p(widetag) &&
             specialized_vector_widetag_p(widetag))
             return ptrans_vector(thing, 0, constant);
-        fprintf(stderr, "Invalid widetag: %d\n", widetag_of(header));
+        fprintf(stderr, "Invalid widetag: %d\n", header_widetag(header));
         /* Should only come across other pointers to the above stuff. */
         gc_abort();
         return NIL;
@@ -475,7 +466,7 @@ pscav(lispobj *addr, long nwords, boolean constant)
 
     while (nwords > 0) {
         thing = *addr;
-        int widetag = widetag_of(thing);
+        int widetag = header_widetag(thing);
         if (is_lisp_pointer(thing)) {
             /* It's a pointer. Is it something we might have to move? */
             if (dynamic_pointer_p(thing)) {
@@ -536,10 +527,10 @@ pscav(lispobj *addr, long nwords, boolean constant)
               case SIMPLE_VECTOR_WIDETAG:
                 // addr[0] : header
                 //     [1] : vector length
-                //     [2] : element[0] = hashtable
+                //     [2] : element[0] = high-water mark
                 //     [3] : element[1] = rehash bit
-                if (is_vector_subtype(thing, VectorValidHashing))
-                    addr[3] = make_fixnum(1);
+                if (is_vector_subtype(thing, VectorAddrHashing))
+                    addr[3] = make_fixnum(1); // just flag it for rehash
                 count = 2;
                 break;
 
@@ -592,7 +583,7 @@ pscav(lispobj *addr, long nwords, boolean constant)
               default:
                 if (other_immediate_lowtag_p(widetag) &&
                     specialized_vector_widetag_p(widetag))
-                    count = sizetab[widetag_of(thing)](addr);
+                    count = sizetab[header_widetag(thing)](addr);
                 else
                     count = 1;
                 break;
@@ -610,12 +601,30 @@ pscav(lispobj *addr, long nwords, boolean constant)
     return addr;
 }
 
+static void
+verify_range(lispobj *base, lispobj *end)
+{
+    lispobj* where = base;
+    while (where < end) {
+        if (widetag_of(where) == CODE_HEADER_WIDETAG) {
+            int n_boxed_words = code_header_words((struct code*)where);
+            int i;
+            for (i = 0; i < n_boxed_words; ++i) {
+                lispobj word = where[i];
+                if (is_lisp_pointer(word) && dynamic_pointer_p(word))
+                    lose("purify failed, obj=%p, where=%p, val=%"OBJ_FMTX,
+                         where, where+i, word);
+            }
+        }
+        where += OBJECT_SIZE(*where, where);
+    }
+}
+
 int
 purify(lispobj static_roots, lispobj read_only_roots)
 {
     lispobj *clean;
     long count, i;
-    struct later *laters, *next;
     struct thread *thread;
 
     if(all_threads->next) {
@@ -641,7 +650,7 @@ purify(lispobj static_roots, lispobj read_only_roots)
         return 0;
     }
 
-    dynamic_space_purify_pointer = dynamic_space_free_pointer;
+    dynamic_space_purify_pointer = get_alloc_pointer();
 
     read_only_end = read_only_free = read_only_space_free_pointer;
     static_end = static_free = static_space_free_pointer;
@@ -658,9 +667,7 @@ purify(lispobj static_roots, lispobj read_only_roots)
     printf(" handlers");
     fflush(stdout);
 #endif
-    pscav((lispobj *) interrupt_handlers,
-          sizeof(interrupt_handlers) / sizeof(lispobj),
-          0);
+    pscav(lisp_sig_handlers, NSIG, 0);
 
 #ifdef PRINTNOISE
     printf(" stack");
@@ -708,22 +715,14 @@ purify(lispobj static_roots, lispobj read_only_roots)
     do {
         while (clean != static_free)
             clean = pscav(clean, static_free - clean, 0);
-        laters = later_blocks;
+        struct later* laters = later_blocks;
         count = later_count;
         later_blocks = NULL;
         later_count = 0;
         while (laters != NULL) {
-            for (i = 0; i < count; i++) {
-                if (laters->u[i].count == 0) {
-                    ;
-                } else if (laters->u[i].count <= LATERMAXCOUNT) {
-                    pscav(laters->u[i+1].ptr, laters->u[i].count, 1);
-                    i++;
-                } else {
-                    pscav(laters->u[i].ptr, 1, 1);
-                }
-            }
-            next = laters->next;
+            for (i = 0; i < count; i++)
+                pscav(laters->ptr[i], laters->count[i], 1);
+            struct later* next = laters->next;
             free(laters);
             laters = next;
             count = LATERBLOCKSIZE;
@@ -734,9 +733,6 @@ purify(lispobj static_roots, lispobj read_only_roots)
     printf(" cleanup");
     fflush(stdout);
 #endif
-#ifdef LISP_FEATURE_HPUX
-    clear_auto_gc_trigger(); /* restore mmap as it was given by os */
-#endif
 
     os_zero((os_vm_address_t) current_dynamic_space, dynamic_space_size);
 
@@ -751,7 +747,7 @@ purify(lispobj static_roots, lispobj read_only_roots)
     read_only_space_free_pointer = read_only_free;
     static_space_free_pointer = static_free;
 
-    dynamic_space_free_pointer = current_dynamic_space;
+    set_alloc_pointer((lispobj)current_dynamic_space);
     set_auto_gc_trigger(bytes_consed_between_gcs);
 
     /* Blast away instruction cache */
@@ -759,6 +755,7 @@ purify(lispobj static_roots, lispobj read_only_roots)
     os_flush_icache((os_vm_address_t)STATIC_SPACE_START, STATIC_SPACE_SIZE);
 
 #ifdef PRINTNOISE
+    verify_range((lispobj*)READ_ONLY_SPACE_START, read_only_space_free_pointer);
     printf(" done]\n");
     fflush(stdout);
 #endif
