@@ -37,7 +37,7 @@
 #include "interr.h"
 #include "alloc.h"
 #include "gc-internal.h"
-#include "pseudo-atomic.h"
+#include "getallocptr.h"
 #include "interrupt.h"
 #include "lispregs.h"
 
@@ -630,7 +630,8 @@ assert_on_stack(struct thread *th, void *esp)
     if (on_altstack_p(th, esp))
         lose("thread %p: esp on altstack: %p", th, esp);
 #endif
-    lose("thread %p: bogus esp: %p", th, esp);
+    lose("thread %p: bogus esp: %p (range=%p..%p)", th, esp,
+         th->control_stack_start, th->control_stack_end);
 }
 
 // returns 0 if skipped, 1 otherwise
@@ -993,7 +994,7 @@ wake_thread_posix(os_thread_t os_thread)
                                  * transition. */
                                 int status = pthread_kill(os_thread, SIGPIPE);
                                 if (status)
-                                    lose("wake_thread_posix: pthread_kill failed with %d\n",
+                                    lose("wake_thread_posix: pthread_kill failed with %d",
                                          status);
                             }
                             break;
